/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <X11/Intrinsic.h> // for XtRealloc
#include "Attributes.h"
#include "VkFont.h"

#include "g.h"
#include "utils.h"

Attributes::Attributes ()
{
  number_of_items = 0;
  max_number_of_items = 0;
  items = NULL;
}

SCM attributes_new ()
{
  Attributes *a = new Attributes ();
  return gh_ulong2scm ((unsigned long) a);
}


void
Attributes::add (char *key, void *value)
{
  int i;

  for (i=0; i<number_of_items; i++)
    if (items[i].type == at_user_defined
	&& !strcmp (key, items[i].key))
      {
	items[i].value = value;
	return;
      }


  while (number_of_items >= max_number_of_items)
    {
      items = (Attribute*)XtRealloc ((char*)items,
				     sizeof (Attribute) *
				     (max_number_of_items + 1) * 2);

      max_number_of_items += 1;
      max_number_of_items *= 2;
    }

  items[number_of_items].key = strdup (key);
  items[number_of_items].value = value;
  items[number_of_items].type = at_user_defined;

  number_of_items ++;
}

SCM attributes_add_s (SCM scm_a, SCM scm_key, SCM scm_value)
{
  Attributes *a = (Attributes *) gh_scm2ulong (scm_a);
  char *key = gh_scm2string (scm_key);
  void *value = (void *) gh_scm2ulong (scm_value);
  a->add (key, value);

  /* free key? */

  return SCM_BOOL_T;
}


void
Attributes::remove (Attr type)
{
  int i;
  int j;

  for (i=0; i<number_of_items; i++)
    if (items[i].type == type)
      {
	for (j=i; j<number_of_items - 1; j++)
	  {
	    items[j] = items[j+1];
	  }
	number_of_items --;
	break;
      }
}

SCM attributes_remove_i (SCM scm_a, SCM scm_type)
{
  printf ("ack\n");

  Attributes *a = (Attributes *) gh_scm2ulong (scm_a);
  Attr type = (Attr) gh_scm2ulong (scm_type);
  a->remove (type);

  printf ("/ack\n");

  return SCM_BOOL_T;
}



void
Attributes::add (Attr type, void *value)
{
  int i;

  for (i=0; i<number_of_items; i++)
    if (items[i].type == type)
      {
	items[i].value = value;
	return;
      }


  //while (number_of_items >= max_number_of_items)
  if (number_of_items >= max_number_of_items)
    {
      max_number_of_items = number_of_items + 4;

      items = (Attribute*) XtRealloc ((char*)items,
				      sizeof (Attribute) *
				      max_number_of_items);
    }

  items[number_of_items].key = NULL;
  items[number_of_items].value = value;
  items[number_of_items].type = type;

  number_of_items ++;
}

SCM attributes_add_i (SCM scm_a, SCM scm_type, SCM scm_value)
{
  Attributes *a = (Attributes *) gh_scm2ulong (scm_a);
  Attr type = (Attr) gh_scm2ulong (scm_type);
  void *value = (void *) gh_scm2ulong (scm_value);
  a->add (type, value);

  return SCM_BOOL_T;
}


int
Attributes::get (char *key, void **ret)
{
  int i;

  for (i=0; i<number_of_items; i++)
    if (items[ i ].type == at_user_defined)
      if (!strcmp (key, items[i].key))
	{
	  *ret= items[i].value;
	  return 1;
	}

  return 0;
}

SCM attributes_get_s (SCM scm_a, SCM scm_key)
{
  void *ret;
  Attributes *a = (Attributes *) gh_scm2ulong (scm_a);
  char *key = gh_scm2string (scm_key);

  if (a->get (key, &ret) == 0)
    return SCM_BOOL_F;

  /* free key? */

  return gh_ulong2scm ((unsigned long) ret);
}



int
Attributes::get (Attr type, void **ret)
{
  int i;

  for (i=0; i<number_of_items; i++)
    if (items[i].type == type)
      {
	*ret = items[i].value;
	return 1;
      }
    
  return 0;
}

SCM attributes_get_i (SCM scm_a, SCM scm_type)
{
  void *ret;
  Attributes *a = (Attributes *) gh_scm2ulong (scm_a);
  Attr type = (Attr) gh_scm2ulong (scm_type);


  if (a->get (type, &ret) == 0)
    return SCM_BOOL_F;

  return gh_ulong2scm ((unsigned long) ret);
}



int
Attributes::count (void)
{
  return number_of_items;
}


SCM attributes_count (SCM scm_a)
{
  Attributes *a = (Attributes *) gh_scm2ulong (scm_a);
  return gh_ulong2scm ((unsigned long) a->count ());
}


Attr *
Attributes::list_keys (void)
{
  Attr *attr_list = (Attr *) malloc (sizeof (Attr) * number_of_items);

  int i;

  for (i=0; i<number_of_items; i++)
    attr_list[ i ] = items[i].type;
    
  return attr_list;
}


/* FINISH ME ... add scheme call to list_attribute_types */


void hook_up_guile_Attributes ()
{
  gh_new_procedure0_0 ("attributes.new", (ghp) attributes_new);
  gh_new_procedure3_0 ("attributes.add-s", (ghp) attributes_add_s);
  gh_new_procedure2_0 ("attributes.remove-i", (ghp) attributes_remove_i);
  gh_new_procedure3_0 ("attributes.add-i", (ghp) attributes_add_i);
  gh_new_procedure2_0 ("attributes.get-s", (ghp) attributes_get_s);
  gh_new_procedure2_0 ("attributes.get-i", (ghp) attributes_get_i);
  gh_new_procedure1_0 ("attributes.count", (ghp) attributes_count);


  /* keep this stuff synced to the C header file Attributes.h */

  /* keys */

  gh_eval_str ("(define at-font 1)");
  gh_eval_str ("(define at-background 2)");
  gh_eval_str ("(define at-foreground 3)");
  gh_eval_str ("(define at-fill-style 4)");
  gh_eval_str ("(define at-direction 5)");
  gh_eval_str ("(define at-overflow-policy 6)");
  gh_eval_str ("(define at-fixed-left 7)");
  gh_eval_str ("(define at-fixed-up 8)");
  gh_eval_str ("(define at-fixed-right 9)");
  gh_eval_str ("(define at-fixed-down 10)");
  gh_eval_str ("(define at-hl-start 11)");
  gh_eval_str ("(define at-hl-end 12)");
  gh_eval_str ("(define at-tag 13)");
  gh_eval_str ("(define at-margin-low 14)");
  gh_eval_str ("(define at-margin-high 15)");
  gh_eval_str ("(define at-show-margins 16)");
  gh_eval_str ("(define at-doc-file 17)");
  gh_eval_str ("(define at-user-defined 18)");


  /* values */

  gh_eval_str ("(define FS-low 1)");
  gh_eval_str ("(define FS-high 2)");
  gh_eval_str ("(define FS-center 3)");
  gh_eval_str ("(define FS-fill 4)");

  gh_eval_str ("(define DIR-lr 5)"); /* left to right */
  gh_eval_str ("(define DIR-ud 6)"); /* up to down */
  gh_eval_str ("(define DIR-rl 7)"); /* right to left */
  gh_eval_str ("(define DIR-du 8)"); /* down to up */

  gh_eval_str ("(define SP-be-as-large-as-allowed 9)");
  gh_eval_str ("(define SP-stay-the-same-size 10)");
  gh_eval_str ("(define SP-stay-small 11)");

  gh_eval_str ("(define OP-push-to-next 12)");
  gh_eval_str ("(define OP-create-a-new-container 13)");

  gh_eval_str ("(define TAG-unknown 14)");
  gh_eval_str ("(define TAG-br 15)");
  gh_eval_str ("(define TAG-hr 16)");
  gh_eval_str ("(define TAG-img 17)");
}
