/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#define FOO 1

#include "XWord.h"
#include "Container.h"
#include "utils.h"
#include "g.h"


Size ZeroSize = {0, 0, 0, 0};


void tabber (int n)
{
  while (n>0)
    {
      printf ("   ");
      n--;
    }
}


Size size_diff (Size a, Size b)
{
  Size c;

  c.left = a.left - b.left;
  c.up = a.up - b.up;
  c.right = a.right - b.right;
  c.down = a.down - b.down;

  return c;
}


int size_compare (Size a, Size b)
{
  if (a.left == b.left &&
      a.up == b.up &&
      a.right == b.right &&
      a.down == b.down)
    return 0;

  return 1;
}


/***************************************************************************/
/***************************************************************************/
/***************************************************************************/
/***************************************************************************/


Container::Container (Attributes *attr)
{
  unsigned long int fixed_left, fixed_up, fixed_right, fixed_down;

  _parent = NULL;
  c_con = NULL;
  /*byte_width = sizeof (Container *);*/
  num_ch = 0;
  max_ch = 0;

  myAttributes = attr;
  tmpAttributes = NULL;

  // if there is no attribute of type FIXED_xxxxx, getAttribute will
  // return a NULL, which will get turned into a zero size.

  if (myAttributes->get (at_fixed_left, (void **) &fixed_left))
    _size.left = fixed_left;
  if (myAttributes->get (at_fixed_up, (void **) &fixed_up))
    _size.up = fixed_up;
  if (myAttributes->get (at_fixed_right, (void **) &fixed_right))
    _size.right = fixed_right;
  if (myAttributes->get (at_fixed_down, (void **) &fixed_down))
    _size.down = fixed_down;
}


SCM container_new (SCM scm_attr)
{
  Attributes *attr = (Attributes *) gh_scm2ulong (scm_attr);
  Container *c = new Container (attr);
  return gh_ulong2scm ((unsigned long) c);
}


Container *
Container::new_of_my_own_type (void)
{
  Container *c = new Container (myAttributes);
  c->tmpAttributes = tmpAttributes;
  return c;
}


SCM container_new_of_my_own_type (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_ulong2scm ((unsigned long) c->new_of_my_own_type ());
}


Container *
Container::dup ()
{
  int i;
  Container *c = new_of_my_own_type ();
  Container *d;

  for (i=0; i<num_ch; i++)
    {
      d = c_con[ i ]->dup ();
      c->Insert_Children (i, 1, &d);
    }

  return c;
}


Container::~Container ()
{
  // do nothing...
}



void Container::ready_for_index (int n)
{
  if (n >= max_ch)
    {
      max_ch = n + 2;
      c_con = (Container **) XtRealloc ((char *) c_con,
					sizeof (Container *) * max_ch);
      if (c_con == NULL)
	{
	  perror ("realloc");
	  fprintf (stderr,
		   "Container couldn't resize children array. bomb.\n");
	  exit (1);
	}
    }
}


Container *Container::first_lowest_container (void)
{
  if (num_ch == 0)
    return this;

  return c_con[ 0 ]->first_lowest_container ();
}


SCM container_first_lowest_container (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_ulong2scm ((unsigned long) c->first_lowest_container ());
}


Container *Container::last_lowest_container (void)
{
  if (num_ch == 0)
    return this;

  return c_con[ num_ch - 1 ]->last_lowest_container ();
}


SCM container_last_lowest_container (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_ulong2scm ((unsigned long) c->last_lowest_container ());
}


Container *
Container::top_of_tree (void)
{
  if (_parent == NULL)
    return this;

  return _parent->top_of_tree ();
}


SCM container_top_of_tree (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_ulong2scm ((unsigned long) c->top_of_tree ());
}


int
Container::are_you_first (void)
{
  int r;

  if (num_ch == 0) return 0;

  r = c_con[ 0 ]->are_you_first () - 1;

  if (r > 0) return r - 1;

  return 0;
}


SCM container_are_you_first (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_long2scm (c->are_you_first ());
}


int
Container::are_you_last (void)
{
  int r;

  if (num_ch == 0) return 0;

  r = c_con[ num_ch-1 ]->are_you_last () - 1;

  if (r > 0) return r - 1;

  return 0;
}


SCM container_are_you_last (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_long2scm (c->are_you_last ());
}


int Container::childs_index (Container *child)
{
  int i;

  for (i = 0; i < num_ch; i++)
    if (c_con[ i ] == (Container *) child)
      return i;

  fprintf (stderr, "child not found in childs_index.\n");

  xw_assert (0);
  return 0; /* reboot */
}


SCM container_childs_index (SCM scm_c, SCM scm_child)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  Container *child = (Container *) gh_scm2ulong (scm_child);
  return gh_int2scm (c->childs_index (child));
}


Container *
Container::get_child (int index)
{
  xw_assert (index < num_ch);
  return (Container *)(c_con[ index ]);
};


SCM container_get_child (SCM scm_c, SCM scm_index)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int index = gh_scm2int (scm_index);
  return gh_ulong2scm ((unsigned long) c->get_child (index));
}


SCM container_get_parent (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_ulong2scm ((unsigned long) c->parent ());
}


int
Container::find_depth (void)
{
  if (_parent == NULL)
    return 1;

  return _parent->find_depth () + 1;
}


SCM container_find_depth (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_int2scm (c->find_depth ());
}


Container *
Container::next_node (void)
{
  if (_parent == NULL)
    return NULL;

  return _parent->next_node_up (this, 1);
}


SCM container_next_node (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  Container *ret = c->next_node ();
  if (ret == NULL) return SCM_BOOL_F;
  return gh_ulong2scm ((unsigned long) ret);
}


Container *
Container::next_node_up (Container *child, int depth)
{
  int child_index = childs_index (child);

  if (child_index == num_ch - 1)
    {
      if (_parent == NULL)
	return NULL;

      return _parent->next_node_up (this, depth + 1);
    }

  if (depth == 0)
    return c_con[ child_index+1 ];
  else
    return c_con[ child_index+1 ]->next_node_down (depth-1);
}


Container *
Container::next_node_down (int depth)
{
  if (depth == 0)
    return this;

  if (num_ch == 0)
    {
      Container *nn = next_node ();
      if (nn == NULL)
	{
	  fprintf (stderr, "Container::next_node_down -- no children.\n");
	  return NULL;
	}

      return nn->next_node_down (depth);
    }

  return c_con[ 0 ]->next_node_down (depth - 1);
}


Container *
Container::prev_node (void)
{
  if (_parent == NULL)
    return NULL;

  return _parent->prev_node_up (this, 1);
}


SCM container_prev_node (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  Container *ret = c->prev_node ();
  if (ret == NULL) return SCM_BOOL_F;
  return gh_ulong2scm ((unsigned long) ret);
}


Container *
Container::prev_node_up (Container *child, int depth)
{
  int child_index = childs_index (child);

  if (child_index == 0)
    {
      if (_parent == NULL)
	return NULL;

      return _parent->prev_node_up (this, depth + 1);
    }

  if (depth == 0)
    return c_con[ child_index-1 ];
  else
    return c_con[ child_index-1 ]->
      prev_node_down (depth - 1);
}


Container *
Container::prev_node_down (int depth)
{
  if (depth == 0)
    return this;

  if (num_ch == 0)
    {
      Container *pn = next_node ();
      if (pn == NULL)
	{
	  fprintf (stderr, "Container::prev_node_down -- no children.\n");
	  return NULL;
	}

      return pn->prev_node_down (depth);
    }

  return c_con[ num_ch-1 ]->prev_node_down (depth-1);
}


/*** call hop_back_position whenever you need to assume that ***/
/*** a position is not at the very start of a container. ***/

Pos
Container::hop_back_position (Pos p)
{
  Container *cur = p.con;
  Container *prev;

  if (p.offset > 0)
    return p;

  prev = cur->prev_node ();

  if (prev != NULL)
    {
      p.con = prev;
      p.offset = prev->num_ch;
      return p;
    }

  // /* insert an empty container before this one? FINISH ME */
  // xw_assert (0);

  return p;
}


SCM container_hop_back_position (SCM scm_c, SCM scm_offset)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int offset = gh_scm2int (scm_offset);
  Pos p;

  p.con = c;
  p.offset = offset;
  p = c->hop_back_position (p);

  return gh_list (gh_ulong2scm ((unsigned long) p.con),
		  gh_int2scm (p.offset), SCM_UNDEFINED);
}

/*********************************************************************/


void Container::dd (int depth) /* debug_dump */
{
  int i;

  tabber (depth);
  printf ("%s-%p (%d,%d,%d,%d) par=%p %d children\n",
	  type (),
	  this,
	  (int)_size.left,
	  (int)_size.up,
	  (int)_size.right,
	  (int)_size.down,
	  _parent,
	  num_ch
	  );

  for (i=0; i<num_ch; i++)
    {
      if (c_con[ i ]->_parent != this)
	{
	  printf ("ERROR -- parent doesn't match this.\n");
	}
      c_con[ i ]->dd (depth+1);
    }
}


/* debug dump */

SCM container_dd (SCM scm_c, SCM scm_depth)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int depth = gh_scm2int (scm_depth);
  c->dd (depth);
  return SCM_BOOL_T;
}



/********************************************************************/

Size
Container::calculate_size (XDisplay */*xd*/,
			   int *constraint_failure, Size *left_over)
{
  Size child_size;
  Attr_Direction direction;
  unsigned long int fixed_left, fixed_up, fixed_right, fixed_down;
  unsigned long int low_margin, high_margin;

  direction = (Attr_Direction) getAttribute (at_direction);

  if (myAttributes->get (at_margin_low, (void **) &low_margin) == 0)
    low_margin = 0;
  if (myAttributes->get (at_margin_high, (void **) &high_margin) == 0)
    high_margin = 0;

  Size the_new_size = ZeroSize;
  (*constraint_failure) = 0;


  switch (direction)
    {
    case DIR_lr:
      the_new_size.right += (low_margin + high_margin);
      break;
    case DIR_ud:
      the_new_size.down += (low_margin + high_margin);
      break;
    case DIR_rl:
      the_new_size.left += (low_margin + high_margin);
      break;
    case DIR_du:
      the_new_size.up += (low_margin + high_margin);
      break;
    }


  for (int lop=0; lop<num_ch; lop++)
    {
      if (lop < num_ch - 1 && c_con[ lop ]->are_you_last() > 0)
	(*constraint_failure) = 1;

      if (lop > 0 && c_con[ lop ]->are_you_first() > 0)
	(*constraint_failure) = 1;

      child_size = c_con[ lop ]->size ();

      switch (direction)
	{
	case DIR_lr:
	  the_new_size.right += child_size.left;
	  the_new_size.right += child_size.right;
	  if (the_new_size.up < child_size.up)
	    the_new_size.up = child_size.up;
	  if (the_new_size.down < child_size.down)
	    the_new_size.down = child_size.down;
	  break;
	case DIR_ud:
	  the_new_size.down += child_size.up;
	  the_new_size.down += child_size.down;
	  if (the_new_size.left < child_size.left)
	    the_new_size.left = child_size.left;
	  if (the_new_size.right < child_size.right)
	    the_new_size.right = child_size.right;
	  break;
	case DIR_rl:
	  the_new_size.left += child_size.left;
	  the_new_size.left += child_size.right;
	  if (the_new_size.up < child_size.up)
	    the_new_size.up = child_size.up;
	  if (the_new_size.down < child_size.down)
	    the_new_size.down = child_size.down;
	  break;
	case DIR_du:
	  the_new_size.up += child_size.up;
	  the_new_size.up += child_size.down;
	  if (the_new_size.left < child_size.left)
	    the_new_size.left = child_size.left;
	  if (the_new_size.right < child_size.right)
	    the_new_size.right = child_size.right;
	  break;
	default:
	  xw_assert (0);
	  break;
	}
    }


  if (myAttributes->get (at_fixed_left, (void **) &fixed_left))
    {
      if (the_new_size.left > fixed_left) (*constraint_failure) = 1;
      the_new_size.left = fixed_left;
    }

  if (myAttributes->get (at_fixed_up, (void **) &fixed_up))
    {
      if (the_new_size.up > fixed_up) (*constraint_failure) = 1;
      the_new_size.up = fixed_up;
    }

  if (myAttributes->get (at_fixed_right, (void **) &fixed_right))
    {
      if (the_new_size.right > fixed_right) (*constraint_failure) = 1;
      the_new_size.right = fixed_right;
    }

  if (myAttributes->get (at_fixed_down, (void **) &fixed_down))
    {
      if (the_new_size.down > fixed_down) (*constraint_failure) = 1;
      the_new_size.down = fixed_down;
    }




  if (left_over != NULL)
    {
      (*left_over) = ZeroSize;
      //left_over->down = _size.down - the_new_size.down;
    }
  

  return the_new_size;
}



/***************************************************************************/
/***************************************************************************/


void *
Container::getAttribute (Attr at)
{
  void *ret = NULL;

  if (myAttributes->get (at, &ret) == 0)
    if (_parent != NULL)
      return _parent->getAttribute (at);

  return ret;
}


void *
Container::getTmpAttribute (Attr at)
{
  void *ret = NULL;

  if (tmpAttributes == NULL)
      return getAttribute (at);

  if (tmpAttributes->get (at, &ret) == 0)
    return getAttribute (at);
  
  return ret;
}



SCM container_getAttribute (SCM scm_c, SCM scm_at)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  Attr at = (Attr) gh_scm2int (scm_at);

  return gh_ulong2scm ((unsigned long) c->getAttribute (at));
}


void Container::setAttributes (Attributes *new_attrs)
{
  myAttributes = new_attrs;
}


SCM container_setAttributes (SCM scm_c, SCM scm_attrs)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  Attributes *attrs = (Attributes *) gh_scm2ulong (scm_attrs);
  c->myAttributes = attrs;
  return SCM_BOOL_T;
}


SCM container_getAttributes (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_ulong2scm ((unsigned long) c->myAttributes);
}


SCM container_set_tmpAttributes (SCM scm_c, SCM scm_attrs)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  Attributes *attrs = (Attributes *) gh_scm2ulong (scm_attrs);
  c->tmpAttributes = attrs;
  return SCM_BOOL_T;
}


SCM container_get_tmpAttributes (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  return gh_ulong2scm ((unsigned long) c->tmpAttributes);
}


void
Container::Insert_Children (int offset, int num, Container **new_children)
{
  int i;

  xw_assert (offset <= num_ch);
  xw_assert (offset >= 0);
  xw_assert (new_children != NULL);
  xw_assert (num >= 0);


  if (num == 0)
    return;

  ready_for_index (num_ch + num);


  for (i = num_ch + num - 1; i > offset; i--)
    c_con[ i ] = c_con[ i - num ];


  for (i = 0; i < num; i++)
    {
      c_con[ i+offset ] = new_children[ i ];
      c_con[ i+offset ]->_parent = this;
    }

  num_ch += num;
}


SCM container_insert_children (SCM scm_c,
			       SCM scm_num,
			       SCM scm_offset,
			       SCM scm_children)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int num = gh_scm2int (scm_num);
  int offset = gh_scm2int (scm_offset);
  Container **children = (Container **) gh_scm2ulong (scm_children);
  c->Insert_Children (offset, num, children);
  return SCM_BOOL_T;
}


SCM container_insert_child (SCM scm_c, SCM scm_offset, SCM scm_child)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int offset = gh_scm2int (scm_offset);
  Container *child = (Container *) gh_scm2ulong (scm_child);
  c->Insert_Children (offset, 1, &child);
  return SCM_BOOL_T;
}


void
Container::Delete_Children (int offset,
			    int num,
			    Container *unhook)
{
  int i;

  xw_assert (offset + num <= num_ch);

  for (i = 0; i < num_ch - num - offset; i++)
    {
      if (unhook)
	unhook->c_con[ unhook->num_ch++ ] = c_con[ i+offset ];

      c_con[ i+offset ] = c_con[ i+offset+num ];
    }

  num_ch -= num;

  /* free some memory */
  max_ch = num_ch + 2;
  c_con = (Container **) XtRealloc ((char *) c_con,
				    sizeof (Container *) * max_ch);
}


SCM container_delete_children (SCM scm_c,
			       SCM scm_offset,
			       SCM scm_num,
			       SCM scm_unhook)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int offset = gh_scm2int (scm_offset);
  int num = gh_scm2int (scm_num);
  Container *unhook = (Container *) gh_scm2ulong (scm_unhook);
  c->Delete_Children (offset, num, unhook);
  return SCM_BOOL_T;
}


void
Container::Unhook_Child (Container *child,
			 Container *unhook,
			 Pos **positions)
{
  xw_assert (child != NULL);

  int c_idx = childs_index (child);

  xw_assert (c_idx >= 0);
  xw_assert (c_idx < num_ch);

  /*** make sure we aren't clobbering one of the positions ***/

  for (int i=0; positions[ i ] != NULL; i++)
    {
      if (positions[ i ]->con == child)
	{
	  Container *p = positions[ i ]->con->prev_node ();
	  if (p == NULL)
	    {
	      Container *n = positions[ i ]->con->next_node ();
	      xw_assert (n != NULL); /* bad. */
	      positions[ i ]->con = n;
	      positions[ i ]->offset = 0;
	    }
	  else
	    {
	      positions[ i ]->con = p;
	      positions[ i ]->offset = p->num_ch;
	    }
	}
    }

  Delete_Children (c_idx, 1, unhook);
}




void
Container::Bump_Children (int cut_off)
{
  Container *next;

  int num_to_bump = num_ch - cut_off;

  xw_assert (num_ch >= 0);
  xw_assert (cut_off <= num_ch);
  xw_assert (num_to_bump < num_ch);

  if (cut_off == num_ch)
    return;

  /*** kick out some children ***/

  if ((Attr_OverFlowPolicy) getAttribute (at_overflow_policy) ==
      OP_push_to_next)
    {
      next = next_node ();
      if (next != NULL)
	if (myAttributes != next->myAttributes)
	  next = NULL;
    }
  else /* OP_create_a_new_container */
    next = NULL;

  if (next == NULL)
    {
      /* create a new container of this one's type */
      next = new_of_my_own_type ();
      xw_assert (_parent != NULL);

      _parent->Insert_Children (_parent->childs_index (this) + 1,
				1,
				&next);
    }
  
  /*** insert all of the extras into ***/
  /*** the start of the next container ***/

  next->Insert_Children (0, num_to_bump,
			 (Container **)
			 (((char *) c_con) + byte_width() * cut_off)
			 );

  num_ch -= num_to_bump;

  /* free some memory */
  max_ch = num_ch + 2;
  c_con = (Container **) XtRealloc ((char *) c_con,
				    sizeof (Container *) * max_ch);
}


void
Container::Wrap_Layer (XDisplay *xd,
		       Container *child_start,
		       Container *child_end,
		       Pos **positions,
		       Container **redraw_container)
{
  int cf; /* constraint_failure */
  Container *parent_start=NULL;
  Container *parent_end, *push_parent_end, *pull_parent_end;
  Container *cur_child;
  Container *del_up_through;
  int changed = 0;
  Size save_size = _size;

  if (child_start == NULL || child_end == NULL)
    return;

  parent_start = child_start->parent ();
  push_parent_end = pull_parent_end = parent_end = child_end->parent ();

  if (parent_start == NULL || parent_end == NULL)
    {
      xw_assert (child_start == child_end);
      child_start->_size = child_start->calculate_size (xd, &cf, NULL);
      xw_assert (cf == 0);
      return;
    }


  /* a container should be redrawn after a wrap if it doesn't change,
     but some of its children change.  By change, I mean: change size
     or participate in a push or a pull.

     FINISH ME  -- still drawing too much -- need to follow the above rule.
     */


  for (cur_child = child_start;
       cur_child != child_end->next_node ();
       cur_child = cur_child->next_node ())
    {
      int save_num_ch = cur_child->num_ch;
      Wrap_Push (xd, cur_child, &child_end);

      push_parent_end = child_end->parent ();

      if (cur_child->num_ch > save_num_ch)
	{
	  cur_child->num_ch = save_num_ch;
	  del_up_through = Wrap_Pull (xd,
				      cur_child,
				      &child_end, &pull_parent_end);

	  if (del_up_through != NULL)
	    {
	      Container *cur_del;
	      Container *cur_del_next = NULL;
	      Container *del_up_through_next = del_up_through->next_node ();
	      
	      for (cur_del = cur_child->next_node ();
		   cur_del != NULL &&
		   cur_del != del_up_through_next;
		   cur_del = cur_del_next)
		{
		  cur_del_next = cur_del->next_node ();
		  cur_del->parent ()->Unhook_Child (cur_del,
						    NULL,
						    positions);
		}

	      if (del_up_through == child_end)
		break;
	    }
	}

      if (cur_child->num_ch != save_num_ch)
	changed = 1;
    }


  if (redraw_container != NULL)
    {
      if (child_start == child_end &&
	  changed == 0 &&
	  (*redraw_container)==NULL &&
	  size_compare (save_size,_size) == 0)
	{
	  (*redraw_container) = child_start;
	}
    }


  /*** figure out which reached farther... the push or the pull ***/
  for (parent_end=parent_start;;parent_end = parent_end->next_node ())
    {
      if (parent_end == push_parent_end)
	{
	  parent_end = pull_parent_end;
	  break;
	}
      if (parent_end == pull_parent_end)
	{
	  parent_end = push_parent_end;
	  break;
	}
    }

  parent_start->Wrap_Layer (xd,
			    parent_start,
			    parent_end,
			    positions,
			    redraw_container);
}


void
Container::Wrap_Push (XDisplay *xd,
		      Container *cur_child, Container **child_end)
{
  int cf; /* constraint_failure */
  int save_num_ch = cur_child->num_ch;

#if FOO
  for (cur_child->num_ch = 0;
#else
  for (cur_child->num_ch = 1;
#endif
       cur_child->num_ch <= save_num_ch;
       cur_child->num_ch ++)
    {
      /*** push... ***/
#if FOO
      cur_child->_size = cur_child->calculate_size (xd, &cf, NULL);
#else
      Container *climb;
      for (climb = cur_child; climb != NULL; climb = climb->_parent)
	{
	  climb->_size = climb->calculate_size (XDisplay *xd, &cf, NULL);
	  if (cf == 1)
		break;
	}
#endif

      if (cf == 1)
	{
	  int break_point;

#if FOO
	  xw_assert (cur_child->num_ch > 0);
#else
	  if (cur_child->num_ch == 0)
	    {
	      /*** re-adjust the bounds of the wrap ***/
	      if (cur_child == (*child_end))
		(*child_end) = cur_child->next_node ();
	      if ((*child_end) == NULL)
		(*child_end) = cur_child;

	      cur_child->_size = cur_child->calculate_size (XDisplay *xd,
							    &cf, NULL);
	      break;
	    }
#endif

	  break_point = cur_child->num_ch - 1;

	  if (break_point < 1) /* word longer than line? */
	    {
	      break_point = 1;
	      xw_assert (cur_child->num_ch > 0);
	    }

	  cur_child->num_ch = save_num_ch;
	  cur_child->Bump_Children (break_point);

	  /*** re-adjust the bounds of the wrap ***/
	  if (cur_child == (*child_end))
	    (*child_end) = cur_child->next_node ();
	  if ((*child_end) == NULL)
	    (*child_end) = cur_child;

	  cur_child->_size = cur_child->calculate_size (xd, &cf, NULL);
	  // xw_assert (cf == 0); this will fail when word longer than line.
	  break;
	}
    }
}


Container *
Container::Wrap_Pull (XDisplay *xd,
		      Container *cur_child,
		      Container **child_end,
		      Container **parent_end)
{
  int cf; /* constraint_failure */
  Container *pull_from;
  Container *del_up_through = NULL;

  for (pull_from = cur_child->next_node ();
       pull_from != NULL;
       pull_from = pull_from->next_node ())
    {
      Container *hop;

      if (cur_child->myAttributes != pull_from->myAttributes)
	{
	  if (cur_child->type_n == pull_from->type_n)
	    {
	      if (cur_child->num_ch == 0)
		cur_child->myAttributes = pull_from->myAttributes;
	      else if (pull_from->num_ch == 0)
		pull_from->myAttributes = cur_child->myAttributes;
	      else
		break;
	    }
	  else
	    break;
	}

      (*parent_end) = pull_from->parent();

      while (pull_from->num_ch > 0)
	{
	  /*** move a child from pull_from to cur_child ***/
	  hop = pull_from->get_child (0);
	  pull_from->Delete_Children (0, 1, NULL);
	  cur_child->Insert_Children (cur_child->num_ch, 1, &hop);

	  /*** see if it fits ***/
#if FOO
	  cur_child->_size = cur_child->calculate_size (xd, &cf, NULL);
#else
	  Container *climb;
	  for (climb = cur_child; climb != NULL; climb = climb->_parent)
	    {
	      climb->_size = climb->calculate_size (XDisplay *xd, &cf, NULL);
	      if (cf == 1)
		break;
	    }
#endif

	  if (cf == 1)
	    {
	      /*** undo the transfer -- didn't fit ***/
	      pull_from->Insert_Children (0, 1, &hop);
	      cur_child->Delete_Children (cur_child->num_ch - 1, 1, NULL);

	      cur_child->_size = cur_child->calculate_size (xd, &cf, NULL);
	      xw_assert (cf == 0);

	      return del_up_through;
	    }

	  /*** if pull_from is empty, delete it ***/
	  if (pull_from->num_ch == 0)
	    del_up_through = pull_from;

	  /*** re-adjust the bounds of the wrap ***/
	  if (pull_from == (*child_end)->next_node() && pull_from != NULL)
	    (*child_end) = pull_from;
	}

      /*** if pull_from is empty, delete it ***/
      if (pull_from->num_ch == 0)
	del_up_through = pull_from;
    }

  return del_up_through;
}


void
Container::Correct_Positions (Pos **positions)
{
  Container *c, *cur;

  for (int i=0; positions[ i ] != NULL; i++)
    {
      cur = positions[ i ]->con;

      while (positions[ i ]->offset > cur->num_ch)
	{
	  positions[ i ]->offset -= cur->num_ch;
	  c = cur->next_node ();
	  if (c == NULL)
	    positions[ i ]->offset = cur->num_ch;
	  else
	    positions[ i ]->con = c;
	}
      
      while (positions[ i ]->offset < 0)
	{
	  c = cur->prev_node ();
	  if (c == NULL)
	    positions[ i ]->offset = 0;
	  else
	    positions[ i ]->con = c;
	  
	  positions[ i ]->offset += c->num_ch;
	}
    }
}


int
Container::X_Erase (XDisplay *xd,
		    Window w,
		    GC gc,
		    int sw, int sh)
{
  int x, y, h;

  get_xy (&x, &y, &h);
  return X_Erase (xd, w, gc, x, y, sw, sh);
}



int
Container::X_Erase (XDisplay *xd,
		    Window w,
		    GC gc,
		    int x, int y,
		    int sw, int sh)
{
  int width, height;

  width = _size.left + _size.right;
  height = _size.up + _size.down;

  /* clip to the window size */
  if (x < 0) {width += x; x = 0;}
  if (y < 0) {height += y; y = 0;}
  if (width > sw) {width = sw;}
  if (height > sh) {height = sh;}

  XSetForeground (xd->display, gc,
		  (Pixel) getAttribute (at_background));
  
  XFillRectangle (xd->display, w, gc,
		  (int) (x - _size.left), (int) (y - _size.up),
		  width, height);

  return 1;
}


int
Container::X_Draw (XDisplay *xd,
		   Window w,
		   GC gc,
		   int sw, int sh)
{
  int x, y, h;
  get_xy (&x, &y, &h);

  return X_Draw (xd, w, gc, x, y, sw, sh);
}


int
Container::Size_Major_Axis (Attr_Direction direction)
{
  int i;
  int lt, up, rt, dn;
  int len=0;

  for (i=0; i<num_ch; i++)
    {
      lt = c_con[i]->size_left ();
      up = c_con[i]->size_up ();
      rt = c_con[i]->size_right ();
      dn = c_con[i]->size_down ();

      switch (direction)
	{
	case DIR_lr:
	case DIR_rl:
	  len += (lt + rt);
	  break;
	case DIR_ud:
	case DIR_du:
	  len += (up + dn);
	  break;
	default:
	  xw_assert (0);
	  break;
	}
    }

  return len;
}



void
Container::handle_fill (Attr_Direction direction, Attr_FillStyle fill,
			int *init_step, int *extra_step)
{
  int extra_space = 0;
  int maj;
  int my_len=0;

  unsigned long int low_margin, high_margin;

  if (myAttributes->get (at_margin_low, (void **) &low_margin) == 0)
    low_margin = 0;
  if (myAttributes->get (at_margin_high, (void **) &high_margin) == 0)
    high_margin = 0;

  (*init_step) = low_margin;
  (*extra_step) = 0;

  if (fill == FS_low) /* do nothing */
    return;

  maj = Size_Major_Axis (direction);
  maj += low_margin + high_margin;

  switch (direction)
    {
    case DIR_lr:
    case DIR_rl:
      my_len = size_left () + size_right ();
      //my_len = _size.left + _size.right;
      break;
    case DIR_ud:
    case DIR_du:
      my_len = size_up () + size_down ();
      //my_len = _size.up + _size.down;
      break;
    }

  extra_space = my_len - maj;
  xw_assert (extra_space >= 0);

  switch (fill)
    {
    case FS_low:
      xw_assert (0);
      break;
    case FS_high:
      (*init_step) += extra_space;
      break;
    case FS_center:
      (*init_step) += extra_space/2;
      break;
    case FS_fill:
      if (num_ch > 1)
	(*extra_step) = extra_space / (num_ch - 1);
      break;
    default:
      xw_assert (0);
      break;
    }
}


SCM container_handle_fill (SCM scm_c)
{
  Container *c = (Container *) gh_scm2ulong (scm_c);
  Attr_Direction direction = (Attr_Direction) c->getAttribute (at_direction);
  Attr_FillStyle fill = (Attr_FillStyle) c->getAttribute (at_fill_style);
  int init_step, extra_step;
  SCM scm_init_step, scm_extra_step;


  c->handle_fill (direction, fill, &init_step, &extra_step);

  scm_init_step = gh_int2scm (init_step);
  scm_extra_step = gh_int2scm (extra_step);

  return gh_list (scm_init_step, scm_extra_step, SCM_UNDEFINED);
}


int
Container::X_Draw (XDisplay *xd,
		   Window w,
		   GC gc,
		   int x, int y,
		   int sw, int sh)
{
  int i;
  int lt, up, rt, dn;
  Attr_Direction direction;
  Attr_FillStyle fill;
  int init_step;
  int extra_step;
  unsigned long int low_margin, high_margin;


  if (x - (int)(_size.left) > sw ||
      y - (int)(_size.up) > sh ||
      x + (int)(_size.right) < 0 ||
      y + (int)(_size.down) < 0)
    return 0;  // out of the window bounds.


  direction = (Attr_Direction) getAttribute (at_direction);
  fill = (Attr_FillStyle) getAttribute (at_fill_style);


  /* draw margins */

  if ((int)(getAttribute (at_show_margins)))
    {
      if (myAttributes->get (at_margin_low, (void **) &low_margin) > 0 &&
	  myAttributes->get (at_margin_high, (void **) &high_margin) > 0)
	{
	  XSetForeground (xd->display, xd->gc,
			  xd->rgb_to_pixel (0x7FFF, 0x7FFF, 0x7FFF));

	  switch (direction)
	    {
	    case DIR_lr:
	    case DIR_rl:
	      XDrawLine (xd->display, w, xd->gc,
			 x - _size.left + low_margin - 1,
			 y - _size.up,
			 x - _size.left + low_margin - 1,
			 y + _size.down);
	      
	      XDrawLine (xd->display, w, xd->gc,
			 x + _size.right - high_margin + 1,
			 y - _size.up,
			 x + _size.right - high_margin + 1,
			 y + _size.down);
	      break;
	    case DIR_ud:
	    case DIR_du:
	      XDrawLine (xd->display, w, xd->gc,
			 x - _size.left,
			 y - _size.up + low_margin - 1,
			 x + _size.right,
			 y - _size.up + low_margin - 1);
	      
	      XDrawLine (xd->display, w, xd->gc,
			 x - _size.left,
			 y + _size.down - high_margin + 1,
			 x + _size.right,
			 y + _size.down - high_margin + 1);
	      break;
	    }
	}
    }


  /* draw children */

  handle_fill (direction, fill, &init_step, &extra_step);
  switch (direction)
    {
    case DIR_lr: x += init_step; break;
    case DIR_ud: y += init_step; break;
    case DIR_rl: x -= init_step; break;
    case DIR_du: y -= init_step; break;
    default: xw_assert (0); break;
    }


  for (i=0; i<num_ch; i++)
    {
      lt = c_con[i]->size_left ();
      up = c_con[i]->size_up ();
      rt = c_con[i]->size_right ();
      dn = c_con[i]->size_down ();

      switch (direction)
	{
	case DIR_lr:
	  x += lt;
	  c_con[i]->X_Draw(xd, w, gc, x, y, sw, sh);
	  x += rt;
	  x += extra_step;
	  break;
	case DIR_ud:
	  y += up;
	  c_con[i]->X_Draw(xd, w, gc, x, y, sw, sh);
	  y += dn;
	  y += extra_step;
	  break;
	case DIR_rl:
	  x -= rt;
	  c_con[i]->X_Draw(xd, w, gc, x, y, sw, sh);
	  x -= lt;
	  x -= extra_step;
	  break;
	case DIR_du:
	  y -= dn;
	  c_con[i]->X_Draw(xd, w, gc, x, y, sw, sh);
	  y -= up;
	  y -= extra_step;
	  break;
	default:
	  xw_assert (0);
	  break;
	}
    }

  return 1; // one because something was drawn.
}



/* get_xy is called by X_Draw and X_Erase */

void
Container::get_xy (int *x, int *y, int *h)
{
  if (_parent != NULL)
    _parent->offset_to_xy (_parent->childs_index (this), x, y, h);
  else
    {
      (*x) = 0;
      (*y) = 0;
      (*h) = 0;
    }
}


/* offset_to_xy */
/* used to find the x and y position of a child */
/* called by get_xy and get_cursor_xy */

void
Container::offset_to_xy (int offset, int *x, int *y, int *h)
{
  int lt, up, rt, dn;
  int i;
  Attr_Direction direction;
  Attr_FillStyle fill;
  int init_step;
  int extra_step;

  xw_assert (offset <= num_ch);

  direction = (Attr_Direction) getAttribute (at_direction);
  fill = (Attr_FillStyle) getAttribute (at_fill_style);

  if (_parent != NULL)
    _parent->offset_to_xy (_parent->childs_index (this), x, y, h);


  handle_fill (direction, fill, &init_step, &extra_step);
  switch (direction)
    {
    case DIR_lr: (*x) += init_step; break;
    case DIR_ud: (*y) += init_step; break;
    case DIR_rl: (*x) -= init_step; break;
    case DIR_du: (*y) -= init_step; break;
    default: xw_assert (0); break;
    }


  if (offset == num_ch)
    {
      fprintf (stderr, "strange offset_to_xy clause in use...\n");
      switch (direction)
	{
	case DIR_lr: (*x) += _size.right; break;
	case DIR_ud: (*y) += _size.down; break;
	case DIR_rl: (*x) -= _size.right; break;
	case DIR_du: (*y) -= _size.down; break;
	default: xw_assert (0); break;
	}
    }
  else
    {
      
      for (i=0; i<=offset; i++)
	{
	  lt = c_con[i]->size_left ();
	  up = c_con[i]->size_up ();
	  rt = c_con[i]->size_right ();
	  dn = c_con[i]->size_down ();
	  
	  switch (direction)
	    {
	    case DIR_lr: (*x) += lt; break;
	    case DIR_ud: (*y) += up; break;
	    case DIR_rl: (*x) -= rt; break;
	    case DIR_du: (*y) -= dn; break;
	    default: xw_assert (0); break;
	    }
	  
	  if (i == offset)
	    break;
	  
	  switch (direction)
	    {
	    case DIR_lr: (*x) += rt; (*x) += extra_step; break;
	    case DIR_ud: (*y) += dn; (*y) += extra_step; break;
	    case DIR_rl: (*x) -= lt; (*x) -= extra_step; break;
	    case DIR_du: (*y) -= up; (*y) -= extra_step; break;
	    default: xw_assert (0); break;
	    }
	}
    }


  (*h) = _size.up;

//  switch (direction)
//    {
//    case DIR_lr: (*h) = _size.up; break;
//    case DIR_ud: (*h) = _size.down; (*x) += lt; break;
//    case DIR_rl: (*h) = _size.up; break;
//    case DIR_du: (*h) = _size.down; (*x) += lt; break;
//    default: xw_assert (0); break;
//    }

  return;
}



/* xy_to_position is usually used figure out where to put the cursor */
/* after a mouse button click */

Pos
Container::xy_to_position (int x, int y)
{
  Pos p;
  int i;
  int current_x = 0;
  int current_y = 0;
  Attr_Direction direction;
  Attr_FillStyle fill;
  int init_step;
  int extra_step;

  direction = (Attr_Direction) getAttribute (at_direction);
  fill = (Attr_FillStyle) getAttribute (at_fill_style);

  p.con = NULL;
  p.offset = 0;


  handle_fill (direction, fill, &init_step, &extra_step);
  switch (direction)
    {
    case DIR_lr: current_x += init_step; break;
    case DIR_ud: current_y += init_step; break;
    case DIR_rl: current_x -= init_step; break;
    case DIR_du: current_y -= init_step; break;
    default: xw_assert (0); break;
    }

  for (i=0; i<num_ch; i++)
    {
      switch (direction)
	{
	case DIR_lr:
	  current_x += c_con[ i ]->size_left ();
	  current_x += c_con[ i ]->size_right ();
	  if (current_x >= x)
	    {
	      current_x -= c_con[ i ]->size_right ();
	      x -= current_x;
	      return c_con[ i ]->xy_to_position (x, y);
	    }
	  current_x += extra_step;
	  break;
	case DIR_ud:
	  current_y += c_con[ i ]->size_up ();
	  current_y += c_con[ i ]->size_down ();
	  if (current_y >= y)
	    {
	      current_y -= c_con[ i ]->size_down ();
	      y -= current_y;
	      return c_con[ i ]->xy_to_position (x, y);
	    }
	  current_y += extra_step;
	  break;
	case DIR_rl:
	  current_x -= c_con[ i ]->size_right ();
	  current_x -= c_con[ i ]->size_left ();
	  if (current_x <= x)
	    {
	      current_x += c_con[ i ]->size_left ();
	      x -= current_x;
	      return c_con[ i ]->xy_to_position (x, y);
	    }
	  current_x -= extra_step;
	  break;
	case DIR_du:
	  current_y -= c_con[ i ]->size_down ();
	  current_y -= c_con[ i ]->size_up ();
	  if (current_y <= y)
	    {
	      current_y += c_con[ i ]->size_up ();
	      y -= current_y;
	      return c_con[ i ]->xy_to_position (x, y);
	    }
	  current_y -= extra_step;
	  break;
	default:
	  xw_assert (0);
	  break;
	}
    }


  if (num_ch > 0)
    {
      switch (direction)
	{
	case DIR_lr:
	  current_x -= c_con[ num_ch-1 ]->size_right ();
	  break;
	case DIR_ud:
	  current_y -= c_con[ num_ch-1 ]->size_down ();
	  break;
	case DIR_rl:
	  current_x += c_con[ num_ch-1 ]->size_left ();
	  break;
	case DIR_du:
	  current_y += c_con[ num_ch-1 ]->size_up ();
	  break;
	default:
	  xw_assert (0);
	  break;
	}

      return c_con[ num_ch-1 ]->xy_to_position (x - current_x,
						y - current_y);
    }

  return p;
}


void
Container::offset_to_index_array (int offset, int *array, int *len)
{
  xw_assert (array != NULL);
  xw_assert (offset <= num_ch);

  if (_parent != NULL)
    _parent->offset_to_index_array (_parent->childs_index (this),
				    array,
				    len);
  array[ *len ] = offset;
  (*len) ++;
}


Pos
Container::index_array_to_offset (int *array, int len)
{
  xw_assert (array != NULL);
  xw_assert (array[ 0 ] < num_ch);
  xw_assert (len >= 1);
  return c_con[ array[ 0 ] ]->index_array_to_offset (array+1, len-1);
}


SCM
Container::size_leaf_ascii (XDisplay *xd, int *sz)
{
  int i;
  int num = 0;
  int r;

  for (i=0; i<num_ch; i++)
    {
      c_con[ i ]->size_leaf_ascii (xd, &r);
      num += r;
    }

  if (sz != NULL)
    (*sz) = num;

  return gh_int2scm (num);
}


SCM
Container::get_leaf_ascii (XDisplay *xd, char **ret, int *len)
{
  int sz, i, rlen;
  char *scratch, *p, *r;
  SCM scm_str;

  size_leaf_ascii (xd, &sz);
  scratch = (char *) malloc (sz + 1);
  p = scratch;

  for (i=0; i<num_ch; i++)
    {
      c_con[ i ]->get_leaf_ascii (xd, &r, &rlen);
      memcpy (p, r, rlen);
      p += rlen;
    }

  (*p) = '\0'; /* null terminate, just for fun! */

  /* gh_str2scm doesn't make it's own copy. */

  scm_str = gh_str2scm (scratch, sz);

  if (ret != NULL) (*ret) = scratch;
  if (len != NULL) (*len) = sz;

  return scm_str;
}


SCM container_get_leaf_ascii (SCM scm_xd, SCM scm_container)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  return container->get_leaf_ascii (xd, NULL, NULL);
}


SCM
Container::set_leaf_ascii (XDisplay * /*xd*/, char *contents, int /*len*/)
{
  fprintf (stderr, "Warning: Container.set_leaf_ascii got %s\n",
	   contents);
  /* free str?  FIX ME */
  return SCM_BOOL_F;
}


SCM container_set_leaf_ascii (SCM scm_xd, SCM scm_container, SCM scm_contents)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  int len;
  char *contents = gh_scm2newstr (scm_contents, &len);


  return container->set_leaf_ascii (xd, contents, len);
}


SCM
Container::size_leaf_xword (XDisplay *xd, int *sz)
{
  int i;
  int num = 0;
  int r;

  for (i=0; i<num_ch; i++)
    {
      c_con[ i ]->size_leaf_xword (xd, &r);
      num += r;
    }

  if (sz != NULL)
    (*sz) = num;

  return gh_int2scm (num);
}


SCM
Container::get_leaf_xword (XDisplay *xd, char **ret, int *len)
{
  int sz, i, rlen;
  char *scratch, *p, *r;
  SCM scm_str;

  size_leaf_xword (xd, &sz);
  scratch = (char *) malloc (sz + 1);
  p = scratch;

  for (i=0; i<num_ch; i++)
    {
      c_con[ i ]->get_leaf_xword (xd, &r, &rlen);
      memcpy (p, r, rlen);
      p += rlen;
    }

  (*p) = '\0';

  /* gh_str2scm does not make it's own copy. */

  scm_str = gh_str2scm (scratch, sz);

  if (ret != NULL) (*ret) = scratch;
  if (len != NULL) (*len) = sz;

  return scm_str;
}


SCM
Container::set_leaf_xword (XDisplay */*xd*/, char *chars, int /*len*/)
{
  fprintf (stderr, "Warning: Container.set_leaf_xword got %s\n", chars);
  /* free str?  FIX ME */
  return SCM_BOOL_F;
}


SCM container_get_leaf_xword (SCM scm_xd, SCM scm_container)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  return container->get_leaf_xword (xd, NULL, NULL);
}


SCM container_set_leaf_xword (SCM scm_xd, SCM scm_container, SCM scm_chars)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  int len;
  char *chars = gh_scm2newstr (scm_chars, &len);
  return container->set_leaf_xword (xd, chars, len);
}


SCM Container::size_leaf_ps (XDisplay */*xd*/, int *sz)
{
  xw_assert (0);
  *sz = 0;
  return gh_int2scm (*sz);
}


SCM Container::get_leaf_ps (XDisplay */*xd*/, char **/*ret*/, int */*len*/)
{
  xw_assert (0);
  return SCM_BOOL_F;
}


SCM Container::set_leaf_ps (XDisplay */*xd*/, char */*chars*/, int /*len*/)
{
  xw_assert (0);
  return SCM_BOOL_F;
}


SCM container_get_leaf_ps (SCM scm_xd, SCM scm_container)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  return container->get_leaf_ps (xd, NULL, NULL);
}


SCM container_set_leaf_ps (SCM scm_xd, SCM scm_container, SCM scm_chars)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  int len;
  char *chars = gh_scm2newstr (scm_chars, &len);
  return container->set_leaf_ps (xd, chars, len);
}



SCM Container::size_leaf_html (XDisplay */*xd*/, int *sz)
{
  xw_assert (0);
  *sz = 0;
  return gh_int2scm (*sz);
}


SCM Container::get_leaf_html (XDisplay */*xd*/, char **/*ret*/, int */*len*/)
{
  xw_assert (0);
  return SCM_BOOL_F;
}


SCM Container::set_leaf_html (XDisplay */*xd*/, char */*chars*/, int /*len*/)
{
  xw_assert (0);
  return SCM_BOOL_F;
}


SCM container_get_leaf_html (SCM scm_xd, SCM scm_container)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  return container->get_leaf_html (xd, NULL, NULL);
}


SCM container_set_leaf_html (SCM scm_xd, SCM scm_container, SCM scm_chars)
{
  Container *container = (Container *) gh_scm2ulong (scm_container);
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xd);
  int len;
  char *chars = gh_scm2newstr (scm_chars, &len);
  return container->set_leaf_html (xd, chars, len);
}



/* returns 1 if a is before b, 0 otherwise */
SCM Container::find_order (Pos a, Pos b)
{
  Container *ca = a.con;
  Container *cb = b.con;

  xw_assert (ca->find_depth () == cb->find_depth ());

  if (ca == cb)
    return (a.offset < b.offset);

  a.con = a.con->parent ();
  b.con = b.con->parent ();
  a.offset = a.con->childs_index (ca);
  b.offset = b.con->childs_index (cb);

  return find_order (a, b);
}


/* returns #t if a is before b */

SCM container_find_order (SCM scm_ca, SCM scm_ia,
			  SCM scm_cb, SCM scm_ib)
{
  Container *ca = (Container *) gh_scm2ulong (scm_ca);
  Container *cb = (Container *) gh_scm2ulong (scm_cb);
  int ia = gh_scm2int (scm_ia);
  int ib = gh_scm2int (scm_ib);
  Pos a, b;

  a.con = ca;
  a.offset = ia;
  b.con = cb;
  b.offset = ib;

  if (ca->find_order (a, b))
    return SCM_BOOL_T;

  return SCM_BOOL_F;
}




SCM container_get_size (SCM scm_container)
{
  SCM left, up, right, down;
  Container *c = (Container *) gh_scm2ulong (scm_container);

  left = gh_int2scm (c->size_left ());
  up = gh_int2scm (c->size_up ());
  right = gh_int2scm (c->size_right ());
  down = gh_int2scm (c->size_down ());

  return gh_list (left, up, right, down, SCM_UNDEFINED);
}



SCM container_get_num_children (SCM scm_container)
{
  SCM scm_num_ch;
  Container *c = (Container *) gh_scm2ulong (scm_container);
  scm_num_ch = gh_int2scm (c->num_ch);
  return scm_num_ch;
}


SCM container_get_type (SCM scm_container)
{
  Container *c = (Container *) gh_scm2ulong (scm_container);
  return gh_str02scm (c->type());
}


void hook_up_guile_Container ()
{
  gh_new_procedure1_0 ("container.new",
		       (ghp) container_new);
  gh_new_procedure1_0 ("container.new-of-my-own-type",
		       (ghp) container_new_of_my_own_type);
  gh_new_procedure1_0 ("container.first-lowest-container",
		       (ghp) container_first_lowest_container);
  gh_new_procedure1_0 ("container.last-lowest-container",
		       (ghp) container_last_lowest_container);
  gh_new_procedure1_0 ("container.top-of-tree",
		       (ghp) container_top_of_tree);
  gh_new_procedure1_0 ("container.are-you-first",
		       (ghp) container_are_you_first);
  gh_new_procedure1_0 ("container.are-you-last",
		       (ghp) container_are_you_last);
  gh_new_procedure2_0 ("container.childs-index",
		       (ghp) container_childs_index);
  gh_new_procedure2_0 ("container.get-child",
		       (ghp) container_get_child);
  gh_new_procedure1_0 ("container.get-parent",
		       (ghp) container_get_parent);
  gh_new_procedure1_0 ("container.find-depth",
		       (ghp) container_find_depth);
  gh_new_procedure1_0 ("container.next-node",
		       (ghp) container_next_node);
  gh_new_procedure1_0 ("container.prev-node",
		       (ghp) container_prev_node);
  gh_new_procedure2_0 ("container.hop-back-position",
		       (ghp) container_hop_back_position);
  gh_new_procedure2_0 ("container.dd",
		       (ghp) container_dd);
  gh_new_procedure2_0 ("container.get-attribute",
		       (ghp) container_getAttribute);
  gh_new_procedure1_0 ("container.get-attributes",
		       (ghp) container_getAttributes);
  gh_new_procedure2_0 ("container.set-attributes",
		       (ghp) container_setAttributes);
  gh_new_procedure1_0 ("container.get-tmp-attributes",
		       (ghp) container_get_tmpAttributes);
  gh_new_procedure2_0 ("container.set-tmp-attributes",
		       (ghp) container_set_tmpAttributes);
  gh_new_procedure4_0 ("container.insert-children",
		       (ghp) container_insert_children);
  gh_new_procedure3_0 ("container.insert-child",
		       (ghp) container_insert_child);
  gh_new_procedure4_0 ("container.delete-children",
		       (ghp) container_delete_children);
  gh_new_procedure2_0 ("container.get-leaf-ascii",
		       (ghp) container_get_leaf_ascii);
  gh_new_procedure3_0 ("container.set-leaf-ascii",
		       (ghp) container_set_leaf_ascii);
  gh_new_procedure2_0 ("container.get-leaf-xword",
		       (ghp) container_get_leaf_xword);
  gh_new_procedure3_0 ("container.set-leaf-xword",
		       (ghp) container_set_leaf_xword);
  gh_new_procedure2_0 ("container.get-leaf-ps",
		       (ghp) container_get_leaf_ps);
  gh_new_procedure3_0 ("container.set-leaf-ps",
		       (ghp) container_set_leaf_ps);
  gh_new_procedure2_0 ("container.get-leaf-html",
		       (ghp) container_get_leaf_html);
  gh_new_procedure3_0 ("container.set-leaf-html",
		       (ghp) container_set_leaf_html);
  gh_new_procedure1_0 ("container.get-size",
		       (ghp) container_get_size);
  gh_new_procedure1_0 ("container.get-num-children",
		       (ghp) container_get_num_children);
  gh_new_procedure1_0 ("container.get-type",
		       (ghp) container_get_type);
  gh_new_procedure1_0 ("container.handle-fill",
		       (ghp) container_handle_fill);
  gh_new_procedure4_0 ("container.find-order",
		       (ghp) container_find_order);
}
