/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#ifndef CONTAINER_H_INCLUDED
#define CONTAINER_H_INCLUDED

#include "XDisplay.h"
#include "Attributes.h"
#include "Attribute_Types.h"
#include "Assoc.h"
#include "g.h"
#include "tag_info.h"

#include <sys/types.h>
#include <stdlib.h>
/*#include <unistd.h>*/ /* exit problem under irix */

class Container;
class Document;
class Page;
class WordSegment;


typedef struct {
  unsigned int left;
  unsigned int up;
  unsigned int right;
  unsigned int down;
} Size;


typedef enum
{
  T_unknown = 1,
  T_char,
  T_Container,
  T_Document,
  T_Line,
  T_Page,
  T_Word,
  T_WordSegment,
  T_Tag
} Container_Type;


typedef struct
{
  Container *con;
  int offset;
} Pos;


void hook_up_guile_Container ();


class Container
{
  /* object oriented! see... */
 public:

  union
    {
      XChar2b *c_char;
      Container **c_con;
      //char **c_tag;
      tag_info **c_tag;
    };

  int num_ch;	/* number of containers in this array */
  int max_ch;	/* current number that there would be space for */

  /*int byte_width;*/


  Container *_parent;


  Size _size;

  Attributes *myAttributes;
  Attributes *tmpAttributes;


  /**************************************************/


  Container (Attributes *attr);
  virtual ~Container ();

  virtual Container *new_of_my_own_type (void);
  virtual Container *dup ();

  virtual int byte_width () {return sizeof (Container *);}

  virtual Container *first_lowest_container (void);
  virtual Container *last_lowest_container (void);
  virtual Container *top_of_tree (void);

  virtual int are_you_first (void);
  virtual int are_you_last (void);

  virtual void ready_for_index (int n);
  virtual int childs_index (Container *child);
  virtual Container *get_child (int index);
  int find_depth (void);

  SCM find_order (Pos a, Pos b);

  virtual Size size (void) {return _size;};
  virtual unsigned int size_left (void) {return _size.left;};
  virtual unsigned int size_up (void) {return _size.up;};
  virtual unsigned int size_right (void) {return _size.right;};
  virtual unsigned int size_down (void) {return _size.down;};

  Container *next_node (void);
  Container *next_node_up (Container *child, int depth);
  Container *next_node_down (int depth);
  Container *prev_node (void);
  Container *prev_node_up (Container *child, int depth);
  Container *prev_node_down (int depth);

  Pos hop_back_position (Pos p);

  Container *parent (void) {return _parent;};

  virtual void dd (int depth);	/* debug_dump */
    
  virtual Size calculate_size (XDisplay *xd,
			       int *constraint_failure, Size *left_over);
    
  void *getAttribute (Attr at);
  void *getTmpAttribute (Attr at);
  void setAttributes (Attributes *new_attrs);

  virtual char *type (void) {return "Container";};
  virtual Container_Type type_n (void) {return T_Container;};

  virtual void Insert_Children (int offset,
				int num,
				Container **new_children);
  virtual void Delete_Children (int offset, int num, Container *unhook);
  virtual void Unhook_Child (Container *child,
			     Container *unhook,
			     Pos **positions);

  virtual void Bump_Children (int cut_off);

  virtual void Correct_Positions (Pos **positions);
  void Wrap_Layer (XDisplay *xd,
		   Container *child_start,
		   Container *child_end,
		   Pos **positions,
		   Container **redraw_container);
  void Wrap_Push (XDisplay *xd,
		  Container *cur_child, Container **child_end);
  Container *Wrap_Pull (XDisplay *xd,
			Container *cur_child,
			Container **child_end,
			Container **parent_end);


  virtual int Size_Major_Axis (Attr_Direction direction);
  virtual void handle_fill (Attr_Direction direction, Attr_FillStyle fill,
			    int *init_step, int *extra_step);


  virtual int X_Erase (XDisplay *xd,
		       Window w,
		       GC gc,
		       int sw, int sh);

  virtual int X_Erase (XDisplay *xd,
		       Window w,
		       GC gc,
		       int x, int y,
		       int sw, int sh);

  virtual int X_Draw (XDisplay *xd,
		      Window w,
		      GC gc,
		      int x, int y,
		      int sw, int sh);

  virtual int X_Draw (XDisplay *xd,
		      Window w,
		      GC gc,
		      int sw, int sh);

  virtual void get_xy (int *x, int *y, int *h);
  virtual Pos xy_to_position (int x, int y);
  virtual void offset_to_xy (int offset, int *x, int *y, int *h);

  /*
  virtual int Archive (int fildes, XDisplay *xd, Assoc *assoc);
  virtual void UnArchive (int fildes, XDisplay *xd, Assoc *assoc);
  */

  virtual void offset_to_index_array (int offset, int *array, int *len);
  virtual Pos index_array_to_offset (int *array, int len);

  virtual SCM size_leaf_ascii (XDisplay *xd, int *sz);
  virtual SCM get_leaf_ascii (XDisplay *xd, char **ret, int *len);
  virtual SCM set_leaf_ascii (XDisplay *xd, char *contents, int len);
  virtual SCM size_leaf_xword (XDisplay *xd, int *sz);
  virtual SCM get_leaf_xword (XDisplay *xd, char **ret, int *len);
  virtual SCM set_leaf_xword (XDisplay *xd, char *chars, int len);
  virtual SCM size_leaf_ps (XDisplay *xd, int *sz);
  virtual SCM get_leaf_ps (XDisplay *xd, char **ret, int *len);
  virtual SCM set_leaf_ps (XDisplay *xd, char *chars, int len);
  virtual SCM size_leaf_html (XDisplay *xd, int *sz);
  virtual SCM get_leaf_html (XDisplay *xd, char **ret, int *len);
  virtual SCM set_leaf_html (XDisplay *xd, char *chars, int len);
};

extern Size ZeroSize;

void tabber (int n);
Size size_diff (Size a, Size b);
int size_compare (Size a, Size b);

/*
Container *Container_UnArchive (int fildes, XDisplay *xd, Assoc *assoc);
*/

#endif // CONTAINER_H_INCLUDED
