/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  


#include "XWord.h"
#include "Document.h"
#include "g.h"
#include "utils.h"



Document::Document (Attributes *attr) : Container (attr)
{
  last_known_x = 0;
  last_known_y = 0;
}


SCM document_new (SCM scm_attr)
{
  Attributes *attr = (Attributes *) gh_scm2ulong (scm_attr);
  Document *c = new Document (attr);
  return gh_ulong2scm ((unsigned long) c);
}


Container *
Document::new_of_my_own_type (void)
{
  Document *c = new Document (myAttributes);
  c->tmpAttributes = tmpAttributes;
  return c;
}



Document::~Document ()
{
  // do nothing...
}


void
Document::set_xy (int x, int y)
{
  last_known_x = x;
  last_known_y = y;
}


void
Document::get_xy (int *x, int *y, int *h)
{
  if (x) (*x) = last_known_x;
  if (y) (*y) = last_known_y;
  if (h) (*h) = _size.up + _size.down;
}


void
Document::offset_to_xy (int offset, int *x, int *y, int *h)
{
  xw_assert (offset <= num_ch);

  (*x) = last_known_x;
  (*y) = last_known_y;

  Container::offset_to_xy (offset, x, y, h);
}


char *Document::get_doc_file ()
{
  char *doc_file = (char *) getAttribute (at_doc_file);

  if (doc_file == NULL)
    {
      doc_file = "untitled.xword";

      fprintf (stderr, "no doc file -- using \"%s\"\n", doc_file);
      /* FIX ME */
      system ("mkdir untitled.xword");
      myAttributes->add (at_doc_file, doc_file);
    }

  return doc_file;
}


SCM document_get_doc_file (SCM scm_doc)
{
  Document *doc = (Document *) gh_scm2ulong (scm_doc);
  return gh_str02scm (doc->get_doc_file ());
}


void Document::set_doc_file (char *doc_file)
{
  char *old_doc_file = (char *) getAttribute (at_doc_file); 

  if (old_doc_file != NULL && strcmp (old_doc_file, doc_file) != 0)
    {
      fprintf (stderr, "FINISH ME -- need to clean up \"%s\"\n", old_doc_file);
    }

  myAttributes->add (at_doc_file, doc_file);
}


SCM document_set_doc_file (SCM scm_doc, SCM scm_doc_file)
{
  Document *doc = (Document *) gh_scm2ulong (scm_doc);
  char *doc_file = gh_scm2string (scm_doc_file);

  doc->set_doc_file (doc_file);

  return SCM_BOOL_T;
}


void hook_up_guile_Document ()
{
  gh_new_procedure1_0 ("document.new", (ghp) document_new);
  gh_new_procedure1_0 ("document.get-doc-file", (ghp) document_get_doc_file);
  gh_new_procedure2_0 ("document.set-doc-file", (ghp) document_set_doc_file);
}
