/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include <stdlib.h>
#include "g.h"
#include "XWord.h"
#include "FontManager.h"
#include "utils.h"


FontManager *theFontManager;

FontManager::FontManager(Display *set_display)
{
    display = set_display;
    number_of_fonts = 0;
    max_number_of_fonts = 30;
    fonts = (VkFont**)malloc(sizeof(VkFont*) * max_number_of_fonts);

    number_of_fontNames = 0;
    max_number_of_fontNames = 30;
    fontNames = (char**)malloc (sizeof(char*) * max_number_of_fontNames);

    number_of_FontConfigEntries = 0;
    max_number_of_FontConfigEntries = 20;
    entries = (FontConfigEntry *)malloc(sizeof(FontConfigEntry) *
					max_number_of_FontConfigEntries);

    scale = 1.0;
}


SCM fontmanager_new (SCM scm_set_display)
{
  Display *set_display = (Display *) gh_scm2ulong (scm_set_display);
  FontManager *fm = new FontManager (set_display);
  return gh_ulong2scm ((unsigned long) fm);
}


void
FontManager::loadFonts()
{
    FILE *inFile = NULL;

    char *filename = find_the_file ("font_config");
    if (filename != NULL)
      inFile = fopen (filename , "r");

    if (!inFile)
      {
	if (filename)
	  {
	    perror ("fopen");
	    fprintf (stderr, "Error: Could not open file '%s' for reading.\n",
		     filename);
	  }
	
	fprintf (stderr,
		 "FontManager could not load config file.\n"
		 "    The only font available is Courier.\n");
	
	/* we don't assume the default font is on the printer */
	addFontConfigEntry ("Courier",
		  "-adobe-courier-medium-r-*-*-%d-*-*-*-p-*-iso8859-1", 0);
	if (query("Courier:10") == NULL)
	  {
	    printf ("Could not even load default font.  Doomed!\n");
	    exit (2);
	  }

	return;
      }


    char psName[40], xlfd[150];
    int present_on_printer;
    
    while (!feof(inFile))
    {
	if (fscanf (inFile, "%s %d %s",
		    psName,
		    &present_on_printer,
		    xlfd) == 3)
	{
	    addFontConfigEntry(psName, xlfd, present_on_printer);
	}
    }

    if (query("Courier:10") == NULL)
    {
	
	printf("Could not load default font.  Check your font_config file\n");
	exit(1);
    }

    fclose(inFile);
}


SCM fontmanager_loadFonts (SCM scm_fm)
{
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  fm->loadFonts ();
  return SCM_BOOL_T;
}



void
FontManager::set_scale (double new_scale)
{
  int i;

  scale = new_scale;

  for (i=0; i<number_of_fonts; i++)
    {
      if (fonts[ i ] != NULL)
	fonts[ i ]->scale = scale;
    }
}


SCM fontmanager_set_scale (SCM scm_fm, SCM scm_new_scale)
{
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  fm->set_scale (gh_scm2double (scm_new_scale));

  return SCM_BOOL_T;
}


void
FontManager::addFontConfigEntry(char *psName, 
				char *xlfd, 
				int present_on_printer)
{
    entries[number_of_FontConfigEntries].name = strdup (psName);
    entries[number_of_FontConfigEntries].format = strdup(xlfd);
    entries[number_of_FontConfigEntries++].present = present_on_printer;

    if (number_of_FontConfigEntries == max_number_of_FontConfigEntries){
	max_number_of_FontConfigEntries *= 2;
	
	entries = (FontConfigEntry*)realloc(entries,
					    max_number_of_FontConfigEntries *
					    sizeof(FontConfigEntry));
    }

    addToFontNames(psName);
}


SCM fontmanager_addFontConfigEntry (SCM scm_fm,
				    SCM scm_psName,
				    SCM scm_xlfd,
				    SCM scm_present_on_printer)
{
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  char *psName = gh_scm2string (scm_psName);
  char *xlfd = gh_scm2string (scm_xlfd);
  int present_on_printer = gh_scm2bool (scm_present_on_printer);
  fm->addFontConfigEntry (psName, xlfd, present_on_printer);

  /* free psName? */
  /* free xlfd? */
  return SCM_BOOL_T;
}


VkFont *
FontManager::add (char */*name*/, VkFont *fs)  /* WHY IS NAME NOT USED HERE? */
{
    fonts[number_of_fonts] = fs;

    number_of_fonts++;

    if (number_of_fonts == max_number_of_fonts)
    {
	max_number_of_fonts += 10;
	fonts = (VkFont**)realloc (fonts, sizeof(VkFont*) *
				   max_number_of_fonts);
    }

    return fs;
}


SCM fontmanager_add (SCM scm_fm, SCM scm_name, SCM scm_fs)
{
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  char *name = gh_scm2string (scm_name);
  VkFont *fs = (VkFont *) gh_scm2ulong (scm_fs);

  fm->add (name, fs);

  /* free name? */
  return scm_fs;
}


VkFont *
FontManager::query (char *name)
{
    int i;
    char actual_name[100];
    char *x = strchr(name, (int)':');
    int y;

    strncpy(actual_name, name, (int)(x-name));
    actual_name[(int)(x-name)] = '\0';

    x+=1;

    y = atoi(x);

    for (i=0; i< number_of_fonts; i++)
	if (!strcmp(name, fonts[i]->displayName()) &&
	    y == (int)(fonts[i]->pointSize()))
	    return fonts[i];

    for (i=0; i< number_of_FontConfigEntries; i++)
    {
	if (!strcmp(actual_name, entries[i].name))
        {
	    char xfontname[400];
	    VkFont *newFont;

	    sprintf(xfontname, entries[i].format,y);

	    newFont = new VkFont(display, xfontname, actual_name, (float)y,
				 (VkFontTraitMask) 0, scale);


	    if (newFont->fs == NULL)
	      newFont->fs = fonts[ 0 ]->fs;  /* FIX ME */

	    return (add(name, newFont));
	}
    }
    return NULL;
}


SCM fontmanager_query (SCM scm_fm, SCM scm_name)
{
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  char *name = gh_scm2string (scm_name);
  VkFont *ret;

  ret = fm->query (name);

  if (ret == NULL)
    return SCM_BOOL_F;

  /* free name? */

  return gh_ulong2scm ((unsigned long) ret);
}


SCM fontmanager_defaultFont (SCM scm_fm)
{
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  VkFont *ret = fm->defaultFont ();

  if (ret == NULL)
    return SCM_BOOL_F;

  return gh_ulong2scm ((unsigned long) ret);
}


void
FontManager::addToFontNames(char *font_name)
{
    int i;
    for (i=0; i<number_of_fontNames; i++)
	if (!strcmp(fontNames[i], font_name))
	    return;

    fontNames[number_of_fontNames++] = strdup(font_name);
    
    if (number_of_fontNames == max_number_of_fontNames)
    {
	max_number_of_fontNames *= 2;
	
	fontNames = (char**)realloc(fontNames,
				    sizeof(char*) * max_number_of_fontNames);
    }
}


SCM fontmanager_addToFontNames (SCM scm_fm, SCM scm_font_name)
{
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  char *font_name = gh_scm2string (scm_font_name);

  fm->addToFontNames (font_name);

  /* free font_name? */

  return SCM_BOOL_T;
}



char **
FontManager::getFontNameList()
{
    return fontNames;
}

int
FontManager::getFontNameListNumber()
{
    return number_of_fontNames;
}


/* FINISH ME */
/* add scheme interface to getFontNameList and getFontNameListNumber */
/* (should only take 1 procedure.) */


void hook_up_guile_FontManager ()
{
  gh_new_procedure1_0 ("fontmanager.new", (ghp) fontmanager_new);
  gh_new_procedure1_0 ("fontmanager.loadFonts", (ghp) fontmanager_loadFonts);
  gh_new_procedure4_0 ("fontmanager.addFontConfigEntry",
		       (ghp) fontmanager_addFontConfigEntry);
  gh_new_procedure3_0 ("fontmanager.add", (ghp) fontmanager_add);
  gh_new_procedure2_0 ("fontmanager.query", (ghp) fontmanager_query);
  gh_new_procedure1_0 ("fontmanager.defaultFont",
		       (ghp) fontmanager_defaultFont);
  gh_new_procedure2_0 ("fontmanager.addToFontNames",
		       (ghp) fontmanager_addToFontNames);
  gh_new_procedure2_0 ("fontmanager.set-display-scale",
		       (ghp) fontmanager_set_scale);
}
