/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include "XWord.h"
#include "Page.h"
#include "utils.h"


/* move border_size into attributes or something....*/
static int border_size = 15;



Page::Page (Attributes *attr) : Container (attr)
{
}


SCM page_new (SCM scm_attr)
{
  Attributes *attr = (Attributes *) gh_scm2ulong (scm_attr);
  Page *c = new Page (attr);
  return gh_ulong2scm ((unsigned long) c);
}


Container *
Page::new_of_my_own_type (void)
{
  Page *c = new Page (myAttributes);
  c->tmpAttributes = tmpAttributes;
  return c;
}


Page::~Page ()
{
  // do nothing...
}


void
Page::handle_fill (Attr_Direction direction, Attr_FillStyle fill,
		   int *init_step, int *extra_step)
{
  Container::handle_fill (direction, fill, init_step, extra_step);
  // add margin stuff here
}


int
Page::X_Erase (XDisplay *xd,
	       Window w,
	       GC gc,
	       int x, int y,
	       int sw, int sh)
{
  int width, height;

  width = _size.left + _size.right + 2*border_size;
  height = _size.up + _size.down + 2*border_size;

  /* clip to the window size */
  if (x < 0) {width += x; x = 0;}
  if (y < 0) {height += y; y = 0;}
  if (width > sw) {width = sw;}
  if (height > sh) {height = sh;}

  XSetForeground (xd->display, gc,
		  (Pixel) getAttribute (at_background));
  
  XFillRectangle (xd->display, w, gc,
		  (int) (x - _size.left), (int) (y - _size.up),
		  width, height);

  return 1;
}


int
Page::X_Draw (XDisplay *xd,
	      Window w,
	      GC gc,
	      int x, int y,
	      int sw, int sh)
{
  if (Container::X_Draw (xd, w, gc,
			 x + border_size, y + border_size,
			 sw, sh))
    {
      Pixel fg = (Pixel) getTmpAttribute (at_foreground);

      XSetForeground (xd->display, xd->gc, fg);

      XSetFillStyle (xd->display, xd->gc, FillTiled);

      XFillRectangle (xd->display, w, xd->gc,
		      x,
		      y,
		      2*border_size + _size.right,
		      border_size);

      XFillRectangle (xd->display, w, xd->gc,
		      x,
		      y + border_size + _size.down,
		      2*border_size + _size.right,
		      border_size);

      XFillRectangle (xd->display, w, xd->gc,
		      x,
		      y + border_size,
		      border_size,
		      _size.down);

      XFillRectangle (xd->display, w, xd->gc,
		      x + border_size + _size.right,
		      y + border_size,
		      border_size,
		      _size.down);


      XSetFillStyle (xd->display, xd->gc, FillSolid);

      return 1;
    }

  return 0;
}



void
Page::get_xy (int *x, int *y, int *h)
{
  Container::get_xy (x, y, h);
  (*x) += border_size;
  (*y) += border_size;
}

void
Page::offset_to_xy (int offset, int *x, int *y, int *h)
{
  Container::offset_to_xy (offset, x, y, h);
  (*x) += border_size;
  (*y) += border_size;
}


Pos
Page::xy_to_position (int x, int y)
{
  x -= border_size;
  y -= border_size;
  return Container::xy_to_position (x, y);
}



/* lie to parent about size so the page can draw it's pretty borders */

Size Page::size (void)
{
  Size ret = _size;
  ret.right += (2*border_size);
  ret.down += (2*border_size);
  return ret;
}

unsigned int Page::size_left (void)
{
  return _size.left;
}

unsigned int Page::size_up (void)
{
  return _size.up;
}

unsigned int Page::size_right (void)
{
  return _size.right + 2*border_size;
}

unsigned int Page::size_down (void)
{
  return _size.down + 2*border_size;
}



void hook_up_guile_Page ()
{
  gh_new_procedure1_0 ("page.new", (ghp) page_new);
}
