/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include "XWord.h"
#include "Document.h"
#include "VkFont.h"
#include "Tag.h"
#include "utils.h"



Tag::Tag (Attributes *attr) : Container (attr)
{
  _size.left = 0;
  _size.up = 8;
  _size.right = 0;
  _size.down = 0;
}


SCM tag_new (SCM scm_attr)
{
  Attributes *attr = (Attributes *) gh_scm2ulong (scm_attr);
  Tag *c = new Tag (attr);
  return gh_ulong2scm ((unsigned long) c);
}


Tag::~Tag ()
{
}


Container *Tag::new_of_my_own_type (void)
{
  Tag *c = new Tag (myAttributes);
  c->tmpAttributes = tmpAttributes;
  return c;
}


Container *Tag::dup ()
{
  int i;
  Container *c = new_of_my_own_type ();

  c->ready_for_index (num_ch);

  for (i=0; i<num_ch; i++)
    c->c_tag[ i ] = c_tag[ i ]->dup ();

  c->num_ch = num_ch;
  c->_size = _size;
  return c;
}


Attr_Tag Tag::to_number ()
{
  /* might change this to use attributes... */

  if (num_ch == 0)
    return TAG_unknown;

  return c_tag[ 0 ]->to_number ();
}


int Tag::are_you_first (void)
{
  if (num_ch == 0)
    return 0;

  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_br:
    case TAG_img:
      return 0;
    case TAG_hr:
      return 3;
    default:
      xw_assert (0);
      return 0;
    }
}

  
int Tag::are_you_last (void)
{
  if (num_ch == 0)
    return 0;

  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_img:
      return 0;
    case TAG_br:
    case TAG_hr:
      return 3;
    default:
      xw_assert (0);
      return 0;
    }
}

  
void Tag::ready_for_index (int n)
{
  if (n >= max_ch)
    {
      max_ch = n + 2;
      c_tag = (tag_info **) XtRealloc ((char *) c_tag,
				       sizeof (tag_info *) * max_ch);
      if (c_tag == NULL)
	{
	  perror ("realloc");
	  fprintf (stderr,
		   "Tag couldn't resize children array. bomb.\n");
	  xw_assert (0);
	}
    }
}


Container *Tag::get_child (int index)
{
  xw_assert (index >= 0);
  xw_assert (index < num_ch);

  //return (Container *) c_tag[ index ];
  return *((Container **) (&(c_tag[ index ])));
}

  
void Tag::dd (int depth)
{
  int i;

  tabber (depth);
  printf ("%s-%p (%d,%d,%d,%d) par=%p, %d children ",
	  type (),
	  this,
	  (int)_size.left,
	  (int)_size.up,
	  (int)_size.right,
	  (int)_size.down,
	  _parent,
	  num_ch
	  );

  for (i=0; i<num_ch; i++)
    c_tag[ i ]->print (stdout);
  printf ("\n");
}


void Tag::get_img_width_height (XDisplay *xd, int *width, int *height)
{
  char *ws;
  char *hs;

  *width = 0;
  *height = 0;

  ensure_pix (xd); 

  if (num_ch == 0)
    return;

  ws = c_tag[ 0 ]->get_value ("width");
  hs = c_tag[ 0 ]->get_value ("height");

  if (ws != NULL && hs != NULL)
    {
      (*width) = atoi (ws);
      (*height) = atoi (hs);
    }
}
  


Size Tag::calculate_size (XDisplay *xd, int *constraint_failure,
			  Size *left_over)
{
  Size my_new_size;
  int fixed_left, fixed_up, fixed_right, fixed_down;

  my_new_size.left = 0;
  my_new_size.up = 8;
  my_new_size.right = 0;
  my_new_size.down = 0;

  if (constraint_failure)
    {
      if (num_ch > 1)
	(*constraint_failure) = 1;
      else
	(*constraint_failure) = 0;
    }

  if (left_over)
    {
      (*left_over).left = 0;
      (*left_over).up = 0;
      (*left_over).right = 0;
      (*left_over).down = 0;
    }


  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_br:
      break;
    case TAG_hr:
      break;
    case TAG_img:
      if (num_ch > 0)
	{
	  int width, height;
	  get_img_width_height (xd, &width, &height);
	  my_new_size.right = width;
	  my_new_size.up = height;
	}
      break;
    default:
      xw_assert (0);
      break;
    }


  if (myAttributes->get (at_fixed_left, (void **) &fixed_left))
    my_new_size.left = fixed_left;
  if (myAttributes->get (at_fixed_up, (void **) &fixed_up))
    my_new_size.up = fixed_up;
  if (myAttributes->get (at_fixed_right, (void **) &fixed_right))
    my_new_size.right = fixed_right;
  if (myAttributes->get (at_fixed_down, (void **) &fixed_down))
    my_new_size.down = fixed_down;

  return my_new_size;
}

  
void Tag::Insert_Children (int offset,
			   int num,
			   Container **new_children)
{
  int i;

  xw_assert (offset <= num_ch);
  xw_assert (offset >= 0);
  xw_assert (new_children != NULL);
  xw_assert (num >= 0);

  if (num == 0)
    return;

  ready_for_index (num_ch + num);

  for (i = num_ch + num - 1; i > offset; i--)
    c_tag[ i ] = c_tag[ i - num ];

  for (i = 0; i < num; i++)
    c_tag[ i+offset ] = ((tag_info **) new_children)[ i ];

  num_ch += num;
}


void Tag::Delete_Children (int offset, int num, Container *unhook)
{
  int i;

  xw_assert (offset + num <= num_ch);

  for (i = 0; i < num_ch - num - offset; i++)
    {
      if (unhook)
	unhook->c_tag[ unhook->num_ch++ ] = c_tag[ i+offset ];

      /* can't free here, else Wrap_Pull gets confused */
      /*
      else
	XtFree (c_tag[ i+offset ]);
	*/

      c_tag[ i+offset ] = c_tag[ i+offset+num ];
    }

  num_ch -= num;
  
  /* free some memory */
  max_ch = num_ch + 2;
  c_tag = (tag_info **) XtRealloc ((char *) c_tag,
				   sizeof (tag_info *) * max_ch);
}


int Tag::X_Draw (XDisplay *xd, Window w, GC gc,
		 int x, int y, int sw, int sh)
{
  switch (to_number ())
    {
    case TAG_unknown:
      return 0;
    case TAG_img:
      return draw_tag_img (xd, w, gc, x, y, sw, sh);
    case TAG_br:
      return draw_tag_br (xd, w, gc, x, y, sw, sh);
    case TAG_hr:
	return draw_tag_hr (xd, w, gc, x, y, sw, sh);
    default:
      xw_assert (0);
      return 0;
    }
}


int Tag::draw_tag_br (XDisplay *xd, Window w, GC gc,
		      int x, int y, int sw, int sh)
{
  if (x - (int)(_size.left) > sw ||
      y - (int)(_size.up) > sh ||
      x + (int)(_size.right) < 0 ||
      y + (int)(_size.down) < 0)
    return 0;  // out of the window bounds.


  XFillRectangle (xd->display, w, gc,
		  x - _size.left,
		  y - _size.up,
		  _size.left + _size.right,
		  _size.up + _size.down);

  return 1;
}


int Tag::draw_tag_hr (XDisplay *xd, Window w, GC gc,
		      int x, int y, int sw, int sh)
{
  Size sz = _parent->_parent->size ();
  int ppw = sz.left + sz.right;

  if (x - (int)(_size.left) > sw ||
      y - (int)(_size.up) > sh ||
      x + ppw < 0 ||
      y + (int)(_size.down) < 0)
    return 0;  // out of the window bounds.

  Pixel fg = (Pixel) getTmpAttribute (at_foreground);

  XSetForeground (xd->display, gc, fg);
  XDrawLine (xd->display, w, gc, x, y-4, x+ppw, y-4);

  return 1;
}


int Tag::draw_tag_img (XDisplay *xd, Window w, GC gc,
		       int x, int y, int sw, int sh)
{
  Pixmap pix;

  if (x - (int)(_size.left) > sw ||
      y - (int)(_size.up) > sh ||
      x + (int)(_size.right) < 0 ||
      y + (int)(_size.down) < 0)
    return 0;  // out of the window bounds.

  pix = ensure_pix (xd);

  if (pix > 0)
    {
      Pixel fg = (Pixel) getTmpAttribute (at_foreground);
      XSetForeground (xd->display, gc, fg);
      XCopyArea (xd->display, pix, w, gc, 0, 0,
		 _size.left + _size.right,
		 _size.down + _size.up,
		 x - _size.left,
		 y - _size.up);

      return 1;
    }

  return 0;
}


  
Pos Tag::xy_to_position (int /*x*/, int y)
{
  Pos pos;
  pos.con = this;
  /* assuming left to right here...? */
  if (y < (int)(_size.left + _size.right))
    pos.offset = 0;
  else
    pos.offset = num_ch;

  return pos;
}


void Tag::offset_to_xy (int offset, int *x, int *y, int *h)
{
  xw_assert (offset >= 0);

  if (_parent != NULL)
    _parent->offset_to_xy (_parent->childs_index (this), x, y, h);
  else
    {
      (*x) = 0;
      (*y) = 0;
    }

  /* assuming left to right here? */
  if (offset > 0)
    (*x) += (_size.left + _size.right);

  (*h) = _size.up;
}


Pos Tag::index_array_to_offset (int *array, int len)
{
  Pos pos;

  xw_assert (array != NULL);
  xw_assert (array[ 0 ] <= num_ch);
  xw_assert (len == 1);

  pos.con = this;
  pos.offset = array[ 0 ];

  return pos;
}


SCM Tag::size_leaf_ascii (XDisplay */*xd*/, int *sz)
{
  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_img:
      (*sz) = 0;
      break;
    case TAG_br:
      (*sz) = 1;
      break;
    case TAG_hr:
      (*sz) = 22;
      break;
    default:
      (*sz) = 0;
      xw_assert (0);
      break;
    }

  return gh_int2scm (*sz);
}


char *tag_scratch = NULL;
int tag_scratch_max = 0;


SCM Tag::get_leaf_ascii (XDisplay *xd, char **ret, int *len)
{
  int length;

  size_leaf_ascii (xd, &length);

  size_scratch (&tag_scratch, &tag_scratch_max, length);
  tag_scratch[ 0 ] = '\0';

  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_img:
      break;
    case TAG_br:
      strcat (tag_scratch, "\n");
      break;
    case TAG_hr:
      strcat (tag_scratch, "\n");
      strcat (tag_scratch, "--------------------");
      strcat (tag_scratch, "\n");
      break;
    default:
      xw_assert (0);
      break;
    }

  if (ret) (*ret) = tag_scratch;
  if (len) (*len) = strlen (tag_scratch);

  return gh_str2scm (tag_scratch, strlen (tag_scratch));
}


SCM Tag::set_leaf_ascii (XDisplay */*xd*/, char * /*contents*/, int /*len*/)
{
  xw_assert (0);
  return SCM_BOOL_T;
}



SCM Tag::size_leaf_xword (XDisplay */*xd*/, int *sz)
{
  int i;
  int num = 0;

  for (i=0; i<num_ch; i++)
    {
      num += c_tag[ i ]->size ();
    }

  num *= 4;

  if (sz != NULL)
    (*sz) = num;

  return gh_int2scm (num);
}


SCM Tag::get_leaf_xword (XDisplay *xd, char **ret, int *len)
{
  int i;
  char *p;
  int length;

  if (to_number () == TAG_img)
    ensure_ppm (xd);

  if (num_ch > 0)
    c_tag[ 0 ]->delete_pair ("pixmap"); /* don't save <img pixmap=0x...> */

  size_leaf_xword (xd, &length);
  size_scratch (&tag_scratch, &tag_scratch_max, length+1);

  for (i=0; i<length; i++)
    tag_scratch[ i ] = ' ';
  tag_scratch[ length ] = '\0';

  p = tag_scratch;
  
  for (i=0; i<num_ch; i++)
    {
      c_tag[ i ]->sprint (p);
      p += strlen (p);
    }

  for (i=strlen (tag_scratch); i>0; i--)
    {
      tag_scratch[ i*4 ] = tag_scratch[ i ];
      tag_scratch[ i ] = ' ';
    }

  if (ret) (*ret) = tag_scratch;
  if (len) (*len) = length;

  return gh_str2scm (tag_scratch, length);
}


SCM Tag::set_leaf_xword (XDisplay */*xd*/, char *chars, int len)
{
  char *p;
  int i;

  size_scratch (&tag_scratch, &tag_scratch_max, len);

  /* collapse the 4 bytes per character */

  for (i=0; i<len; i++)
    tag_scratch[ i ] = chars[ i*4 ];

  p = tag_scratch;
  len /= 4;

  /* FINISH ME -- free old children here... */
  num_ch = 0;

  while (p < tag_scratch + len)
    {
      xw_assert (*p == '<');

      for (i=0; ;i++)
	{
	  if ((p + i + 1) >= (tag_scratch + len) ||
	      p[ i ] == '>')
	    {
	      ready_for_index (num_ch);

	      c_tag[ num_ch ] = new tag_info (p, i+1);
	      p += i;
	      p ++;
	      num_ch ++;
	      break;
	    }
	}
    }

  return SCM_BOOL_T;
}



SCM Tag::size_leaf_ps (XDisplay *xd, int *sz)
{
  int num = 0;

  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_br:
      num = strlen (" skip:");
      break;
    case TAG_img:
      num = strlen ("shell:pnmtops -rle -nocenter -noturn -scale 0.72 ");
      /*
      ensure_ppm (xd);
      num += strlen (c_tag[ 0 ]->get_value ("ppm"));
      */
      num += strlen (ppm_file_name (xd));
      num += 5;
      break;
    case TAG_hr:
      num = strlen ("   ps: 0 rlineto closepath stroke\n");
      num += strlen (itoa (100));
      break;
    default:
      xw_assert (0);
      num = strlen (" skip:");
      return 0;
    }


  if (sz != NULL)
    (*sz) = num;

  return gh_int2scm (num);
}


Pixmap Tag::ensure_pix (XDisplay *xd)
{
  char *ppm;
  char *pn;
  int width, height;

  if (num_ch == 0)
    return 0;

  Pixmap pix = c_tag[ 0 ]->get_pixmap ();

  if (pix != 0)
    return pix; /* already have it. */

  pn = c_tag[ 0 ]->get_value ("pixmap");
  if (pn != NULL)
    {
      char *widths, *heights;
      pix = atol (pn);
      widths = c_tag[ 0 ]->get_value ("width");
      heights = c_tag[ 0 ]->get_value ("height");
      if (widths == NULL || heights == NULL)
	pix = 0; /* sigh. */
      else
	{
	  width = atoi (widths);
	  height = atoi (heights);
	}
    }

  if (pix == 0 && ensure_ppm (xd) == 1)
    {
      /*ppm = c_tag[ 0 ]->get_value ("ppm");*/
      ppm = ppm_file_name (xd);
      pix = xd->read_ppm_file_to_pixmap (ppm, &width, &height);
    }

  if (pix == 0)
    {
      width = 15;
      height = 15;
      /* not a ppm file? */ /* make a little question mark */
      pix = XCreatePixmap (xd->display,
			   RootWindow (xd->display, xd->screen),
			   width, height,
			   xd->depth);
      XSetForeground (xd->display, xd->gc,
		      (Pixel) getAttribute (at_background));
      XFillRectangle (xd->display, pix, xd->gc, 0, 0, width, height);
      XSetForeground (xd->display, xd->gc,
		      (Pixel) getAttribute (at_foreground));
      XDrawRectangle (xd->display, pix, xd->gc, 0, 0, width, height);
      XDrawString (xd->display, pix, xd->gc, 2, 14, "?", 1);
    }

  c_tag[ 0 ]->set_pair ("width", itoa (width));
  c_tag[ 0 ]->set_pair ("height", itoa (height));
  c_tag[ 0 ]->set_pixmap (pix);
  _size.right = width;
  _size.up = height;

  return pix;
}


int Tag::ensure_ppm (XDisplay *xd)
{
  char *ppm = c_tag[ 0 ]->get_value ("ppm");
  Pixmap pix;
  char *src;

  if (ppm != NULL)
    return 1; /* already have it. */

  if (num_ch == 0)
    return 0; /* can't get it. */

  pix = c_tag[ 0 ]->get_pixmap ();

  if (pix != 0)
    {
      char fn[ 2048 ];
      int add = 0;
      Document *doc = (Document *) top_of_tree ();
      char *doc_file = doc->get_doc_file ();
      char *fl; /* fn without the path */

      do
	{
	  /* FINISH ME -- check for > 2048 */
	  sprintf (fn, "%s/xword-%ld.ppm",
		   doc_file,
		   (unsigned long int) pix + add);
	  add ++;
	} while (file_exists (fn));

      if ((fl = strrchr (fn, '/')) == NULL) fl = fn; else fl ++;

      xd->write_ppm_file_from_pixmap (fn, _size.right, _size.up, pix);
      c_tag[ 0 ]->set_pair ("ppm", fl);

      return 1;
    }

  src = c_tag[ 0 ]->get_value ("src");

  if (src != NULL)
    {
      char *com;
      char fn[ 2048 ];
      int add = 0;
      Document *doc = (Document *) top_of_tree ();
      char *doc_file = doc->get_doc_file ();
      char *fl; /* fn without the path */

      do
	{
	  /* FINISH ME -- check for > 2048 */
	  sprintf (fn, "%s/xword-%ld.ppm",
		   doc_file, (unsigned long int) pix + add);
	  add ++;
	} while (file_exists (fn));

      if ((fl = strrchr (fn, '/')) == NULL) fl = fn; else fl ++;

      com = (char *) XtMalloc (strlen (src) + strlen (fn) + 20);
      sprintf (com, "anytopnm %s > %s", src, fn);
      system (com);

      if (!file_readable (fn))
	return 0;

      c_tag[ 0 ]->set_pair ("ppm", fl);

      return 1;
    }

  return 0; /* give up. */
}


SCM Tag::get_leaf_ps (XDisplay *xd, char **ret, int *len)
{
  int length;

  if (to_number () == TAG_img)
    ensure_ppm (xd);

  size_leaf_ps (xd, &length);
  size_scratch (&tag_scratch, &tag_scratch_max, length+1);


  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_br:
      sprintf (tag_scratch, " skip:");
      break;
    case TAG_img:
      /*ensure_ppm (xd);*/
      sprintf (tag_scratch,
	       "shell:pnmtops -rle -nocenter -noturn -scale 0.66 %s",
	       ppm_file_name (xd));
	       /*c_tag[ 0 ]->get_value ("ppm"));*/
      break;
    case TAG_hr:
      {
	Size sz = _parent->_parent->size ();
	int ppw = sz.left + sz.right;
	sprintf (tag_scratch, "   ps:%d 0 rlineto closepath stroke\n",
		 (int)(ppw * 0.66));
      }
      break;
    default:
      xw_assert (0);
      sprintf (tag_scratch, " skip:");
      return 0;
    }

  length = strlen (tag_scratch);

  if (ret) (*ret) = tag_scratch;
  if (len) (*len) = length;

  return gh_str2scm (tag_scratch, length);
}


SCM Tag::set_leaf_ps (XDisplay */*xd*/, char */*chars*/, int /*len*/)
{
  xw_assert (0);
  return SCM_BOOL_T;
}



SCM Tag::size_leaf_html (XDisplay */*xd*/, int *sz)
{
  int num = 0;

  if (num_ch > 0)
    num = c_tag[ 0 ]->size ();

  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_img:
      break;
    case TAG_hr:
      num ++;
      /* fall through */
    case TAG_br:
      num ++;
      break;
    default:
      xw_assert (0);
      break;
    }


  if (sz != NULL)
    (*sz) = num;

  return gh_int2scm (num);
}


SCM Tag::get_leaf_html (XDisplay *xd, char **ret, int *len)
{
  int length;

  if (to_number () == TAG_img)
    ensure_ppm (xd);

  if (num_ch > 0)
    c_tag[ 0 ]->delete_pair ("pixmap"); /* don't save <img pixmap=0x...> */

  size_leaf_html (xd, &length);

  size_scratch (&tag_scratch, &tag_scratch_max, length);
  tag_scratch[ 0 ] = '\0';

  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_img:
    case TAG_br:
      break;
    case TAG_hr:
      strcat (tag_scratch, "\n");
      break;
    default:
      xw_assert (0);
      break;
    }


  if (num_ch > 0)
    c_tag[ 0 ]->sprint (tag_scratch + strlen (tag_scratch));


  switch (to_number ())
    {
    case TAG_unknown:
    case TAG_img:
      break;
    case TAG_hr:
    case TAG_br:
      strcat (tag_scratch, "\n");
      break;
    default:
      xw_assert (0);
      break;
    }


  if (ret) (*ret) = tag_scratch;
  if (len) (*len) = strlen (tag_scratch);

  return gh_str2scm (tag_scratch, strlen (tag_scratch));
}


SCM Tag::set_leaf_html (XDisplay */*xd*/, char *chars, int len)
{
  if (num_ch == 0)
    {
      num_ch = 1;
      ready_for_index (num_ch);
      c_tag[ 0 ] = new tag_info ();
    }

  c_tag[ 0 ]->parse (chars, len);

  return SCM_BOOL_T;
}



char *Tag::ppm_file_name (XDisplay *xd)
{
  static char *fn = NULL;
  static unsigned int fn_len = 0;
  Document *doc;
  char *doc_file;
  char *ppm;

  ensure_ppm (xd);

  doc = (Document *) top_of_tree ();
  doc_file = doc->get_doc_file ();
  ppm = c_tag[ 0 ]->get_value ("ppm");

  if (fn_len < strlen (doc_file) + strlen (ppm))
    {
      fn_len = strlen (doc_file) + strlen (ppm) + 5;
      fn = XtRealloc (fn, fn_len);
    }

  strcpy (fn, doc_file);
  strcat (fn, "/");
  strcat (fn, ppm);

  return fn;
}



void hook_up_guile_Tag ()
{
  gh_new_procedure1_0 ("tag.new", (ghp) tag_new);
}
