/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

/**
 *
 * VkFont.C
 *
 * This file contains the class implementation
 * VkFont, which encapsulates screen (as well
 * as printer) fonts, and provides an interface
 * to manipulating them.  The majority of operations,
 * however, should be performed with the VkFontManager
 * instead of individual fonts, so the application
 * maintains information of these changes.
 *
 * This class is roughly analogous to NSFont in
 * the OpenStep Specification, which inspired it.
 *
 **/

#include "VkFont.h"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "utils.h"
#include "g.h"


VkFont::VkFont (Display *display,
		char *xlfd,
		char *fontName,
                float pointSize,
                VkFontTraitMask,
		double new_scale)
{
    _pointSize = pointSize;

    _screenName = strdup(xlfd);

/*
    if ((fp = fopen(VkFormat("/usr/public/lib/ViewKit/AFM/%s", fontName), "r")) == NULL)
*/
	_hasAFM = False;
    /*else 
    {
	_hasAFM = parseAFM(fp);
	fclose(fp);
    }*/

    if (!_hasAFM)
        _fontName = strdup(fontName);

    //printf ("VkFont: screen name: '%s'\n", _screenName);

    set_scale (new_scale);

    //fs = XLoadQueryFont(display, _screenName);
    fs = XLoadQueryFont(display, _scaleName);

    if (fs == NULL)
    {
      fprintf (stderr, "font '%s' didn't load.\n", _screenName);
    }
}


void
VkFont::set_scale (double new_scale)
{
  scale = new_scale;

  int hy_count;
  int sz_start;
  int sz_end;
  int i;

  char old_sz_s[ 5 ];
  int old_sz_i;

  for (i=0, hy_count=0; hy_count < 7; i++)
      if (_screenName[ i ] == '-')
	hy_count ++;

  sz_start = i;

  for (;;i++)
    if (_screenName[ i ] == '-')
      break;

  sz_end = i;

  //printf (":::: %s --> ", _screenName);

  //for (i=sz_start; i<sz_end; i++)
    //printf ("%c", _screenName[ i ]);

  //printf ("\n");

  /* free old _scaleName */

  _scaleName = (char *) XtMalloc (strlen (_screenName) +
				  (sz_end - sz_start) + 5);


  strncpy (old_sz_s, _screenName + sz_start, sz_end - sz_start);
  old_sz_s[ sz_end - sz_start ] = '\0';

  old_sz_i = atoi (old_sz_s);

  strncpy (_scaleName, _screenName, sz_start);

  sprintf (_scaleName + sz_start, "%d", (int) (old_sz_i * scale));
  strcat (_scaleName, _screenName + sz_end);

  //printf ("::: '%s'\n", _scaleName);
  

  /* free old font here */
  /* XUnloadFont */
  //fs = here;
}



Boolean
VkFont::parseAFM(FILE *fp)
{
    char buf[1000];
    char first_field[1000];

    fgets(buf, 1000, fp);

    if (strncmp(buf, "StartFontMetrics", 16))
	return False;

    while (!feof(fp))
    {
	fscanf(fp, "%s ", first_field);

	if (!strcmp(first_field, "FamilyName"))
	{
	    fscanf(fp, "%s ", first_field);
	    _familyName = strdup(first_field);
	    continue;
	}
	else if (!strcmp(first_field, "FontName"))
	{
	    fscanf(fp, "%s ", first_field);
	    _fontName = strdup(first_field);
	    continue;
	}
	else 
	{
	    
	    fgets(buf, 1000, fp);
	    continue;
	}
    }
    return True;
}

Boolean 
VkFont::hasAFMDictionary()
{
    return _hasAFM;
}

char *
VkFont::displayName()
{
    return _fontName;
}

SCM vk_font_display_name (SCM scm_vkf)
{
  VkFont *vkf = (VkFont *) gh_scm2ulong (scm_vkf);
  return gh_str02scm (vkf->displayName ());
}

const char *
VkFont::familyName()
{
  return NULL;
}

float 
VkFont::pointSize()
{
    return _pointSize;
}


SCM vk_font_point_size (SCM scm_vkf)
{
  VkFont *vkf = (VkFont *) gh_scm2ulong (scm_vkf);
  return gh_int2scm ((int) vkf->pointSize ());
}


VkFont *
VkFont::printerFont()
{
    return this;
}

XFontStruct *
VkFont::screenFont()
{
    return fs;
}

float 
VkFont::stringWidth(char *string)
{
    if (_hasAFM)
	return 0.0;/* do some nifty font metric things */
    else
	return (float)XTextWidth(fs, string, strlen(string));
}

float
VkFont::stringWidth(char *string, int num_chars)
{
    if (_hasAFM)
	return 0.0;/* do some nifty font metric things */
    else
	return (float)XTextWidth(fs, string, num_chars);
}

float
VkFont::stringWidth16 (XChar2b *string, int num_chars)
{
    if (_hasAFM)
	return 0.0;/* do some nifty font metric things */
    else
	return (float)XTextWidth16(fs, (XChar2b *) string, num_chars);
}

const float *
VkFont::widths()
{
  return NULL;
}

float 
VkFont::xAdvancementForGlyph(char aGlyph, 
			     char)
{
    if (_hasAFM)
	return 0.0;/* do some nifty thing with font metrics */
    else
	return XTextWidth(fs, &aGlyph, 1);
}

float 
VkFont::yAdvancementForGlyph(char)
{
    if (_hasAFM)
	return 0.0;/* do some nifty thing with font metrics */
    else
	return fs->ascent + fs->descent;
}

float 
VkFont::heightOfBoundingRect(char c)
{
    if (isspace (c))
	return 0;

    if (_hasAFM)
	return 0.0;/* do some nifty thing with font metrics */
    else
	return fs->ascent + fs->descent;
}

float 
VkFont::widthOfBoundingRect(char aGlyph)
{
    if (_hasAFM)
	return 0.0;/* do some nifty thing with font metrics */
    else
	return XTextWidth(fs, &aGlyph, 1);
}

float
VkFont::ascent()
{
    if (_hasAFM)
	return 0.0;/* do some nifty thing with the font metrics */
    else
	return fs->ascent;
}

float
VkFont::descent()
{
    if (_hasAFM)
	return 0.0;/* do some nifty thing with the font metrics */
    else
	return fs->descent;
}


void hook_up_guile_VkFont ()
{
  gh_new_procedure1_0 ("vk-font.display-name",
		       (ghp) vk_font_display_name);
  gh_new_procedure1_0 ("vk-font.point-size",
		       (ghp) vk_font_point_size);
}
