/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
  

/**
 *
 * $Id: VkFont.h,v 1.7 1997/05/12 15:51:01 alves Exp $
 *
 * This file contains the class description of 
 * VkFont, which encapsulates screen (as well
 * as printer) fonts, and provides an interface
 * to manipulating them.  The majority of operations,
 * however, should be performed with the VkFontManager
 * instead of individual fonts, so the application
 * maintains information of these changes.
 *
 *
 **/

#ifndef VKFONT_H
#define VKFONT_H

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <stdio.h>

typedef unsigned int VkFontTraitMask;

/* for now we only support the following traits */
/* should change as soon as more of the font machinery
   is in place  -- Chris */

#define VK_ITALIC_FONT_MASK 1
#define VK_BOLD_FONT_MASK 2

void hook_up_guile_VkFont ();

class VkFont {
private:
    char *_fontName;
    float _pointSize;
    char *_familyName;
    char *_screenName;
    char *_scaleName;

    Boolean _hasAFM;

    Boolean parseAFM(FILE *fp);

public:

    XFontStruct *fs;

    double scale;

    VkFont (Display *display,
	    char *xlfd,
	    char *fontName,
	    float pointSize,
	    VkFontTraitMask traitMask,
	    double set_scale);

	    //char *fontName = "Courier", 
	    //float pointSize = 11.0,
	    //VkFontTraitMask traitMask = 0,
	    //double set_scale = 1.0);


    void set_scale (double new_scale);


    Boolean hasAFMDictionary();
    char *displayName();
    const char *familyName();
    
    float pointSize();
    
    VkFont *printerFont();
    XFontStruct *screenFont();
    
    float stringWidth(char *string);
    float stringWidth(char *string, int num_chars);
    float stringWidth16 (XChar2b *string, int num_chars);


    const float *widths();
    
    float xAdvancementForGlyph(char aGlyph, char fromAGlyph = '\0');
    float yAdvancementForGlyph(char aGlyph);
    
    float heightOfBoundingRect(char aGlyph = (char)0);
    float widthOfBoundingRect(char aGlyph = (char)0);

    float ascent();
    float descent();
};

#endif // VKFONT_H
