/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include "XWord.h"
#include "VkFont.h"
#include "WordSegment.h"
#include "utils.h"


WordSegment::WordSegment (Attributes *attr) : Container (attr)
{
  /*byte_width = sizeof (short);*/
}


SCM wordsegment_new (SCM scm_attr)
{
  Attributes *attr = (Attributes *) gh_scm2ulong (scm_attr);
  WordSegment *c = new WordSegment (attr);
  return gh_ulong2scm ((unsigned long) c);
}


Container *
WordSegment::new_of_my_own_type (void)
{
  WordSegment *c = new WordSegment (myAttributes);
  c->tmpAttributes = tmpAttributes;
  return c;
}



Container *
WordSegment::dup ()
{
  int i;
  Container *c = new_of_my_own_type ();

  c->ready_for_index (num_ch);

  for (i=0; i<num_ch; i++)
    c->c_char[ i ] = c_char[ i ];

  c->num_ch = num_ch;

  return c;
}



WordSegment::~WordSegment ()
{
  // do nothing...
}



void WordSegment::ready_for_index (int n)
{
  if (n >= max_ch)
    {
      max_ch = n + 2;
      c_char = (XChar2b *) XtRealloc ((char *) c_char,
				      sizeof (XChar2b) * max_ch);
      if (c_char == NULL)
	{
	  perror ("realloc");
	  fprintf (stderr,
		   "WordSegment couldn't resize children array. bomb.\n");
	  xw_assert (0);
	}
    }
}


Container *
WordSegment::get_child (int index)
{
  xw_assert (index < num_ch);
  /* I don't care if you believe me, this _is_ needed: */
  return *((Container **) (&(c_char[ index ])));
  /* I love C */
};


void WordSegment::dd (int depth) /* debug_dump */
{
  int i;

  tabber (depth);
  printf ("%s-%p (%d,%d,%d,%d) par=%p, %d children '",
	  type (),
	  this,
	  (int)_size.left,
	  (int)_size.up,
	  (int)_size.right,
	  (int)_size.down,
	  _parent,
	  num_ch
	  );
  
  for (i=0; i<num_ch; i++)
    {
      switch (xc2b_to_xk (c_char[ i ]))
	{
	case XK_Linefeed: printf ("\\n"); break;
	case XK_Return: printf ("\\r"); break;
	case XK_Tab: printf ("\\t"); break;
	case XK_backslash: printf ("\\"); break;
        default:
	  printf ("%c", (int)(xc2b_to_xk (c_char[ i ]) & 0xFF));
	  break;
	}
    }

  printf ("'\n");
}


/********************************************************************/



XChar2b *flip_string_scratch_space = NULL;
int flip_string_max_size = 0;

XChar2b *
flip_string (XChar2b *str, int len)
{
  int i;
  
  int number_flag = 0;
  int space_flag = 0;


  if (len == 0)
    return flip_string_scratch_space;

  if (len >= flip_string_max_size)
    {
      flip_string_max_size = len + 10;
      flip_string_scratch_space =
	(XChar2b *) XtRealloc ((char *) flip_string_scratch_space,
			       flip_string_max_size *
			       sizeof (XChar2b));
    }

  /* numbers still go from left to right. */

  for (i=0; i < len; i++)
    if (xc2b_to_xk (str[i]) >= XK_0 && xc2b_to_xk (str[i]) <= XK_9)
      {
	number_flag = 1;
	break;
      }

  if (xc2b_to_xk (str[ len - 1 ]) == XK_space ||
      xc2b_to_xk (str[ len - 1 ]) == XK_Linefeed ||
      xc2b_to_xk (str[ len - 1 ]) == XK_Return ||
      xc2b_to_xk (str[ len - 1 ]) == 12)
    {
      space_flag = 1;
    }


  switch (number_flag * 2 + space_flag)
    {
    case 0:
    case 1:
      for (i=len-1; i>=0; i--)
	flip_string_scratch_space [ len-i-1 ] = str[ i ];
      break;
    case 2:
      return str;
    case 3:
      flip_string_scratch_space [ 0 ] = str[ len-1 ];
      for (i=0; i<len-1; i++)
	flip_string_scratch_space [ i+1 ] = str[ i ];
      break;
    default:
      xw_assert (0);
      break;
    }

  return flip_string_scratch_space;
}



Size
WordSegment::calculate_size (XDisplay */*xd*/,
			     int *constraint_failure, Size *left_over)
{
    Size my_new_size;
    int num_to_count;
    Attr_Direction direction;

    direction = (Attr_Direction) getAttribute (at_direction);

    (*constraint_failure) = 0;

    VkFont *f;
    if ((VkFont *)myAttributes->get(at_font, (void **) &f) == 0)
    {
	fprintf (stderr, "WordSegment::WordSegment -- "
		 "attributes don't container a font.\n");
	xw_assert (0);
	return _size;
    }

    xw_assert (f != NULL);

    for (num_to_count=0; num_to_count < num_ch-1; num_to_count++)
    {
      KeySym xk = xc2b_to_xk (c_char[ num_to_count ]);

      if (xk == XK_Linefeed ||
	  xk == XK_Return ||
	  xk == XK_space ||
	  xk == 12) /* ctrl-L... new page */
	{
	  (*constraint_failure) = 1;
	  break;
	}
    }

    if (num_ch > 0)
      num_to_count ++;

    my_new_size.up = (short) f->ascent ();
    my_new_size.down = (short) f->descent ();

    switch (direction)
      {
      case DIR_rl:
	my_new_size.left = (short) f->stringWidth16(c_char + 0, num_to_count);
	my_new_size.right = 0;
	break;
      default:
	my_new_size.left = 0;
	my_new_size.right = (short) f->stringWidth16(c_char + 0, num_to_count);
	break;
      }

    xw_assert (my_new_size.up+my_new_size.down == f->heightOfBoundingRect());

    if (left_over != NULL)
	(*left_over) = ZeroSize;

    return my_new_size;
}


int
WordSegment::are_you_first (void)
{
  return 0;
}


int
WordSegment::are_you_last (void)
{
  if (num_ch == 0)
    return 0;

  KeySym xk = xc2b_to_xk (c_char[ num_ch-1 ]);

  if (xk == XK_space)
    return 2;		    
			    
  if (xk == XK_Linefeed ||
      xk == XK_Return)
    return 3;		    
			    
  if (xk == 12)
    return 4;

  return 0;
}



/***************************************************************************/
/***************************************************************************/


void
WordSegment::Insert_Children (int offset, int num, Container **new_children)
{
    int i;

    xw_assert (offset <= num_ch);
    xw_assert (offset >= 0);
    xw_assert (new_children != NULL);
    xw_assert (num >= 0);

    if (num == 0)
      return;

    ready_for_index (num_ch + num);

    for (i = num_ch + num - 1; i > offset; i--)
	c_char[ i ] = c_char[ i - num ];

    for (i = 0; i < num; i++)
      c_char[ i+offset ] = ((XChar2b *) new_children)[ i ];

    num_ch += num;
}


void
WordSegment::Delete_Children (int offset, int num, Container *unhook)
{
    int i;

    xw_assert (offset + num <= num_ch);

    for (i = 0; i < num_ch - num - offset; i++)
    {
	if (unhook)
	    unhook->c_char[ unhook->num_ch++ ] = c_char[ i+offset ];

	c_char[ i+offset ] = c_char[ i+offset+num ];
    }

    num_ch -= num;

    /* free some memory */
    max_ch = num_ch + 2;
    c_char = (XChar2b *) XtRealloc ((char *) c_char,
				    sizeof (XChar2b) * max_ch);
}



int
WordSegment::X_Draw (XDisplay *xd,
		     Window w,
		     GC gc,
		     int x, int y,
		     int sw, int sh)
{
  Attr_Direction direction;

  direction = (Attr_Direction) getAttribute (at_direction);

  if (x - (int)(_size.left) > sw ||
      y - (int)(_size.up) > sh ||
      x + (int)(_size.right) < 0 ||
      y + (int)(_size.down) < 0)
    return 0;  // out of the window bounds.

    if (num_ch && gc)
    {
	VkFont *f = (VkFont *) getAttribute (at_font);

	if (f == NULL)
	{
	    fprintf (stderr, "WordSegment::WordSegment --"
		     " attributes don't contain a font.\n");
	    return 0;
	}


	/* handle strange background colors... */
	{
	  Pixel bp = (Pixel) getTmpAttribute (at_background);
	  int hls = (int) getTmpAttribute (at_hl_start);
	  int hle = (int) getTmpAttribute (at_hl_end);

	  if (bp != BlackPixel (xd->display, 0))
	    {
	      int width;
	      int xoff = 0;

	      if ((hls == 0) && (hle == 0))
		width = _size.left + _size.right;
	      else
		{
		  xoff = (int) f->stringWidth16 (c_char, num_ch - hls);
		  width = (int) (f->stringWidth16 (c_char, num_ch - hle))
		    - xoff;
		}
	      int height = _size.up + _size.down;

	      XSetForeground (xd->display, gc, bp);
  
	      XFillRectangle (xd->display, w, gc,
			      (int) (x - _size.left) + xoff,
			      (int) (y - _size.up),
			      width, height);
	    }
	}


	XSetForeground (xd->display, gc,
			(Pixel) getTmpAttribute (at_foreground));
                        //WhitePixel (xd->display, 0));


	XSetFont (xd->display, gc, f->screenFont()->fid);

	switch (direction)
	  {
	  case DIR_rl:
	    XDrawString16 (xd->display,
			   w,
			   gc,
			   (int) (x - _size.left), (int) y,
			   (XChar2b *) flip_string (c_char, num_ch),
			   num_ch);
	    break;
	  default:

	    //for (int i=0; i<num_ch; i++)
	    //{
	    //char *p = (char *) (c_char + i);
	    //char t = p[ 0 ];
	    //p[ 0 ] = p[ 1 ];
	    //p[ 1 ] = t;
	    //}

	    XDrawString16 (xd->display,
			   w,
			   gc,
			   (int)x, (int)y,
			   (XChar2b *) c_char,
			   num_ch);

	    break;
	  }

    }

  return 1;
}



Pos
WordSegment::xy_to_position (int x, int)
{
  int i;
  Pos pos;
  Attr_Direction direction;

  direction = (Attr_Direction) getAttribute (at_direction);

  pos.con = this;
  pos.offset = 0;

  VkFont *f = NULL;
  if ((VkFont *)myAttributes->get(at_font, (void **) &f) == 0)
    {
      xw_assert (0);
    }

  //if (f == NULL)
    //return pos;


  for (i=0; i<num_ch; i++)
    {
/*
      printf ("i=%d/%d, x=%f, sw=%f --> %d && %d\n",
	      i,
	      num_ch,
	      x,
	      f->stringWidth16(c_char, i),
	      direction == DIR_rl,
	      - f->stringWidth16(c_char, i) < x);
*/

      if (direction == DIR_rl)
	{
	  if (- f->stringWidth16 (c_char, i) < x)
	    break;
	}
      else if (f->stringWidth16 (c_char, i) > x)
	break;
    }


  pos.offset = i-1;

  switch (direction)
    {
    case DIR_rl:
      if ( - f->stringWidth16 (c_char, num_ch) > x)
	pos.offset = 0;
      break;
    default:
      if (f->stringWidth16 (c_char, num_ch) < x)
	pos.offset = num_ch;
      break;
    }


  if (pos.offset > 0)
    {
      KeySym xk = xc2b_to_xk (c_char[ pos.offset-1 ]);
      if (xk == XK_Linefeed ||
	  xk == XK_Return ||
	  xk == 12)
	pos.offset --;
    }

  if (pos.offset < 0)
    pos.offset = 0;

  return pos;
}


void
WordSegment::offset_to_xy (int offset, int *x, int *y, int *h)
{
  VkFont *f;
 Attr_Direction direction;

  direction = (Attr_Direction) getAttribute (at_direction);

  xw_assert (offset <= num_ch);

  if (_parent != NULL)
    _parent->offset_to_xy (_parent->childs_index (this), x, y, h);
  else
    {
      (*x) = 0;
      (*y) = 0;
    }

  if (offset == num_ch)
    {
      if (direction == DIR_rl)
	(*x) -= _size.left;
      else
	(*x) += _size.right;
    }
  else
    {
      f = (VkFont *) getAttribute (at_font);
      xw_assert (f != NULL);

      if (direction == DIR_rl)
	(*x) -= (short) f->stringWidth16 (c_char + 0, offset);
      else
	(*x) += (short) f->stringWidth16 (c_char + 0, offset);
    }

  (*h) = _size.up;
}


Pos
WordSegment::index_array_to_offset (int *array, int len)
{
  Pos pos;

  xw_assert (array != NULL);
  xw_assert (array[ 0 ] <= num_ch);
  xw_assert (len == 1);

  pos.con = this;
  pos.offset = array[ 0 ];

  return pos;
}


SCM
WordSegment::size_leaf_ascii (XDisplay */*xd*/, int *sz)
{
  int num = num_ch;

  if (sz != NULL)
    (*sz) = num;

  return gh_int2scm (num);
}


SCM
WordSegment::get_leaf_ascii (XDisplay */*xd*/, char **ret, int *len)
{
  int i;
  /* convert the array of shorts into a string */
  /* ... we loose any weird chars this way. */
  unsigned char *scratch;
  unsigned long int xk;
  Attr_Direction direction;

  if (num_ch >= flip_string_max_size)
    {
      flip_string_max_size = num_ch + 10;
      flip_string_scratch_space =
	(XChar2b *) XtRealloc ((char *) flip_string_scratch_space,
			       flip_string_max_size *
			       sizeof (XChar2b));
    }

  scratch = (unsigned char *) flip_string_scratch_space;

  direction = (Attr_Direction) getAttribute (at_direction);

  for (i=0; i<num_ch; i++)
    {
      XChar2b c;
      if (direction == DIR_rl)
	c = c_char[ num_ch - i - 1 ];
      else
	c = c_char[ i ];


      xk = xc2b_to_xk (c);
      switch (xk)
	{
	case XK_Delete      : scratch[ i ] = 255; break;
	case XK_BackSpace   : scratch[ i ] = 8; break;
	case XK_Tab         : scratch[ i ] = '\t'; break;
	case XK_Linefeed    : scratch[ i ] = '\n'; break;
	case XK_Return      : scratch[ i ] = '\r'; break;
	case XK_Clear       : scratch[ i ] = 11; break;
	case XK_Pause       : scratch[ i ] = 19; break;
	case XK_Scroll_Lock : scratch[ i ] = 20; break;
	//case XK_Sys_Req     : scratch[ i ] = 21; break;
	case XK_Escape      : scratch[ i ] = 27; break;
	default: scratch[ i ] = (xk & 0xFF);
	}
    }


  if (ret != NULL) (*ret) = (char *) scratch;
  if (len != NULL) (*len) = num_ch;
      

  /* hmmm... will gh_str2scm make it's own copy? */

  return gh_str2scm ((char *) scratch, num_ch);
}


SCM
WordSegment::set_leaf_ascii (XDisplay */*xd*/, char *contents, int len)
{
  int i;

  /* only does 8 bit values... (Latin) */

  ready_for_index (len);

  for (i=0; i<len; i++)
    {
      int xk = contents[ i ];
      switch (xk)
	{
	case 255:  c_char[ i ] = xk_to_2b (XK_Delete); break;
	case 8:    c_char[ i ] = xk_to_2b (XK_BackSpace); break;
	case '\t': c_char[ i ] = xk_to_2b (XK_Tab); break;
	case '\n': c_char[ i ] = xk_to_2b (XK_Linefeed); break;
	case '\r': c_char[ i ] = xk_to_2b (XK_Return); break;
	case 11:   c_char[ i ] = xk_to_2b (XK_Clear); break;
	case 19:   c_char[ i ] = xk_to_2b (XK_Pause); break;
	case 20:   c_char[ i ] = xk_to_2b (XK_Scroll_Lock); break;
	//case 21:   c_char[ i ] = xk_to_2b (XK_Sys_Req); break;
	case 27:   c_char[ i ] = xk_to_2b (XK_Escape); break;
	default:   c_char[ i ] = xk_to_2b (contents[ i ]); break;
	}
    }

  num_ch = len;

  /* free contents? */

  return SCM_BOOL_T;
}


SCM
WordSegment::size_leaf_xword (XDisplay */*xd*/, int *sz)
{
  int num = num_ch * 4;

  if (sz != NULL)
    (*sz) = num;

  return gh_int2scm (num);
}



SCM
WordSegment::get_leaf_xword (XDisplay */*xd*/, char **ret, int *len)
{
  int i;
  char *scratch;
  char *p;
  Attr_Direction direction;

  if (2 * (num_ch + 2) >= flip_string_max_size)
    {
      flip_string_max_size = 2 * (num_ch + 4);
      flip_string_scratch_space =
	(XChar2b *) XtRealloc ((char *) flip_string_scratch_space,
			       flip_string_max_size *
			       sizeof (XChar2b));
    }

  scratch = (char *) flip_string_scratch_space;
  p = scratch;

  direction = (Attr_Direction) getAttribute (at_direction);


  for (i=0; i<num_ch; i++)
    {
      XChar2b c;

      if (direction == DIR_rl)
	c = c_char[ num_ch - i - 1 ];
      else
	c = c_char[ i ];

      sprintf (p, "%02x", c.byte1);
      p += 2;		   
      sprintf (p, "%02x", c.byte2);
      p += 2;
    }

  if (ret != NULL) (*ret) = scratch;
  if (len != NULL) (*len) = num_ch * 4;

  /* hmmm... will gh_str2scm make it's own copy? */

  return gh_str2scm ((char *) scratch, num_ch * 4);
}


SCM WordSegment::set_leaf_xword (XDisplay */*xd*/, char *contents, int len)
{
  int i;
  char *p = contents;
  int byte1, byte2;

  len /= 4;

  ready_for_index (len);

  for (i=0; i<len; i++)
    {
      sscanf (p, "%02x%02x", &byte1, &byte2);
      c_char[ i ].byte1 = byte1;
      c_char[ i ].byte2 = byte2;
      p += 4;
    }

  /* free contents? */

  num_ch = len;

  return SCM_BOOL_T;
}



SCM WordSegment::size_leaf_ps (XDisplay *xd, int *sz)
{
  return size_leaf_ascii (xd, sz);
}


SCM WordSegment::get_leaf_ps (XDisplay *xd, char **ret, int *len)
{
  return get_leaf_ascii (xd, ret, len);
}


SCM WordSegment::set_leaf_ps (XDisplay *xd, char *chars, int len)
{
  return set_leaf_ascii (xd, chars, len);
}


SCM WordSegment::size_leaf_html (XDisplay *xd, int *sz)
{
  return size_leaf_ascii (xd, sz);
}


SCM WordSegment::get_leaf_html (XDisplay *xd, char **ret, int *len)
{
  return get_leaf_ascii (xd, ret, len);
}


SCM WordSegment::set_leaf_html (XDisplay *xd, char *chars, int len)
{
  return set_leaf_ascii (xd, chars, len);
}



void hook_up_guile_WordSegment ()
{
  gh_new_procedure1_0 ("wordsegment.new", (ghp) wordsegment_new);
}




#if 0
void
WordSegment::bump_cursor_past_cr (Position &where, rec_info *ri)
{
    /*** if the last character on this line is a line break, ***/
    /*** put the cursor on the next line. ***/

    if (_parent != NULL &&
	where.offset-1 >= 0 &&
	where.offset-1 < num_ch &&
	(c_char[ where.offset-1 ] == XK_Linefeed ||
	 c_char[ where.offset-1 ] == XK_Return ||
	 c_char[ where.offset-1 ] == 12 /* page break */ ))
    {
	Container *nxt = next_node ();
	if (nxt == NULL)
	{
	    WordSegment *after_me = new WordSegment ("", myAttributes);
	    _parent->this_doesnt_fit (this, after_me, ri);
	    nxt = next_node ();
	}

	if (nxt == NULL)
	    fprintf (stderr, "Couldn't place cursor on next line.\n");
	else
	{
	    where.offset = 0;
	    where.con = nxt;
	}
    }
}

#endif /* 0 */
