/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include "XCutBuffer.h"
#include "g.h"


XCutBuffer::XCutBuffer (Display *set_dpy)
{
    dpy = set_dpy;

    /*** Set up the atoms that we already know about. ***/

    natoms = 8;
    atom = (Atom *) XtMalloc (natoms * sizeof (Atom));
    atom[0] = XA_CUT_BUFFER0;
    atom[1] = XA_CUT_BUFFER1;
    atom[2] = XA_CUT_BUFFER2;
    atom[3] = XA_CUT_BUFFER3;
    atom[4] = XA_CUT_BUFFER4;
    atom[5] = XA_CUT_BUFFER5;
    atom[6] = XA_CUT_BUFFER6;
    atom[7] = XA_CUT_BUFFER7;

    root = RootWindow (dpy, DefaultScreen (dpy));
}

SCM xcutbuffer_new (SCM scm_dpy)
{
  Display *dpy = (Display *) gh_scm2ulong (scm_dpy);
  XCutBuffer *xcb = new XCutBuffer (dpy);
  return gh_ulong2scm ((unsigned long) xcb);
}


XCutBuffer::~XCutBuffer()
{
}



/********************************************/
/* Add an atom to the program's atom cache. */
/********************************************/

Atom
XCutBuffer::get_atom (int n, int ifexists)
{
    char tmp[32];

    if (n >= natoms)
    {
	atom = (Atom *) XtRealloc ( (char *) atom, (n+1) * sizeof (Atom));
	while (natoms < n+1)
	    atom[natoms++] = 0;
    }

    if (!atom[n])
    {
	sprintf (tmp, "CUT_BUFFER%d", n);
	atom[n] = XInternAtom (dpy, tmp, (Bool) ifexists);
    }
    return atom[n];
}



/************************************************************/
/* Fetch the contents of cut buffer n from the root window. */
/************************************************************/

char *
XCutBuffer::fetch_buffer (Atom a, unsigned long int *nb)
{
    Atom actl_type;
    int actl_format;
    unsigned long nitems, after;
    char *data;

    *nb = 0;
    if (XGetWindowProperty (dpy, root, a,
			   0L, 10000000L, False, XA_STRING,
			   &actl_type, &actl_format, &nitems,
			   &after, (unsigned char **) &data) != Success)
	return NULL;

    if (actl_type == XA_STRING && actl_format != 32)
    {
	*nb = nitems;
	return data;
    }


    printf ("non string info.\n");

    if (data != NULL)
	XFree (data);
    return NULL;
}



/************************************************************/
/* Store the string p into cut buffer n on the root window. */
/************************************************************/

void
XCutBuffer::store_buffer (char *p, int nb, Atom atom)
{
    XChangeProperty(dpy, root, atom, XA_STRING,
		    8, PropModeReplace, (unsigned char *) p, nb);
}



/************************************************/
/* Retrieve a value from one of the cut buffers */
/*  the return of doget should be XFree'ed by   */
/*  the caller of doget                         */
/*  n = which cut buffer (0-7)                  */
/*  *nb is set to the length                    */
/************************************************/

char *
XCutBuffer::doget (int n, unsigned long int *nb)
{
    Atom a;
    char *ptr;

    a = get_atom (n, True);
    if (a)
    {
	ptr = fetch_buffer (a, nb);
	return ptr;
    }

    return NULL;
}


SCM xcutbuffer_get (SCM scm_xcb, SCM scm_n)
{
  XCutBuffer *xcb = (XCutBuffer *) gh_scm2ulong (scm_xcb);
  int n = gh_scm2int (scm_n);
  char *ret;
  unsigned long int len;
  SCM scm_ret;

  ret = xcb->doget (n, &len);
  scm_ret = gh_str2scm (ret, len);
  XFree (ret);

  return scm_ret;
}



/*************************************************/
/* Load a new value into one of the cut buffers. */
/*************************************************/

void
XCutBuffer::doset(int n, char *ptr, int nb)
{
    store_buffer(ptr, nb, get_atom(n, False));
}


SCM xcutbuffer_set (SCM scm_xcb, SCM scm_n, SCM scm_str)
{
  XCutBuffer *xcb = (XCutBuffer *) gh_scm2ulong (scm_xcb);
  int n = gh_scm2int (scm_n);
  char *str;
  int len;

  str = gh_scm2newstr (scm_str, &len);
  xcb->doset (n, str, (unsigned long int) len);
  /* free str? */
  return SCM_BOOL_T;
}


void hook_up_guile_XCutBuffer ()
{
  gh_new_procedure1_0 ("xcutbuffer.new", (ghp) xcutbuffer_new);
  gh_new_procedure2_0 ("xcutbuffer.get", (ghp) xcutbuffer_get);
  gh_new_procedure3_0 ("xcutbuffer.set", (ghp) xcutbuffer_set);
}
