/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include <stdlib.h>
#include "XDisplay.h"
#include "g.h"
#include "utils.h"

extern verbose;

XDisplay::XDisplay (Display *set_display)
{
  display = set_display;

  if (display == NULL)
    {
      char *dpyname = getenv("DISPLAY");

      if (dpyname == NULL)
	/* well, just try something then. */
	dpyname=":0";

      display=XOpenDisplay(dpyname);

      if (display==NULL)
	{
	  fprintf (stderr, "cannot open X Display\n");
	  fprintf (stderr, "try setting the DISPLAY env variable.\n");
	  return;
	}
    }

  screen = DefaultScreen (display);
  colormap = DefaultColormap (display, screen);
  visual = DefaultVisual (display, screen);
  depth = DefaultDepth (display, screen);
  gc = DefaultGC (display, screen);

  fm = new FontManager (display);
  //fm->loadFonts ();

  if (depth == 8)
    {
      find_color_cube_on_current_colormap ();
      if (verbose)
	printf ("done finding color cube\n");
    }

  /* set a tile pattern */
  Pixmap tile = XCreatePixmap (display, RootWindow (display, screen),
			       4, 2, depth);

  if (verbose)
    printf ("creating tile pixmap...\n");

  XSetForeground (display, gc, BlackPixel (display, screen));
  XFillRectangle (display, tile, gc, 0, 0, 4, 2);
  XSetForeground (display, gc, WhitePixel (display, screen));
  XDrawPoint (display, tile, gc, 0, 0);
  XDrawPoint (display, tile, gc, 2, 0);
  XDrawPoint (display, tile, gc, 1, 1);
  XDrawPoint (display, tile, gc, 3, 1);

  if (verbose)
    printf ("setting tile pixmap...\n");

  XSetTile (display, gc, tile);

  if (verbose)
    printf ("done with tile pixmap.\n");
}


SCM xdisplay_new (SCM scm_set_display)
{
  Display *display = (Display *) gh_scm2ulong (scm_set_display);
  SCM ret;

  if (verbose)
    printf ("new XDisplay\n");

  XDisplay *xd = new XDisplay (display);

  if (verbose)
    printf ("after new XDisplay\n");

  ret = gh_ulong2scm ((unsigned long) xd);

  if (verbose)
    printf ("returning XDisplay\n");

  return ret;
}


SCM xdisplay_get_display (SCM scm_xdisplay)
{
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xdisplay);
  return gh_ulong2scm ((unsigned long) xd->display);
}

SCM xdisplay_get_screen (SCM scm_xdisplay)
{
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xdisplay);
  return gh_ulong2scm ((unsigned long) xd->screen);
}

SCM xdisplay_get_fm (SCM scm_xdisplay)
{
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xdisplay);
  return gh_ulong2scm ((unsigned long) xd->fm);
}



SCM xdisplay_set_display (SCM scm_xdisplay, SCM scm_display)
{
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xdisplay);
  Display *display = (Display *) gh_scm2ulong (scm_display);
  xd->display = display;
  return SCM_BOOL_T;
}

SCM xdisplay_set_screen (SCM scm_xdisplay, SCM scm_screen)
{
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xdisplay);
  int screen = (int) gh_scm2long (scm_screen);
  xd->screen = screen;
  return SCM_BOOL_T;
}

SCM xdisplay_set_fm (SCM scm_xdisplay, SCM scm_fm)
{
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xdisplay);
  FontManager *fm = (FontManager *) gh_scm2ulong (scm_fm);
  xd->fm = fm;
  return SCM_BOOL_T;
}



#define GICK(x) ((unsigned short int)(0xFFFF * x / 6))


static int find_closest_color (XColor *std_cm, int red, int green, int blue)
{
  int std_i;
  int best_diff = 65536, cur_diff;
  int best_index = 0;

  red = GICK (red);
  green = GICK (green);
  blue = GICK (blue);

  for (std_i = 0; std_i < 256; std_i ++)
    {
      cur_diff = Iabs (std_cm[ std_i ].red - red) +
	Iabs (std_cm[ std_i ].green - green) +
	Iabs (std_cm[ std_i ].blue - blue);

      if (cur_diff <= best_diff)
	{
	  best_diff = cur_diff;
	  best_index = std_i;
	}
    }

  return best_index;
}



void XDisplay::find_color_cube_on_current_colormap (void)
{
  int red, green, blue;
  int cube_i;
  int i;


  if (verbose)
    printf ("Finding a color cube...\n");

  if (depth != 8)
    {
      fprintf (stderr, "FINISH ME!!! depth not 8!\n");
    }

  color_cube = (XColor *) XtMalloc (sizeof (XColor) * 256);


  /* find out what the default colormap is like */

  for (i=0; i<256; i++)
    color_cube[ i ].pixel = i;

  XQueryColors (display, colormap, color_cube, 256);

  for (red=0; red<6; red++)
    for (green=0; green<6; green++)
      for (blue=0; blue<6; blue++)
	{
	  if (verbose)
	    printf ("color cube: %d %d %d\n", red, green, blue);
	  cube_i = red*36 + green*6 + blue;
	  color_hop[ cube_i ] = find_closest_color (color_cube,
						    red, green, blue);
	}
}



Pixel XDisplay::rgb_to_pixel (unsigned short int red,
			      unsigned short int green,
			      unsigned short int blue)
{
  if (depth == 8)
    {
      return rgb_color_cube_match (red, green, blue);
    }
  else
    {
      XColor xcolor;
      xcolor.red = red;
      xcolor.green = green;
      xcolor.blue = blue;
      xcolor.flags = DoRed | DoGreen | DoBlue;

      XAllocColor (display, colormap, &xcolor);
      return xcolor.pixel;
    }
}


Pixel XDisplay::rgb_color_cube_match (unsigned short int red,
				      unsigned short int green,
				      unsigned short int blue)
{
  if (color_cube == NULL)
    {
      fprintf (stderr, "Why are you trying to match on a color cube?\n");
      return (Pixel)0;
    }

  red /= (0xFFFF/6);
  green /= (0xFFFF/6);
  blue /= (0xFFFF/6);

  return color_cube[ color_hop[ (int) (red*36 + green*6 + blue) ]].pixel;
}


Pixmap XDisplay::image_to_pixmap (XImage *ximage)
{
  Pixmap pixmap;

  if (verbose)
    printf ("image_to_pixmap: depth is %d\n", depth);

  /* build the pixmap here */
  pixmap = XCreatePixmap (display, 
			  RootWindow (display, 0),
			  ximage->width, ximage->height,
			  depth);


  XPutImage (display,
	     pixmap,
	     gc,
	     ximage,
	     0, 0,
	     0, 0,
	     ximage->width, ximage->height);

  return pixmap;
}


int eat_ppm_err (char *m)
{
  fprintf (stderr, "%s\n", m);
  return 0;
}


int eat_ppm_header (FILE *fp, int *width, int *height, int *max_value)
{
  char buf[1024];
  int ppm_type;

  fgets (buf, 1024, fp);

  /* try to make sure this is a ppm file */
  if (strlen (buf) != 3) return eat_ppm_err ("didn't get 3 chars");
  if (buf[ 0 ] != 'P') return eat_ppm_err ("didn't get P");
  if (buf[ 1 ] < '0' || buf[ 1 ] > '9')
    return eat_ppm_err ("didn't get ppm_type");

  ppm_type = atoi (buf + 1);

  fgets (buf, 1024, fp);

  while (buf[0] == '#')
    fgets (buf, 1024, fp);

  if (sscanf (buf, "%d %d", width, height) != 2)
    return eat_ppm_err ("didn't get width and height");

  if ((*width) < 1 || (*height) < 1)
    return eat_ppm_err ("got 0 size");

  if (ppm_type == 6 ||
      ppm_type == 3)
    {
      if (fscanf (fp, "%d", max_value) != 1)
	return eat_ppm_err ("didn't get max value");

      if ((*max_value) < 1)
	return eat_ppm_err ("got 0 max value");
    }

  fgetc (fp);		/* eat up the last newline here */


  if (verbose)
    printf ("Reading a type %d ppm to (%dx%d) pixmap...\n",
	    ppm_type, *width, *height);

  return ppm_type;
}


XImage *XDisplay::read_ppm_file_to_ximage (char *filename)
{
  FILE *fp;
  int max_value;
  register int x, y;
  unsigned char *color_buffer;
  long int index;
  XImage *ximage;
  unsigned char *image_color_buffer;
  int dumby;
  int rshift=0, gshift=0, bshift=0;
  unsigned long rmask=0, gmask=0, bmask=0;
  unsigned long r, g, b;
  int bytes_per_pixel;
  int width, height;
  int ppm_type;
  int p4bit = 0x80;


  /* open the PPM file here */

  fp = fopen (filename, "r");
  if (!fp)
    {
      fprintf (stderr, "Could not open PPM file. (%s)\n", filename);
      return NULL;
    }

  if ((ppm_type = eat_ppm_header (fp, &width, &height, &max_value)) == 0)
    {
      fprintf (stderr, "Not in PPM formap? (%s)\n", filename);
      return NULL;
    }


  /* find out how many bytes_per_pixel we are going to end up with */
  /* THIS is a yikky hack -- how is this properly done? */

  ximage = XCreateImage (display,
			 visual,
			 depth,
			 ZPixmap,
			 0,
			 (char *)NULL,
			 0, 0,
			 8,
			 0);

  bytes_per_pixel = ceil ((double)(ximage->bits_per_pixel) / 8.0);
  XDestroyImage (ximage);

  /* allocate space for the color buffers */

  color_buffer = (unsigned char *) XtMalloc (3 * width * height);

  image_color_buffer = (unsigned char *)
    XtMalloc (bytes_per_pixel * width * height);


  /* read the PPM file here */
  if (ppm_type == 6)
    fread (color_buffer, 1, 3 * width * height, fp);
  if (ppm_type == 4)
    fread (color_buffer, 1, width * height, fp);


  ximage = XCreateImage (display,
			 visual,
			 depth,
			 ZPixmap,
			 0,
			 (char *)image_color_buffer,
			 width, height,
			 8,
			 0);


  if (visual->c_class == TrueColor || visual->c_class == DirectColor)
    {
      rmask = visual->red_mask;
      gmask = visual->green_mask;
      bmask = visual->blue_mask;
      
      rshift = 15 - highbit (rmask);
      gshift = 15 - highbit (gmask);
      bshift = 15 - highbit (bmask);

      if (verbose)
	printf ("visual type is TrueColor or DirectColor.\n"
		"  masks=(%lx, %lx, %lx), shifts = (%d, %d, %d)\n",
		rmask, gmask, bmask, rshift, gshift, bshift);
    }


  dumby = 0;
  index = 0;
  for (y=0; y<height; y++)
    {
      for (x=0; x< width; x++)
	{
	  r = g = b = 0;

	  if (ppm_type == 6)
	    {
	      r = color_buffer[index++] << 8;
	      g = color_buffer[index++] << 8;
	      b = color_buffer[index++] << 8;
	    }
	  if (ppm_type == 4)
	    {
	      if (p4bit == 0)
		{
		  index ++;
		  p4bit = 0x80;
		}

	      if (color_buffer[index] & p4bit) r = 0xFFFF;
	      if (color_buffer[index] & p4bit) g = 0xFFFF;
	      if (color_buffer[index] & p4bit) b = 0xFFFF;
	      p4bit >>= 1;
	    }
	  if (ppm_type == 3)
	    {
	      fscanf (fp, "%ld ", &r);
	      fscanf (fp, "%ld ", &g);
	      fscanf (fp, "%ld ", &b);
	      /*
	      r <<= 8;
	      g <<= 8;
	      b <<= 8;
	      */
	    }

	  if (ximage->bits_per_pixel == 32)
	    {
	      /* HERE, we are assuming that bytes_per_pixel = 4, */
	      /* and that an unsigned long is 4 bytes. */
		
	      if (rshift<0) r <<= (-rshift); else r >>= rshift;
	      if (gshift<0) g <<= (-gshift); else g >>= gshift;
	      if (bshift<0) b <<= (-bshift); else b >>= bshift;

	      ((unsigned long *)image_color_buffer)[ dumby++ ] = r | g | b;
	    }
	  if (ximage->bits_per_pixel == 16)
	    {
	      /* HERE, we are assuming that bytes_per_pixel = 2, */
	      /* and that an unsigned short is 2 bytes. */

	      if (rshift<0) r <<= (-rshift); else r >>= rshift;
	      if (gshift<0) g <<= (-gshift); else g >>= gshift;
	      if (bshift<0) b <<= (-bshift); else b >>= bshift;

	      r &= rmask;
	      g &= gmask;
	      b &= bmask;

	      ((unsigned short *)image_color_buffer)[ dumby++ ] = r | g | b;
	    }
	  else if (ximage->bits_per_pixel == 8)
	    {
	      /* HERE, we are assuming that bytes_per_pixel = 1. */
		
	      image_color_buffer[(unsigned short int)(dumby++)] =
		(unsigned char) rgb_color_cube_match ((unsigned short int) r,
						      (unsigned short int) g,
						      (unsigned short int) b);
	    }
	  else
	    /* slow, but reliable... */
	    XPutPixel (ximage, x, y, rgb_to_pixel ((unsigned short int) r,
						   (unsigned short int) g,
						   (unsigned short int) b));
	}
    }


  XtFree ((char *) color_buffer);

  fclose (fp);

  return (ximage);
}


Pixmap XDisplay::read_ppm_file_to_pixmap (char *filename, 
					  int *width,
					  int *height)
{
  Pixmap pixmap;
  XImage *ximage;
    
  ximage = read_ppm_file_to_ximage (filename);

  if (ximage == NULL)
    return (Pixmap) 0;

  pixmap = image_to_pixmap (ximage);

  *width = ximage->width;
  *height = ximage->height;

  XDestroyImage (ximage);
    
  return pixmap;
}


SCM xdisplay_read_ppm_file_to_pixmap (SCM scm_xdisplay, SCM scm_filename)
{
  XDisplay *xd = (XDisplay *) gh_scm2ulong (scm_xdisplay);
  char *filename = gh_scm2string (scm_filename);
  int width;
  int height;
  SCM ret;
  Pixmap p = xd->read_ppm_file_to_pixmap (filename, &width, &height);

  ret = gh_list (gh_ulong2scm ((unsigned long) p),
		 gh_int2scm (width),
		 gh_int2scm (height),
		 SCM_UNDEFINED);

  return ret;
}


void XDisplay::write_ppm_file_from_pixmap (char *filename,
					   int width, 
					   int height,
					   Pixmap pixmap)
{
  XImage *ximage;

  ximage = XGetImage (display,
		      pixmap,
		      0, 0,
		      width, height,
		      ~0,
		      ZPixmap);
    
  if (ximage == NULL)
    {
      fprintf (stderr, "XGetImage failed.\n");
      return;
    }

  write_ppm_file_from_ximage (filename, ximage);

  XDestroyImage (ximage);
}



void XDisplay::write_ppm_file_from_ximage (char *filename, XImage *ximage)
{
  FILE *fp;
  XColor xcolor;
  int x, y;
    
  fp = fopen (filename, "w");
  if (!fp)
    {
      fprintf (stderr, "Could not open file %s for output!\n", filename);
      return;
    }

  fprintf (fp, "P6\n");
  fprintf (fp, "%d %d\n", ximage->width, ximage->height);
  fprintf (fp, "255\n");

  if (verbose)
    printf ("width = %d, height = %d\n", ximage->width, ximage->height);

  for (y=0;y<ximage->height;y++)
    {
      for (x=0; x<ximage->width; x++)
	{
	  xcolor.pixel = XGetPixel (ximage, x, y);
	  /* xcolor.pixel = (unsigned long int)
	     (& (ximage->data[ y * ximage->width + x * 4 ])); */

	  xcolor.flags = DoRed|DoGreen|DoBlue;
	  XQueryColor (display, colormap, &xcolor);
	  
	  fprintf (fp, "%c%c%c",
		   xcolor.red >> 8,
		   xcolor.green >> 8,
		   xcolor.blue >> 8);
	}
    }

  fclose (fp);
}



void hook_up_guile_XDisplay ()
{
  gh_new_procedure1_0 ("xdisplay.new", (ghp) xdisplay_new);
  gh_new_procedure1_0 ("xdisplay.get-display", (ghp) xdisplay_get_display);
  gh_new_procedure1_0 ("xdisplay.get-screen", (ghp) xdisplay_get_display);
  gh_new_procedure1_0 ("xdisplay.get-fm", (ghp) xdisplay_get_fm);
  gh_new_procedure2_0 ("xdisplay.set-display", (ghp) xdisplay_set_display);
  gh_new_procedure2_0 ("xdisplay.set-screen", (ghp) xdisplay_set_display);
  gh_new_procedure2_0 ("xdisplay.set-fm", (ghp) xdisplay_set_fm);
  gh_new_procedure2_0 ("xdisplay.read-ppm-file-to-pixmap",
		       (ghp) xdisplay_read_ppm_file_to_pixmap);
}
