/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include <Xm/ScrollBar.h>
#include <X11/Xmu/StdSel.h>

#include "XWord_Window.h"
#include "Assoc.h"

#include "Document.h"
#include "Page.h"
#include "WordSegment.h"

#include "Tag.h"
#include "utils.h"


extern int verbose;

int horizontal_scroll_bar_position (Widget w, XWord_Window *xw);
int vertical_scroll_bar_position (Widget w, XWord_Window *xw);

XWord_Window::XWord_Window (XDisplay *set_xd,
			    Window set_w,
			    Container *set_doc,
			    Widget set_vsb, Widget set_hsb)
{
  int i;
  Pos *positions[ 3 ];

  xd = set_xd;
  w = set_w;
  gc = XCreateGC (xd->display, w, 0, (XGCValues *) NULL);

  scroll_x = scroll_y = 0;
  vsb = set_vsb;
  hsb = set_hsb;


  doc = set_doc;

  if (verbose)
    {
      printf ("sizeof unsigned long = %d, sizeof void * = %d\n",
	      sizeof (unsigned long int),
	      sizeof (void *));

      printf ("sizeof WordSegment = %d\n", sizeof (Container));
    }


  max_cut_slots = 16;
  num_cut_slots = 0;
  cuts = (Container **) XtMalloc (sizeof (Container *) * max_cut_slots);
  for (i=0; i<max_cut_slots; i++)
    cuts[ i ] = NULL;

  cur.con = doc->last_lowest_container ();
  cur.offset = 0;
  mark.con = doc->last_lowest_container ();
  mark.offset = 0;

  positions[ 0 ] = &cur;
  positions[ 1 ] = &mark;
  positions[ 2 ] = NULL;

  ReWrap (positions);

  paste_tag_attributes = NULL;
  paste_ws_attributes = NULL;

  num_rec = 0;
  max_rec = 0;
  selection_cur = NULL;
  rec_ids = NULL;
}


SCM xword_window_new (SCM scm_xd, SCM scm_w, SCM scm_doc,
		      SCM scm_vsb, SCM scm_hsb)
{
  XWord_Window *xw;
  XDisplay *set_xd = (XDisplay *) gh_scm2ulong (scm_xd);
  Window set_w = (Window) gh_scm2ulong (scm_w);
  Container *set_doc = (Container *) gh_scm2ulong (scm_doc);
  Widget set_vsb = (Widget) gh_scm2ulong (scm_vsb);
  Widget set_hsb = (Widget) gh_scm2ulong (scm_hsb);

  xw = new XWord_Window (set_xd, set_w, set_doc, set_vsb, set_hsb);

  return gh_ulong2scm ((unsigned long) xw);
}



SCM xword_window_set_paste_attributes (SCM scm_xw,
				       SCM scm_ws_attr,
				       SCM scm_tag_attr)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  xw->paste_tag_attributes = (Attributes *) gh_scm2ulong (scm_tag_attr);
  xw->paste_ws_attributes = (Attributes *) gh_scm2ulong (scm_ws_attr);
  return SCM_BOOL_T;
}

SCM xword_window_get_cursor_position (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  SCM scm_con = gh_ulong2scm ((unsigned long) xw->cur.con);
  SCM scm_pos = gh_int2scm (xw->cur.offset);
  return gh_list (scm_con, scm_pos, SCM_UNDEFINED);
}


SCM xword_window_get_mark_position (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  SCM scm_con = gh_ulong2scm ((unsigned long) xw->mark.con);
  SCM scm_pos = gh_int2scm (xw->mark.offset);
  return gh_list (scm_con, scm_pos, SCM_UNDEFINED);
}



SCM xword_window_set_cursor_position (SCM scm_xw, SCM scm_c, SCM scm_p)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int p = gh_scm2int (scm_p);

  xw->cur.con = c;
  xw->cur.offset = p;

  return SCM_BOOL_T;
}


SCM xword_window_set_mark_position (SCM scm_xw, SCM scm_c, SCM scm_p)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *c = (Container *) gh_scm2ulong (scm_c);
  int p = gh_scm2int (scm_p);

  xw->mark.con = c;
  xw->mark.offset = p;

  return SCM_BOOL_T;
}



SCM xword_window_set_attributes (SCM scm_xw,
				 SCM scm_new_attributes,
				 SCM scm_parent,
				 SCM scm_depth)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Attributes *new_a = (Attributes *) gh_scm2ulong (scm_new_attributes);
  Container *parent = (Container *) gh_scm2ulong (scm_parent);
  int depth = gh_scm2int (scm_depth);

  xw = NULL; // shut up the compiler

  Container *first = parent->first_lowest_container ();
  Container *last = parent->last_lowest_container ();
  Container *cur;
  Container *dont;

  while (first->find_depth () > depth)
    first = first->parent ();

  while (last->find_depth () > depth)
    last = last->parent ();

  dont = last->next_node ();

  for (cur = first; cur != dont; cur = cur->next_node ())
    cur->setAttributes (new_a);

  //Pos *positions[ 3 ];
  //positions[ 0 ] = &(xw->cur);
  //positions[ 1 ] = NULL;

  //xw->ReWrap (positions);
  //xw->Redraw ();

  return SCM_BOOL_T;
}


#if 0
SCM xword_window_set_leaf_ascii (SCM scm_xw, SCM scm_leaf, SCM scm_contents)
{
  Pos *positions[ 4 ];
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *leaf = (Container *) gh_scm2ulong (scm_leaf);
  Container *wrap_from = leaf;
  Container *wrap_to = leaf;
  Pos p;

  char *contents = gh_scm2string (scm_contents);
  int len = strlen (contents);


  p.con = leaf;
  p.offset = 0;

  if (wrap_from->prev_node () != NULL)
    wrap_from = wrap_from->prev_node ();

  positions[ 0 ] = &(xw->cur);
  positions[ 1 ] = &(xw->mark);
  positions[ 2 ] = &p;
  positions[ 3 ] = NULL;

  leaf->set_leaf_ascii (contents, len);

  leaf->Correct_Positions (positions);

  Container *fee = NULL;
  leaf->Wrap_Layer (wrap_from, wrap_to, positions, &fee);

  //if (fee == NULL)
  //fee = xw->doc;

  //xw->Redraw ();

  //return SCM_BOOL_T;
  return gh_ulong2scm ((unsigned long) p.con);
}
#endif


SCM xword_window_set_doc (SCM scm_xw, SCM scm_doc)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *set_doc = (Container *) gh_scm2ulong (scm_doc);
  xw->doc = set_doc;
  return SCM_BOOL_T;
}


SCM xword_window_set_cursor_xy (SCM scm_xw, SCM scm_x, SCM scm_y)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  int x = gh_scm2int (scm_x);
  int y = gh_scm2int (scm_y);

  xw->Set_Cursor_Position (x, y);

  return SCM_BOOL_T;
}


SCM xword_window_set_mark_xy (SCM scm_xw, SCM scm_x, SCM scm_y)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  int x = gh_scm2int (scm_x);
  int y = gh_scm2int (scm_y);

  xw->Set_Mark_Position (x, y);

  return SCM_BOOL_T;
}


void
XWord_Window::Set_Cursor_Position (int x, int y)
{
  x -= horizontal_scroll_bar_position (hsb, this);
  y -= vertical_scroll_bar_position (vsb, this);
  Pos p = doc->xy_to_position (x, y);
  if (p.con == NULL)
    return;

  cur = p;
}


void
XWord_Window::Set_Mark_Position (int x, int y)
{
  x -= horizontal_scroll_bar_position (hsb, this);
  y -= vertical_scroll_bar_position (vsb, this);
  Pos p = doc->xy_to_position (x, y);
  if (p.con == NULL)
    return;

  mark = p;
}


SCM xword_window_get_cursor_xy (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  int x, y, h;
  SCM scm_x, scm_y;

  xw->cur.con->offset_to_xy (xw->cur.offset, &x, &y, &h);
  scm_x = gh_int2scm (x);
  scm_y = gh_int2scm (y);

  return gh_list (scm_x, scm_y, SCM_UNDEFINED);
}


SCM xword_window_get_mark_xy (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  int x, y, h;
  SCM scm_x, scm_y;

  xw->mark.con->offset_to_xy (xw->mark.offset, &x, &y, &h);
  scm_x = gh_int2scm (x);
  scm_y = gh_int2scm (y);

  return gh_list (scm_x, scm_y, SCM_UNDEFINED);
}


SCM xword_window_delete_character (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *ret = xw->Delete_Character ();
  return gh_ulong2scm ((unsigned long) ret);
}


Container *
XWord_Window::Delete_Character (void)
{
  Pos *positions[ 3 ];
  Container *wrap_from;

  cur = cur.con->hop_back_position (cur);

  if (cur.offset == 0)
    {
      /* printf ("can't delete.\n"); */
      return doc;
    }

  cur.offset --;
  cur.con->Delete_Children (cur.offset, 1, NULL);

  wrap_from = cur.con->prev_node ();
  if (wrap_from == NULL)
    wrap_from = cur.con;

  positions[ 0 ] = &cur;
  positions[ 1 ] = &mark;;
  positions[ 2 ] = NULL;

  Container *fee = NULL;
  cur.con->Wrap_Layer (xd, wrap_from, cur.con, positions, NULL);
  cur.con->Correct_Positions (positions);
  if (fee == NULL)
    fee = doc;

  // printf ("topped at %p\n", fee);

  return fee;
}


SCM xword_window_insert_xk (SCM scm_xw, SCM scm_xk, SCM scm_at)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  KeySym c = (KeySym) gh_scm2ulong (scm_xk);
  Attributes *at = (Attributes *) gh_scm2ulong (scm_at);
  Container *ret = xw->Insert_Character (c, at);
  return gh_ulong2scm ((unsigned long) ret);
}


/* FIX ME!  insert_tag and insert_character are very
   similar -- make routine for common code */


Container *
XWord_Window::Insert_Tag (char *tag, Attributes *current_attributes)
{
  Pos *positions[ 3 ];
  Container *wrap_from = cur.con;
  Container *wrap_to = cur.con;


  if (cur.con->myAttributes != current_attributes)
    {
      /*** may have done a style change mid-word... ***/
      Container *t = new Tag (current_attributes);
      int ws_index; /* current con's index into parent's children */

      cur = cur.con->hop_back_position (cur);

      if (cur.con->myAttributes != current_attributes)
	{
	  if (cur.offset == 0)
	    {
	      /*** insert the new empty word segment ***/
	      ws_index = cur.con->parent ()->childs_index (cur.con);
	      cur.con->parent ()->Insert_Children (ws_index, 1, &t);
	    }
	  else
	    {
	      /*** split the current word segment into 2 parts ***/
	      cur.con->Bump_Children (cur.offset);
	      wrap_to = cur.con->next_node ();
	      if (wrap_to == NULL)
		wrap_to = t;
	      
	      /*** insert the new empty con ***/
	      ws_index = cur.con->parent ()->childs_index (cur.con);
	      cur.con->parent ()->Insert_Children (ws_index+1, 1, &t);
	    }
	  
	  /*** move cursor to new empty word segment ***/
	  cur.con = t;
	  cur.offset = 0;
	}
    }

  tag_info *ti = new tag_info (tag);

  cur.con->Insert_Children (cur.offset, 1, (Container **) (&ti));
  cur.offset ++;

  if (wrap_from->prev_node () != NULL)
    wrap_from = wrap_from->prev_node ();


  positions[ 0 ] = &cur;
  positions[ 1 ] = &mark;
  positions[ 2 ] = NULL;

  Container *fee = NULL;
  cur.con->Wrap_Layer (xd, wrap_from, wrap_to, positions, &fee);

  cur.con->Correct_Positions (positions);

  if (fee == NULL)
    fee = doc;

  // printf ("topped at %p\n", fee);

  return fee;
}



SCM xword_window_insert_tag (SCM scm_xw, SCM scm_tag, SCM scm_at)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  char *tag = gh_scm2string (scm_tag);
  Attributes *at = (Attributes *) gh_scm2ulong (scm_at);
  Container *ret;

  ret = xw->Insert_Tag (tag, at);
  return gh_ulong2scm ((unsigned long) ret);
}


Container *
XWord_Window::Insert_Character (KeySym c, Attributes *current_attributes)
{
  Pos *positions[ 3 ];
  Container *wrap_from = cur.con;
  Container *wrap_to = cur.con;
  //short char_c = (short) c;


  if (cur.con->myAttributes != current_attributes)
    {
      /*** may have done a style change mid-word... ***/
      Container *t = new WordSegment (current_attributes);
      int ws_index; /* current word segment's index into parent's children */

      cur = cur.con->hop_back_position (cur);

      if (cur.con->myAttributes != current_attributes)
	{
	  if (cur.offset == 0)
	    {
	      /*** insert the new empty word segment ***/
	      ws_index = cur.con->parent ()->childs_index (cur.con);
	      cur.con->parent ()->Insert_Children (ws_index, 1, &t);
	    }
	  else
	    {
	      /*** split the current word segment into 2 parts ***/
	      cur.con->Bump_Children (cur.offset);
	      wrap_to = cur.con->next_node ();
	      if (wrap_to == NULL)
		wrap_to = t;
	    
	      /*** insert the new empty word segment ***/
	      ws_index = cur.con->parent ()->childs_index (cur.con);
	      cur.con->parent ()->Insert_Children (ws_index+1, 1, &t);
	    }
      
	  /*** move cursor to new empty word segment ***/
	  cur.con = t;
	  cur.offset = 0;
	}
    }

  XChar2b xc = xk_to_2b (c);
  cur.con->Insert_Children (cur.offset, 1, (Container **) (&xc));
  cur.offset ++;

  if (wrap_from->prev_node () != NULL)
    wrap_from = wrap_from->prev_node ();


  positions[ 0 ] = &cur;
  positions[ 1 ] = &mark;
  positions[ 2 ] = NULL;

  Container *fee = NULL;
  cur.con->Wrap_Layer (xd, wrap_from, wrap_to, positions, &fee);

  cur.con->Correct_Positions (positions);

  if (fee == NULL)
    fee = doc;

  // printf ("topped at %p\n", fee);

  return fee;
}


SCM xword_window_insert_character (SCM scm_xw, SCM scm_c, SCM scm_at)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  int c = gh_scm2char (scm_c);
  Attributes *at = (Attributes *) gh_scm2ulong (scm_at);
  Container *ret;
  KeySym xk = c;

  switch (c)
    {
    case 255:  xk = XK_Delete; break;
    case 8:    xk = XK_BackSpace; break;
    case '\t': xk = XK_Tab; break;
    case '\n': xk = XK_Linefeed; break;
    case '\r': xk = XK_Return; break;
    case 11:   xk = XK_Clear; break;
    case 19:   xk = XK_Pause; break;
    case 20:   xk = XK_Scroll_Lock; break;
    //case 21:   xk = XK_Sys_Req; break;
    case 27:   xk = XK_Escape; break;
    default:   break;
    }

  ret = xw->Insert_Character (xk, at);
  return gh_ulong2scm ((unsigned long) ret);
}



/***************************/
/**** Cut and Paste Code ***/
/***************************/


#define NUM_ATOMS 6
enum atoms
{
  A_TEXT,
  A_TARGETS,
  A_MULTIPLE,
  A_XWORD,
  A_FILENAME,
  A_COMPOUND_TEXT
};


static Atom get_atom (Display *display, int atom_num)
{
    static Atom atom_list[ NUM_ATOMS ] = {0, 0, 0, 0, 0, 0};
    static char *atom_names[ NUM_ATOMS ] = {"TEXT",
					    "TARGETS",
					    "MULTIPLE",
					    "XWORD",
					    "FILENAME",
					    "COMPOUND_TEXT"};
    
    if (atom_list[ atom_num ] == 0)
      atom_list[ atom_num ] =
	XInternAtom (display, atom_names[ atom_num ], False);

    return atom_list[ atom_num ];
}




void
XWord_Window::set_number_of_cut_slots (int n)
{
  int i;

  if (max_cut_slots == n)
    return;

  if (max_cut_slots < n)
    {
      for (i=n; i<max_cut_slots; i++)
	if (cuts[ i ] != NULL)
	  delete cuts[ i ];
    }

  max_cut_slots = n;
  if (num_cut_slots > n)
    num_cut_slots = n;

  cuts = (Container **) realloc (cuts, sizeof (Container *) * max_cut_slots);

  for (i=num_cut_slots; i<max_cut_slots; i++)
    cuts[ i ] = NULL;
}



void
XWord_Window::rotate_cut_slots ()
{
  int i;
  Container *t;

  t = cuts[ max_cut_slots-1 ];

  for (i=max_cut_slots-1; i>0; i--)
    cuts[ i ] = cuts[ i-1 ];

  cuts[ 0 ] = t;
}


void
XWord_Window::rotate_used_cut_slots ()
{
  int i;
  Container *t;

  if (num_cut_slots == 0)
    return;

  t = cuts[ num_cut_slots-1 ];

  for (i=num_cut_slots-1; i>0; i--)
    cuts[ i ] = cuts[ i-1 ];

  cuts[ 0 ] = t;

  //selection_cur = NULL;
  selection_first.con = cuts[ 0 ]->first_lowest_container ();
  selection_first.offset = 0;
  selection_last.con = cuts[ 0 ]->last_lowest_container ();
  selection_last.offset = selection_last.con->num_ch;
}


SCM xword_window_rotate_used_cut_slots (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  xw->rotate_used_cut_slots ();
  return SCM_BOOL_T;
}


/* this should only be called from receiver_get */
void XWord_Window::receiver_add (XtRequestId receiver_id)
{
  if (num_rec == max_rec)
    {
      max_rec += 2;
      selection_cur =
	(Container **) XtRealloc ((char *) selection_cur,
				  sizeof (Container *) * max_rec);
      rec_ids = (XtRequestId *) XtRealloc ((char *) rec_ids,
					   sizeof (XtRequestId) * max_rec);
    }

  rec_ids[ num_rec ] = receiver_id;
  selection_cur[ num_rec ] = NULL;

  if (verbose)
    printf ("   adding receiver %ld in slot %d\n",
	    (unsigned long int) receiver_id, num_rec);

  num_rec ++;
}

int XWord_Window::receiver_exists (XtRequestId receiver_id)
{
  int i;

  for (i=0; i<num_rec; i++)
    if (rec_ids[ i ] == receiver_id)
      return 1;

  return 0;
}


Container **XWord_Window::receiver_get (XtRequestId receiver_id)
{
  int i;

  for (i=0; i<num_rec; i++)
    if (rec_ids[ i ] == receiver_id)
      return selection_cur + i;

  /* didn't find it.  add a new one */
  receiver_add (receiver_id);
  return receiver_get (receiver_id);
}


void XWord_Window::receiver_rm (XtRequestId receiver_id)
{
  int i;

  for (i=0; i<num_rec; i++)
    if (rec_ids[ i ] == receiver_id)
      {
	int j;

	if (verbose)
	  printf ("   removing receiver %ld in slot %d, "
		  "leaving %d receivers\n",
		  (unsigned long int) receiver_id, i, num_rec-1);

	for (j=i+1; j<num_rec; j++)
	  rec_ids[ j-1 ] = rec_ids[ j ];
	num_rec --;
	return;
      }

  fprintf (stderr, "warning -- tried to remove a non existant receiver.\n");
}


/*******************/
/*** data source ***/
/*******************/


Boolean _convert_proc (Widget w,
		       Atom *selection,
		       Atom *target,
		       Atom *type,
		       XtPointer *value,
		       unsigned long *length,
		       int *format,
		       unsigned long *max_length,
		       XtPointer client_data,
		       XtRequestId *receiver_id)
{
  XWord_Window *xw = (XWord_Window *) client_data;
  return xw->convert_proc (w, selection, target, type,
			   value, length, format, max_length,
			   client_data, receiver_id);
}


Boolean
XWord_Window::convert_proc (Widget /*w*/,
			    Atom *selection,
			    Atom *target,
			    Atom *type,
			    XtPointer *value,
			    unsigned long *length,
			    int *format,
			    unsigned long *max_length,
			    XtPointer /*client_data*/,
			    XtRequestId *receiver_id)
{
  /*** for debugging ***/
  if (verbose)
  {
    char *sel_name = XGetAtomName (xd->display, *selection);
    char *target_name = XGetAtomName (xd->display, *target);

    printf ("get_primary: selection = '%s', target = '%s', rqid = %ld/%ld\n",
	    sel_name, target_name,
	    (long int)receiver_id, (long int) (*receiver_id));

    printf ("   max_length = %ld\n", *max_length);

    XFree (sel_name);
    XFree (target_name);
  }


  if (*target == get_atom (xd->display, A_TARGETS))
    {
      /* "MECHANISM, BUT NOT POLICY" */
      Atom *targets;
      int num_targets = 5;

      /* this is a bit bogus -- multiple receivers
	 at once could wedge things */

      if (! receiver_exists (*receiver_id))
	{
	  receiver_add (*receiver_id);

	  if (verbose)
	    printf ("convert_proc got XA_TARGETS\n");
	  targets = (Atom *)XtMalloc(sizeof(Atom) * num_targets);
	  targets[0] = XA_STRING;
	  targets[1] = get_atom (xd->display, A_TEXT);
	  targets[2] = get_atom (xd->display, A_TARGETS);
	  targets[3] = get_atom (xd->display, A_MULTIPLE);
	  targets[4] = get_atom (xd->display, A_XWORD);
	  *type = XA_ATOM;
	  *value = (XtPointer) targets;
	  *length = num_targets;
	  *format = 32;
	  if (verbose)
	    printf ("  end XWord_Window::convert_proc -- targets\n");
	  return True;
	}
      else
	{
	  //receiver_rm (*receiver_id);
	  *value = "done";
	  *length = (unsigned long int) 0;
	  *type = XA_ATOM;
	  *format = 32;
	  return True;
	}
    }
  else if (*target == XA_STRING ||
	   *target == get_atom (xd->display, A_TEXT) ||
	   *target == get_atom (xd->display, A_XWORD))
    {
      Container *first = selection_first.con;
      Container **curnt;
      Container *last = selection_last.con;
      int len;
      int bts;
      char *foo;

      if (*target == XA_STRING || *target == get_atom (xd->display, A_TEXT))
	bts = 1; /* 1 byte per character in ascii format */
      else
	bts = 4; /* 4 bytes per character in the XWord format */


      curnt = receiver_get (*receiver_id);
      if ((*curnt) == NULL)
	(*curnt) = first;
      else
	(*curnt) = (*curnt)->next_node ();

      *type = *target;

      if ((*curnt) == NULL || (*curnt) == last->next_node ())
	{
	  *value = "done";
	  *length = (unsigned long int) 0;
	  *format = bts * 8;
	  return True;
	}


      if (bts == 1)
	(*curnt)->get_leaf_ascii (xd, &foo, &len);
      else if (bts == 4)
	(*curnt)->get_leaf_xword (xd, &foo, &len);


      /** **/
      if ((*curnt) == selection_first.con)
	{
	  len -= (selection_first.offset * bts);
	  *value = foo + (selection_first.offset * bts);
	}
      else
	*value = foo;

      if ((*curnt) == selection_last.con)
	len -= ((selection_last.con->num_ch - selection_last.offset) * bts);
      /** **/


      *format = bts * 8; /* format is in bits. */

      *length = (unsigned long int) len / bts;

      if ((unsigned int) len >= (*max_length))
	{
	  if (verbose)
	    fprintf (stderr, "convert_proc: len > max_length\n");
	  *length = *max_length; /* chop */
	}

      return True;
    }
  else if (*target == XA_PIXMAP ||
	   *target == XA_BITMAP)
    {
      if (! receiver_exists (*receiver_id))
	{
	  Container *c;
	  Pixmap *ret;
	  Pixmap np, pix;

	  receiver_add (*receiver_id);

	  /* look for a pixmap in the selected area */

	  for (c = selection_first.con;
	       c != NULL && c != selection_last.con->next_node ();
	       c = c->next_node ())
	    {
	      if (c->type_n () == T_Tag &&
		  c->num_ch > 0 &&
		  c->c_tag[ 0 ]->to_number () == TAG_img)
		{
		  pix = c->c_tag[ 0 ]->get_pixmap ();
		  
		  np = XCreatePixmap (xd->display, w,
				      c->_size.right, c->_size.up,
				      xd->depth);
		  XCopyArea (xd->display, pix, np, xd->gc,
			     0, 0, c->_size.right, c->_size.up, 0, 0);
		  

		  if (verbose)
		    printf ("sent (%d, %d, %ld), depth=%d\n",
			    c->_size.right, c->_size.up, np, xd->depth);
		    
		  //ret = XtNew (Pixmap);
		  ret = (Pixmap *) XtMalloc (sizeof (Pixmap));

		  *ret = np;
		  *value = (XtPointer) ret;
		    
		  *type = XA_PIXMAP;
		  *length = 1;
		  *format = 32;
		    
		  return True;
		}
	    }
	  return False; /* no pictures to send. */
	}
      else
	{
	  /* we've met before... haven't we! */
	  //receiver_rm (*receiver_id);
	  *value = "done";
	  *length = (unsigned long int) 0;
	  *type = XA_PIXMAP;
	  *format = 32;
	  return True;
	}
    }


  if (verbose)
    printf ("got strange target...\n");

  if (XmuConvertStandardSelection (vsb,
				   XtLastTimestampProcessed (xd->display),
				   selection, target, type,
				   (caddr_t *)value, length, format))
    return True;

  return False;
}


void lose_selection (Widget /*w*/,
		     Atom */*selection*/,
		     XtPointer client_data)
{
  XWord_Window *xw = (XWord_Window *) client_data;
  xw->unhighlight ();
  //xw->selection_cur = NULL;
  if (verbose)
    printf ("lost selection.\n");
  xw->Redraw ();
}

void done_proc (Widget /* widget */,
		Atom * /* selection */,
		Atom * /* target */,
		XtRequestId * receiver_id,
		XtPointer client_data)

{
  XWord_Window *xw = (XWord_Window *) client_data;
  //xw->selection_cur = NULL;
  xw->receiver_rm (*receiver_id);
  if (verbose)
    printf ("done proc called.\n");
}

void cancel_proc (Widget /* widget */,
		  Atom * /* selection */,
		  Atom * /* target */,
		  XtRequestId *receiver_id,
		  XtPointer client_data)
{
  XWord_Window *xw = (XWord_Window *) client_data;
  //xw->selection_cur = NULL;
  xw->receiver_rm (*receiver_id);
  if (verbose)
    fprintf (stderr, "cancel proc called.\n");
}


void
XWord_Window::Copy_To_Primary (Container *first, int offset_first,
			       Container *last, int offset_last)
{
  Boolean ok;

  ok = XtOwnSelectionIncremental ((Widget) vsb,
				  XA_PRIMARY,
				  //CurrentTime,
				  XtLastTimestampProcessed (xd->display),
				  _convert_proc,
				  lose_selection,
				  done_proc,
				  cancel_proc,
				  (XtPointer) this);

  if (!ok)
    {
      fprintf (stderr, "couldn't get ownership of Primary Selection\n");
      return;
    }

  //selection_cur = NULL;
  selection_first.con = first;
  selection_first.offset = offset_first;
  selection_last.con = last;
  selection_last.offset = offset_last;
}


SCM xword_window_copy_to_primary (SCM scm_xw,
				  SCM scm_c0, SCM scm_i0,
				  SCM scm_c1, SCM scm_i1)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *c0 = (Container *) gh_scm2ulong (scm_c0);
  int i0 = gh_scm2int (scm_i0);
  Container *c1 = (Container *) gh_scm2ulong (scm_c1);
  int i1 = gh_scm2int (scm_i1);

  xw->Copy_To_Primary (c0, i0, c1, i1);
  return SCM_BOOL_T;
}



void
XWord_Window::copy_to_cut_slot (Container *first, int offset_first,
				Container *last, int offset_last)
{
  Boolean ok;
  Container *c;
  Container *cut;
  Container *end = last->next_node ();

  rotate_cut_slots ();
  if (cuts[ 0 ] != NULL)
    delete (cuts[ 0 ]);

  cuts[ 0 ] = new Document (doc->myAttributes);
  cut = cuts[ 0 ];

  for (c=first; c != end; c=c->next_node ())
    {
      Container *d = c->dup ();
      cut->Insert_Children (cut->num_ch, 1, &d);
    }



  ok = XtOwnSelectionIncremental ((Widget) vsb,
				  XA_PRIMARY,
				  //CurrentTime,
				  XtLastTimestampProcessed (xd->display),
				  _convert_proc,
				  lose_selection,
				  done_proc,
				  cancel_proc,
				  (XtPointer) this);

  if (!ok)
    {
      fprintf (stderr, "couldn't get ownership of Primary Selection\n");
      return;
    }

  //selection_cur = NULL;
  selection_first.con = cuts[ 0 ]->first_lowest_container ();
  selection_first.offset = offset_first;
  selection_last.con = cuts[ 0 ]->last_lowest_container ();
  selection_last.offset = offset_last;
}


SCM xword_window_copy_to_cut_slot (SCM scm_xw,
				   SCM scm_c0, SCM scm_i0,
				   SCM scm_c1, SCM scm_i1)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *c0 = (Container *) gh_scm2ulong (scm_c0);
  int i0 = gh_scm2int (scm_i0);
  Container *c1 = (Container *) gh_scm2ulong (scm_c1);
  int i1 = gh_scm2int (scm_i1);

  xw->copy_to_cut_slot (c0, i0, c1, i1);
  return SCM_BOOL_T;
}



/*****************/
/*** data sink ***/
/*****************/


int atom_is_in_list (Atom a, Atom *data, int len)
{
  int i;

  for (i=0; i<len; i++)
    if (data[ i ] == a)
      return 1;

  return 0;
}


void _get_primary_proc (Widget w,
			XtPointer client_data,
			Atom *selection,
			Atom *type,
			XtPointer value,
			unsigned long *length,
			int *format)
{
  XWord_Window *xw = (XWord_Window *) client_data;
  xw->get_primary_proc (w, client_data, selection,
			type, value, length, format);
}


void XWord_Window::get_primary_proc (Widget /*w*/,
				     XtPointer /*client_data*/,
				     Atom *selection,
				     Atom *type,
				     XtPointer value,
				     unsigned long *length,
				     int *format)
{
  unsigned long int i;

  if (*type == 0 /*XT_CONVERT_FAIL*/ || *length == 0 || value == NULL)
    {
      fprintf (stderr, "get_primary_proc: convert failed?\n");
      return;
    }

  /*** for debugging ***/
  if (verbose)
  {
    char *sel_name = XGetAtomName (xd->display, *selection);
    char *type_name = XGetAtomName (xd->display, *type);

    printf ("get_primary_proc: selection = '%s', "
	    "type = '%s', *format = %d\n",
	    sel_name, type_name, *format);

    XFree (sel_name);
    XFree (type_name);
  }


  /* this happens after user does a paste. */
  if (*type == XA_ATOM && *format == 32)
    {
      int i;
      int len = (int) (*length);
      Atom *data = (Atom *) value;
      Atom use = 0;

      if (verbose)
	{
	  printf ("----------------------------\n");
	  printf ("targets are coming back... \n");
	  for (i=0; i<len; i++)
	    {
	      char *nm = XGetAtomName (xd->display, data[ i ]);
	      printf ("  %s\n", nm);
	      XFree (nm);
	    }
	  printf ("----------------------------\n");
	}

      /* pick our favorite one */

      if (atom_is_in_list (get_atom (xd->display, A_XWORD), data, len))
	use = get_atom (xd->display, A_XWORD);
      else if (atom_is_in_list (get_atom (xd->display, A_COMPOUND_TEXT),
				data, len))
	use = get_atom (xd->display, A_COMPOUND_TEXT);
      else if (atom_is_in_list (XA_STRING, data, len))
	use = XA_STRING;
      else if (atom_is_in_list (get_atom (xd->display, A_TEXT), data, len))
	use = get_atom (xd->display, A_TEXT);
      else if (atom_is_in_list (XA_PIXMAP, data, len))
	use = XA_PIXMAP;
      else if (atom_is_in_list (XA_BITMAP, data, len))
	use = XA_BITMAP;
      else if (atom_is_in_list (get_atom (xd->display, A_FILENAME),
				data, len))
	  use = get_atom (xd->display, A_FILENAME);
      else
	fprintf (stderr, "can't deal with any available target types.\n");

      if (use != 0)
	{
	  char *nm = XGetAtomName (xd->display, use);
	  if (verbose)
	    printf ("  requesting target type '%s' from owner.\n", nm);
	  XFree (nm);

	  XtGetSelectionValue (vsb,
			       XA_PRIMARY,
			       use,
			       _get_primary_proc,
			       (XtPointer) this,
			       XtLastTimestampProcessed (xd->display)
			       );
	}
    }
  /******/
  else if (*type == XA_STRING ||
	   *type == get_atom (xd->display, A_TEXT) ||
	   *type == get_atom (xd->display, A_COMPOUND_TEXT))
    {
      for (i=0; i<(*length); i++)
	{
	  int c = ((char *)value)[ i ];
	  KeySym xk = c;
	  switch (c)
	    {
	    case 255:  xk = XK_Delete; break;
	    case 8:    xk = XK_BackSpace; break;
	    case '\t': xk = XK_Tab; break;
	    case '\n': xk = XK_Linefeed; break;
	    case '\r': xk = XK_Return; break;
	    case 11:   xk = XK_Clear; break;
	    case 19:   xk = XK_Pause; break;
	    case 20:   xk = XK_Scroll_Lock; break;
	    case 27:   xk = XK_Escape; break;
	    default:   break;
	    }
	  
	  Insert_Character (xk, paste_ws_attributes);
	}
    }
  /******/
  else if (*type == get_atom (xd->display, A_XWORD))
    {
      static char *sp_scratch = NULL;
      static int sp_max = 0;
      static int sp_read = 0;
      static int sp_write = 0;

      int bts = (*format) >> 3; /* format is in bits */
      int len = (*length) * bts;

      int start=0, end=0;
      Container_Type ct;

      size_scratch (&sp_scratch, &sp_max, sp_write + len);

      memcpy ((void *) (sp_scratch + sp_write), value, len);
      sp_write += len;


      while (1)
	{
	  parse_stream (sp_scratch, &sp_read, &sp_write,
			&start, &end, &ct, bts);

	  switch (ct)
	    {
	    case T_unknown:
	      return;
	    case T_Tag:
 	      Insert_Tag ("", paste_tag_attributes);
 	      cur.con->set_leaf_xword (xd, sp_scratch + start, end - start);
	      break;
	    case T_WordSegment:
	      {
		int byte1, byte2;
		KeySym xk;
		sscanf (sp_scratch + start, "%02x%02x", &byte1, &byte2);
		xk = (byte1 << 8) | byte2;
		Insert_Character (xk, paste_ws_attributes);
		break;
	      }
	    default:
	      xw_assert (0);
	      break;
	    }
	}
    }
  /******/
  else if (*type == XA_PIXMAP ||
	   *type == XA_BITMAP)
    {
      static char pm_scratch[ 400 ]; /* FIX ME -- arbitrary limit */
      Pixmap pix, np;
      int x, y;
      unsigned int pwidth, pheight, border_width, depth;
      Window root;

      pix = *((Pixmap *) value);

      XGetGeometry (xd->display, pix, &root,
		    &x, &y, &pwidth, &pheight, &border_width, &depth);
		    

      np = XCreatePixmap (xd->display, w,
			  pwidth, pheight,
			  xd->depth);
      XCopyArea (xd->display, pix, np, xd->gc,
		 0, 0, pwidth, pheight, 0, 0);
		  

      sprintf (pm_scratch, "<img pixmap=%ld width=%d height=%d>",
	       np, pwidth, pheight);
      Insert_Tag (pm_scratch, paste_tag_attributes);

      /* destroy pix? */
    }
  /******/
  else if (*type == get_atom (xd->display, A_FILENAME))
    {
      printf ("got FILENAME selection...?\n");
    }
  /******/
  else
    {
      char *type_name = XGetAtomName (xd->display, *type);
      printf ("ignoring selection of type '%s'\n", type_name);
      XFree (type_name);
    }


  /* XtFree((char *)value); */  /* ??? */
  Redraw (doc);
}


void
XWord_Window::Get_Primary_Targets ()
{
  XtGetSelectionValue (vsb,
		       XA_PRIMARY,
		       get_atom (xd->display, A_TARGETS),
		       _get_primary_proc,
		       (XtPointer) this,
		       XtLastTimestampProcessed (xd->display)
		       );

  return;
}


SCM xword_window_get_primary_targets (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  xw->Get_Primary_Targets ();
  return SCM_BOOL_T;
}


void
XWord_Window::Get_Primary ()
{
  XtGetSelectionValue (vsb,
		       XA_PRIMARY,
		       XA_STRING,
		       _get_primary_proc,
		       (XtPointer) this,
		       //CurrentTime);
		       XtLastTimestampProcessed (xd->display)
		       );

  return;
}


SCM xword_window_get_primary (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  xw->Get_Primary ();
  return SCM_BOOL_T;
}


/*********************************/
/*** end of cut and paste code ***/
/*********************************/



void
XWord_Window::highlight (Container *first, int offset_first,
			 Container *last, int offset_last,
			 Attributes *attr_s,
			 Attributes *attr,
			 Attributes *attr_e)
{
  Pos a, b;
  Container *c;
  Container *end;

  a.con = first;
  a.offset = offset_first;
  b.con = last;
  b.offset = offset_last;

  /*
  if (!first->find_order (a, b))
    {
      first = b.con;
      offset_first = b.offset;
      last = a.con;
      offset_last = a.offset;
    }
    */

  for (c=first, end=last->next_node(); c != end; c=c->next_node())
    c->tmpAttributes = attr;

  first->tmpAttributes = attr_s;
  last->tmpAttributes = attr_e;
}


SCM xword_window_highlight (SCM scm_xw,
			    SCM scm_c0, SCM scm_i0,
			    SCM scm_c1, SCM scm_i1,
			    SCM scm_attr_s,
			    SCM scm_attr,
			    SCM scm_attr_e)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *c0 = (Container *) gh_scm2ulong (scm_c0);
  int i0 = gh_scm2int (scm_i0);
  Container *c1 = (Container *) gh_scm2ulong (scm_c1);
  int i1 = gh_scm2int (scm_i1);
  Attributes *a_s = (Attributes *) gh_scm2ulong (scm_attr_s);
  Attributes *a = (Attributes *) gh_scm2ulong (scm_attr);
  Attributes *a_e = (Attributes *) gh_scm2ulong (scm_attr_e);

  xw->highlight (c0, i0, c1, i1, a_s, a, a_e);

  return SCM_BOOL_T;
}



/* this is broken... shouldn't assume mark. */

void
XWord_Window::unhighlight ()
{
  Container *c;

  for (c=mark.con->next_node ();
       (c != NULL) && (c->tmpAttributes != NULL);
       c = c->next_node ())
    c->tmpAttributes = NULL;


  for (c=mark.con->prev_node ();
       (c != NULL) && (c->tmpAttributes != NULL);
       c = c->prev_node ())
    c->tmpAttributes = NULL;

  mark.con->tmpAttributes = NULL;

  //selection_cur = NULL;
}


SCM xword_window_unhighlight (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);

  xw->unhighlight ();

  return SCM_BOOL_T;
}



void
XWord_Window::Size_Notify (int set_width, int set_height)
{
  width = set_width;
  height = set_height;
}


SCM xword_window_size_notify (SCM scm_xw, SCM scm_width, SCM scm_height)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Dimension width = (Dimension) gh_scm2ulong (scm_width);
  Dimension height = (Dimension) gh_scm2ulong (scm_height);

  xw->Size_Notify (width, height);

  return SCM_BOOL_T;
}


int
XWord_Window::Document_Width (void)
{
  return (int) (doc->size_left () + doc->size_right ());
}


int
XWord_Window::Document_Height (void)
{
  return (int) (doc->size_up () + doc->size_down ());
}


int
XWord_Window::Document_Up (void)
{
  return (int)doc->size_up ();
}


int
XWord_Window::Document_Left (void)
{
  return (int)doc->size_left ();
}


int
XWord_Window::Document_Right (void)
{
  return (int)doc->size_right ();
}


int
XWord_Window::Document_Down (void)
{
  return (int)doc->size_down ();
}


void
XWord_Window::Redraw (void)
{
  doc->X_Erase (xd, w, gc, width, height);
  ((Document *)doc)->set_xy (scroll_x, scroll_y);
  doc->X_Draw (xd, w, gc, width, height);
  Draw_Cursor ();
  XFlush (xd->display);
}


SCM xword_window_redraw (SCM scm_xw, SCM scm_top)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *top = (Container *) gh_scm2ulong (scm_top);

  xw->Redraw (top);

  return SCM_BOOL_T;
}


void
XWord_Window::Redraw (Container *top)
{
  top->X_Erase (xd, w, gc, width, height);
  ((Document *)doc)->set_xy (scroll_x, scroll_y);
  top->X_Draw (xd, w, gc, width, height);
  Draw_Cursor ();
  XFlush (xd->display);
}


int
XWord_Window::Scroll_To_Cursor (void)
{
  int cursor_x, cursor_y, cursor_h;
  Size sz = cur.con->size ();
  int desc = sz.down;

  cur.con->offset_to_xy (cur.offset, &cursor_x, &cursor_y, &cursor_h);

  if (cursor_y - cursor_h <= 0)
    {
      scroll_y -= cursor_y - cursor_h; /* -1 ? */
      set_vertical_scrollbar ();
      return 1;
    }

  if (cursor_y+desc >= height)
    {
      scroll_y -= cursor_y + desc - height;
      scroll_y -= 2;
      set_vertical_scrollbar ();
      return 1;
    }

  return 0;
}


SCM xword_window_scroll_to_cursor (SCM scm_xw)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  if (xw->Scroll_To_Cursor ())
    return SCM_BOOL_T;
  return SCM_BOOL_F;
}


void
XWord_Window::set_vertical_scrollbar (void)
{
  int scroll_bar_pos;
  int max_pos;
  int slider_size;

  xw_assert (scroll_y <= 0);


  XtVaGetValues (vsb,
		 XmNvalue, &scroll_bar_pos,
		 XmNmaximum, &max_pos,
		 XmNsliderSize, &slider_size,
		 NULL);


  if (Document_Height () == height)
    return;  /* RUN AWAY! */

  scroll_bar_pos =
    (int)((double)((slider_size - max_pos) * scroll_y) /
	  (double)(Document_Height () - height));


  XtVaSetValues (vsb,
		 XmNvalue, scroll_bar_pos,
		 NULL);
}


void
XWord_Window::Draw_Cursor (void)
{
  int cursor_x, cursor_y, cursor_h;

  cur.con->offset_to_xy (cur.offset, &cursor_x, &cursor_y, &cursor_h);
  XDrawLine (xd->display, w, gc,
	     (int) cursor_x,
	     (int) cursor_y,
	     (int) cursor_x,
	     (int) (cursor_y - cursor_h));

  /***/
  /*
  mark.con->offset_to_xy (mark.offset, &cursor_x, &cursor_y, &cursor_h);
  XDrawLine (xd->display, w, gc,
	     (int) cursor_x,
	     (int) cursor_y,
	     (int) cursor_x,
	     (int) (cursor_y - cursor_h));
	     */
}



void
XWord_Window::ReWrap (Pos **positions)
{
  Container *flc, *llc;
  Pos *no_positions[ 1 ];

  if (positions == NULL)
    {
      no_positions[ 0 ] = NULL;
      positions = no_positions;
    }

  flc = doc->first_lowest_container ();
  llc = doc->last_lowest_container ();
  flc->Wrap_Layer (xd, flc, llc, positions, NULL);
  flc->Correct_Positions (positions);
}



SCM xword_window_rewrap (SCM scm_xw, SCM scm_first, SCM scm_last)
{
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);
  Container *first = (Container *) gh_scm2ulong (scm_first);
  Container *last = (Container *) gh_scm2ulong (scm_last);

  Pos *positions[ 3 ];
  positions[ 0 ] = &(xw->cur);
  positions[ 1 ] = &(xw->mark);
  positions[ 2 ] = NULL;
  Container *fee;

  first->Wrap_Layer (xw->xd, first, last, positions, &fee);
  first->Correct_Positions (positions);

  return gh_ulong2scm ((unsigned long) fee);
}


void hook_up_guile_XWord_Window ()
{
  gh_new_procedure5_0 ("xword-window.new",
		       (ghp) xword_window_new);
  gh_new_procedure2_0 ("xword-window.set-doc",
		       (ghp) xword_window_set_doc);
  gh_new_procedure4_0 ("xword-window.set-attributes",
		       (ghp) xword_window_set_attributes);
  gh_new_procedure3_0 ("xword-window.set-paste-attributes",
		       (ghp) xword_window_set_paste_attributes);
  gh_new_procedure1_0 ("xword-window.get-cursor-position",
		       (ghp) xword_window_get_cursor_position);
  gh_new_procedure3_0 ("xword-window.set-cursor-position",
		       (ghp) xword_window_set_cursor_position);
  gh_new_procedure1_0 ("xword-window.get-mark-position",
		       (ghp) xword_window_get_mark_position);
  gh_new_procedure3_0 ("xword-window.set-mark-position",
		       (ghp) xword_window_set_mark_position);
  gh_new_procedure3_0 ("xword-window.insert-xk",
		       (ghp) xword_window_insert_xk);
  gh_new_procedure3_0 ("xword-window.insert-tag",
		       (ghp) xword_window_insert_tag);
  gh_new_procedure3_0 ("xword-window.insert-character",
		       (ghp) xword_window_insert_character);
  gh_new_procedure1_0 ("xword-window.delete-character",
		       (ghp) xword_window_delete_character);
  gh_new_procedure1_0 ("xword-window.scroll-to-cursor",
		       (ghp) xword_window_scroll_to_cursor);
  gh_new_procedure2_0 ("xword-window.redraw",
		       (ghp) xword_window_redraw);
  gh_new_procedure3_0 ("xword-window.rewrap",
		       (ghp) xword_window_rewrap);
  gh_new_procedure3_0 ("xword-window.set-cursor-xy",
		       (ghp) xword_window_set_cursor_xy);
  gh_new_procedure1_0 ("xword-window.get-cursor-xy",
		       (ghp) xword_window_get_cursor_xy);
  gh_new_procedure3_0 ("xword-window.set-mark-xy",
		       (ghp) xword_window_set_mark_xy);
  gh_new_procedure1_0 ("xword-window.get-mark-xy",
		       (ghp) xword_window_get_mark_xy);
  gh_new_procedure5_0 ("xword-window.copy-to-primary",
		       (ghp) xword_window_copy_to_primary);
  gh_new_procedure ("xword-window.highlight", (ghp) xword_window_highlight,
		    8, 0, 0);
  gh_new_procedure1_0 ("xword-window.unhighlight",
		       (ghp) xword_window_unhighlight);
  gh_new_procedure3_0 ("xword-window.size-notify",
		       (ghp) xword_window_size_notify);
  gh_new_procedure5_0 ("xword-window.copy-to-cut-slot",
		       (ghp) xword_window_copy_to_cut_slot);
  gh_new_procedure1_0 ("xword-window.rotate-used-cut-slots",
		       (ghp) xword_window_rotate_used_cut_slots);
  gh_new_procedure1_0 ("xword-window.get-primary",
		       (ghp) xword_window_get_primary);
  gh_new_procedure1_0 ("xword-window.get-primary-targets",
		       (ghp) xword_window_get_primary_targets);
}

