; Copyright (C) 1997 The Hungry Programmers - hungry@hungry.com
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


; i just can't seem to make friends with motif's file-selection widget.
; also, lesstif's version still has some sizing issues to work out.


(define path-separator #\/)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; main entry point
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (file-selection-box title path ok-cb ok-args)
  (let* ((fsb (create-file-selection-box title path ok-cb ok-args)))
    (fsb-refresh fsb)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; convert a file system path from ("a" "b" "c") to "/a/b/c/"
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (path-list-to-string path-list)
  (let ((path-string ""))
    (for-each (lambda (x)
		(set! path-string (string-append path-string
						 (string path-separator)))
		(set! path-string (string-append path-string x)))
	      path-list)
    ; if the string is empty, return "/"
    (if (eqv? (length path-string) 0)
	(string path-separator)
	path-string)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; get the current working directory
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;help!

(define (pwd)
  ;(getenv "PWD") ; why does this fail?

  ; this is so pretty!
  (let* ((p (pipe))
	 (stdout (car (fdes->ports 1))))
    (if (eqv? 0 (primitive-fork))
	(begin
	  (close-port (car p))
	  (redirect-port (cdr p) stdout)
	  (execlp "pwd" "pwd")))
    (close-port (cdr p))
    (let ((dir (read-line (car p))))
      (close-port (car p))
      dir)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; convert a file system path from "/a/b/c/" to ("a" "b" "c")
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (path-string-to-list path-string)

  ; kill trailing slashes
  (if (> (length path-string) 0)
      (do () ((or
	       (eqv? (length path-string) 0)
	       (not (char=?
		     path-separator
		     (string-ref path-string (- (length path-string) 1)))))
	      #t)
	(set! path-string (substring
			   path-string
			   0
			   (- (length path-string) 1)))))

  ; break the string into parts
  (do ((i 0 (+ i 1))
       ;(path-list (list ""))
       (path-list (list))
       ; if path-string is relative, tack $PWD to beginning
       (path-string
	(if (eqv? (length path-string) 0) ""
	    (if (char=? (string-ref path-string 0) path-separator)
		path-string
		(string-append (pwd)
			       (string path-separator)
			       path-string)))
	; path-string gets 1 char shorter each loop interation
	(substring path-string 1 (length path-string))))
      ; exit clause
      ((eqv? 0 (length path-string)) path-list)
    ; loop body
    (let ((first-char (string-ref path-string 0))
	  (path-last (- (length path-list) 1)))
      (if (char=? path-separator first-char)
	  ; first char is a /
	  (if (and (>= path-last 0)
		   (> (length (list-ref path-list path-last)) 0))
	      (set! path-list (list-append path-list (list ""))))
	  ; first char is not a /
	  (if (>= path-last 0)
	      (list-set! path-list path-last
			 (string-append (list-ref path-list path-last)
					(string first-char)))
	      (set! path-list (list (string first-char))))))))
  



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; clean up the text of a path string
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (clean-path-string path-string)
  (path-list-to-string (path-string-to-list path-string)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; tests for the code defined above
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (test-path-subrs)

  (display (path-list-to-string (list "ack" "bat" "cat")))
  (newline)
  (display (path-list-to-string (list)))
  (newline)
  (display "----------------------------")
  (newline)

  (display (path-string-to-list "/dog//egg/foo/guts///"))
  (newline)
  (display (path-string-to-list "/dog//egg/foo/guts"))
  (newline)
  (display (path-string-to-list "dog//egg/foo/guts"))
  (newline)
  (display (path-string-to-list ""))
  (newline)
  (display (path-string-to-list "dog//egg/foo/guts//"))
  (newline)
  (display (path-string-to-list "/"))
  (newline)
  (display (path-string-to-list "////"))
  (newline)

  (display "----------------------------")
  (newline)

  (display (clean-path-string "dog//egg/foo/guts//"))
  (newline)
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; put together a file selection box
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (create-file-selection-box title path ok-cb ok-args)
  (let* ((fsb-shell (xt-app-create-shell xdisplay title))
	 (fsb-form (xt-create-widget "fsb form"
				     (xm-form-widget-class) fsb-shell #f))
	 (fsb-ok (xt-create-widget
		  "OK"
		  (xm-push-button-widget-class)
		  fsb-form
		  (list (list (xm-n-left-attachment) (xm-attach-form))
			(list (xm-n-top-attachment) (xm-attach-none))
			(list (xm-n-right-attachment) (xm-attach-none))
			(list (xm-n-bottom-attachment) (xm-attach-form)))))

	 (fsb-cancel (xt-create-widget
	       "Cancel"
	       (xm-push-button-widget-class)
	       fsb-form
	       (list (list (xm-n-left-attachment) (xm-attach-widget))
		     (list (xm-n-left-widget) fsb-ok)
		     (list (xm-n-top-attachment) (xm-attach-none))
		     (list (xm-n-right-attachment) (xm-attach-none))
		     (list (xm-n-bottom-attachment) (xm-attach-form)))))


	 (fsb-tf (xt-create-widget "fsb tf"
				   (xm-text-field-widget-class)
				   fsb-form
				   (list (list (xm-n-left-attachment)
					       (xm-attach-form))
					 (list (xm-n-top-attachment)
					       (xm-attach-none))
					 (list (xm-n-right-attachment)
					       (xm-attach-form))
					 (list (xm-n-columns)
					       60)
					 (list (xm-n-bottom-attachment)
					       (xm-attach-widget))
					 (list (xm-n-bottom-widget)
					       fsb-ok))))


	 (fsb-list (xm-create-scrolled-list
		    fsb-form "fsb list"
		    (list (list (xm-n-left-attachment) (xm-attach-form))
			  (list (xm-n-top-attachment) (xm-attach-form))
			  (list (xm-n-right-attachment) (xm-attach-form))
			  (list (xm-n-bottom-attachment) (xm-attach-widget))
			  (list (xm-n-bottom-widget) fsb-tf)
			  (list (xm-n-visible-item-count) 20)))))


    (xt-add-callback fsb-list
		     (xm-n-browse-selection-callback)
		     fsb-list-callback
		     (list fsb-shell))

    (xt-add-callback fsb-ok
		     (xm-n-activate-callback)
		     ok-cb ok-args)

    (xt-add-callback fsb-cancel
		     (xm-n-activate-callback)
		     fsb-cancel-callback
		     (list fsb-shell))

    (xt-manage-child fsb-list)
    (xt-manage-child fsb-tf)
    (xt-manage-child fsb-ok)
    (xt-manage-child fsb-cancel)
    (xt-manage-child fsb-form)
    (xt-realize-widget fsb-shell)

    (xt-popup fsb-shell (xt-grab-none))


    (xm-text-set-string fsb-tf path)

    fsb-shell)) ; return the shell


;access? path how 
;   Variable: R_OK test for read permission. 
;   Variable: W_OK test for write permission. 
;   Variable: X_OK test for execute permission. 
;   Variable: F_OK test for existence of the file. 


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; refresh the file selection box
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (fsb-file-is-interesting file)
  #t)
; hmmm
;  (cond ((has-suffix? file ".xword") #t)
;	((has-suffix? file ".gif") #t)
;	((has-suffix? file ".pgm") #t)
;	((has-suffix? file ".ppm") #t)
;	((has-suffix? file ".txt") #t)
;	((has-suffix? file ".txt") #t)))


(define (fsb-refresh fsb)
  (let* ((fsb-tf (xt-name-to-widget fsb "*fsb tf"))
	 (fsb-list (xt-name-to-widget fsb "*fsb list"))
	 (fsb-form (xt-name-to-widget fsb "*fsb form"))
	 (pf-string (xm-text-get-string fsb-tf))
	 (pf-list (path-string-to-list pf-string))
	 (path-and-file (get-path-and-file pf-list))
	 (path-list (car path-and-file))
	 (path-string (path-list-to-string path-list)))
    
    (if (access? path-string (+ R_OK X_OK))
	(let ((dir (opendir path-string)))
	  (xm-list-delete-all-items fsb-list)
      
	  (xt-unmanage-child fsb-list)
	  (xt-unmanage-child fsb-form)
      
	  (do ((dir-entry (readdir dir) (readdir dir))
	       (entries (list)))
	      ;
	      ((eof-object? dir-entry)
	       (xm-list-add-entries fsb-list 
				    (sort entries string-comparison) 0))
	    ;
	    (if (fsb-is-file? (list-append path-list (list dir-entry)))
		; a file
		(if (fsb-file-is-interesting dir-entry)
		    (set! entries (list-append entries (list dir-entry))))
		; directory
		(set! entries
		      (list-append entries
				   (list (string-append dir-entry "/"))))))
      
	  (xt-manage-child fsb-list)
	  (xt-manage-child fsb-form)
	  (closedir dir))
	(begin
	  (xm-list-delete-all-items fsb-list)
	  (xm-list-add-entry fsb-list "./" 0)
	  (xm-list-add-entry fsb-list "../" 0)
	  (display "Can't list directory.")
	  (newline)))))


;;;;;;;;;;
;;
;;;;;;;;;;

(define (fsb-is-file? path-list)
  (if (null? path-list)
      #f
      (let* ((path-last (car (reverse path-list)))
	     (path-string (path-list-to-string path-list)))
	(if (or (has-suffix? path-last ".xword")
		(not (file-exists? path-string)))
	    #t
	    (if (file-is-directory? path-string)
		#f
		#t)))))
	    

(define (get-path-and-file path-list)
  (if (fsb-is-file? path-list)
      (list (reverse (cdr (reverse path-list)))
	    (car (reverse path-list)))
      (list path-list #f)))


(define (fsb-up-dir fsb)
  (let* ((fsb-tf (xt-name-to-widget fsb "*fsb tf"))
	 (path-string (xm-text-get-string fsb-tf))
	 (path-list (path-string-to-list path-string))
	 (path-and-file (get-path-and-file path-list))
	 (up-one-list (car path-and-file)))

    (if (not (null? up-one-list))
	(set! up-one-list (reverse (cdr (reverse up-one-list)))))

    (if (cadr path-and-file)
	(set! up-one-list
	      (list-append up-one-list (list (cadr path-and-file)))))

    (xm-text-set-string fsb-tf (path-list-to-string up-one-list))))



(define (fsb-down-dir fsb dir)

  (let* ((fsb-tf (xt-name-to-widget fsb "*fsb tf"))
	 (path-string (xm-text-get-string fsb-tf))
	 (path-list (path-string-to-list path-string))
	 (path-and-file (get-path-and-file path-list))
	 (down-one-list (reverse (cons dir (reverse (car path-and-file))))))

    ; if there was a file selected before, and the directory
    ; has changed, tack the file back onto the end of the path
    (if (and (eqv? #f (cadr (get-path-and-file down-one-list)))
	     (not (eqv? #f (cadr path-and-file))))
	(set! down-one-list
	      (list-append down-one-list (list (cadr path-and-file)))))

    (xm-text-set-string fsb-tf (path-list-to-string down-one-list))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; the ok callback needs a way to get the selected file name
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (fsb-get-file-name widget)
  (let* ((fsb-tf (xt-name-to-widget (xt-parent widget) "*fsb tf")))
    (xm-text-get-string fsb-tf)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; file selection box callbacks
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (fsb-list-callback widget client-data call-data)

  (let* ((sp (xt-get-value widget (xm-n-selected-items)))
	 (s (dereference sp))
	 (str (xm-string-get-l-to-r s (xm-font-list-default-tag)))
	 (parent (xt-parent (xt-parent (xt-parent widget))))
	 (fsb-tf (xt-name-to-widget parent "*fsb tf")))

    (if str ; xm-string-get-l-to-r can sometimes return #f
	(begin
	  ; trim any trailing slash
	  (if (char=? path-separator (string-ref str (- (length str) 1)))
	      (set! str (substring str 0 (- (length str) 1))))

	  ; check for .. or .
	  (if (string=? str "..")
	      (fsb-up-dir parent)
	      (if (not (string=? str "."))
		  (fsb-down-dir parent str)))

	  ; need to xt-free str?  FINISH ME
	  (fsb-refresh parent)))))


(define (fsb-cancel-callback widget client-data call-data)
  (xt-destroy-widget (xt-parent (xt-parent widget))))
