/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>


typedef struct _node
{
  char *str;
  struct _node **children;
  struct _node *parent;
  int num_children;
  int max_children;
} node;



node *new_node (void)
{
  node *n;
  n = (node *) malloc (sizeof (node));
  n->str = NULL;
  n->parent = NULL;
  n->num_children = 0;
  n->max_children = 2;


  n->children = (node **) malloc (sizeof (node *) * n->max_children);
  if (n->children == NULL)
    {
      perror ("malloc of children");
      exit (1);
    }

  return n;
}


node *lookup_or_add (node *p, char *str)
{
  int i;

  for (i=0; i<p->num_children; i++)
    {
      if (strcasecmp (p->children[i]->str, str) == 0)
	{
	  return p->children[i];
	}
    }

  if (p->num_children == p->max_children)
    {
      p->max_children ++;
      p->max_children *= 2;
      p->children = (node **) realloc ((void *) p->children,
				      sizeof (node *) * p->max_children);
      if (p->children == NULL)
	{
	  perror ("realloc of children");
	  exit (1);
	}
    }

  p->children[ p->num_children ] = new_node ();
  p->children[ p->num_children ]->str = strdup (str);
  p->children[ p->num_children ]->parent = p;
  p->num_children ++;

  return p->children[ p->num_children - 1 ];
}


void break_down_str (node *n, char *str, int depth)
{
  char *s = str + 1;
  char *e;
  char seg[ 400 ];
  int len;
  node *child;

  /*  printf ("breaking down '%s'\n", str);*/

  if (str[0] != '-')
      return; /* done. */

  if (*s == '\0')
    return; /* reached the end? */

  e = strchr (s, '-');

  if (e == NULL)
    {
      len = strlen (s);
      e = "";
    }
  else
    len = e - s;


  switch (depth)
    {
    case 6:
      strcpy (seg, "%d");
      break;
    case 7:
    case 8:
    case 9:
    case 11:
      strcpy (seg, "*");
      break;
    default:
      strncpy (seg, s, len);
      seg[ len ] = '\0';
      break;
    }

  child = lookup_or_add (n, seg);

  break_down_str (child, e, depth + 1);
}


void dump_from_leaf (node *n)
{
  if (n->parent != NULL)
    dump_from_leaf (n->parent);

  if (n->str != NULL)
    printf ("-%s", n->str);
}


void print_nice_name (node *n)
{
  int i;
  char *spacing;
  char *slant;
  char *weight;
  char *family;

  for (i=0; i<9; i++)
    n = n->parent;

  spacing = n->str;
  slant = n->parent->str;
  weight = n->parent->parent->str;
  family = n->parent->parent->parent->str;

  family[0] = toupper(family[0]);

  for (i=0; i<strlen (family); i++)
    {
      if (family[i] == ' ')
	family[i] = '_';
    }


  printf ("%s", family);

  if (strcmp (weight, "bold") == 0)
    printf ("-Bold");
  if (strcmp (slant, "o") == 0)
    printf ("-Oblique");
  if (strcmp (slant, "i") == 0)
    printf ("-Italic");
}


void dump (node *n, int depth)
{
  int i;

  if (n->num_children == 0)
    {
      print_nice_name (n);
      printf (" 1 ");
      dump_from_leaf (n);
      printf ("\n");
    }

  for (i=0; i<n->num_children; i++)
    {
      dump (n->children[i], depth+1);
    }
}


int main (int argc, char *argv[])
{
  Widget topLevel;
  XtAppContext app_context;
  int n;
  char **f;
  int lop;
  node *root;

  root = new_node ();

  topLevel = XtVaAppInitialize (&app_context,
				"Generate_Font_Config",
				NULL, 0,
				&argc, argv,
				NULL, NULL);


  f = XListFonts (XtDisplay (topLevel),
		  "-adobe-*-*-*-*--*-*-*-*-*-*-*-*",
		  1000,
		  &n);

  for (lop=0; lop<n; lop++)
    {
      /* printf ("    %s\n", f[lop]);*/
      break_down_str (root, f[lop], 0);
    }


  dump (root, 0);

  return 0;
}
