/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include "utils.h"
#include "tag_info.h"

static int debug = 0;


tag_info::tag_info ()
{
  tag = XtMalloc (1);
  tag[ 0 ] = '\0';
  num_pairs = 0;
  pair_names = NULL;
  pair_values = NULL;
  pix = 0;
}


tag_info::tag_info (char *txt)
{
  tag = NULL;
  pix = 0;
  parse (txt, strlen (txt));
}


tag_info::tag_info (char *txt, int len)
{
  tag = NULL;
  pix = 0;
  parse (txt, len);
}


void tag_info::parse (char *txt, int len)
{
  char *p = txt;
  char *e;

  if (debug)
    {
      int i;
      printf ("tag_info parsing: '");
      for (i=0; i<len; i++)
	printf ("%c", txt[ i ]);
      printf ("'\n");
    }


  if (tag != NULL)
    XtFree (tag);

  tag = (char *) XtMalloc (1);
  tag[ 0 ] ='\0';
  num_pairs = 0;
  pair_names = NULL;
  pair_values = NULL;

  p = skip_ws (p, txt + len);

  if (p == txt + len)
    {
      if (debug)
	printf ("    incomplete tag #0\n");
      return;
    }

  if (*p != '<')
    {
      if (debug)
	printf ("    didn't get '<' #1\n");
      return;
    }

  p++;

  if (p == txt + len)
    {
      if (debug)
	printf ("    incomplete tag #2\n");
      return;
    }

  p = skip_ws (p, txt + len);
  e = skip_to_char_or_ws (p, '>', txt + len);

  if (p == e)
    {
      if (debug)
	printf ("    empty tag #3\n");
      return;
    }

  tag = (char *) XtRealloc (tag, e - p + 1);
  strncpy (tag, p, e - p);
  tag[ e-p ] = '\0';

  if (debug)
    printf ("    tag = '%s'\n", tag);

  p = skip_ws (e, txt + len);

  for (;;)
    {
      if (p >= txt + len)
	{
	  if (debug)
	    printf ("    incomplete tag #4\n");
	  return;
	}

      p = skip_ws (p, txt + len);

      if (p >= txt + len)
	{
	  if (debug)
	    printf ("    incomplete tag #4.5\n");
	  return;
	}

      if (*p == '>')
	{
	  if (debug)
	    printf ("    end of tag #5\n");
	  return;
	}

      pair_names = (char **) XtRealloc ((char *) pair_names,
					sizeof (char *) * num_pairs + 1);
      pair_values = (char **) XtRealloc ((char *) pair_values,
					 sizeof (char *) * num_pairs + 1);

      e = skip_to_char2_or_ws (p, '=', '>', txt + len);
      if (p == e)
	{
	  if (debug)
	    printf ("    incomplete tag #6\n");
	  return;
	}

      pair_names[ num_pairs ] = (char *) XtMalloc (e - p + 1);
      strncpy (pair_names[ num_pairs ], p, e - p);
      pair_names[ num_pairs ][ e-p ] = '\0';


      pair_values[ num_pairs ] = (char *) XtMalloc (1);
      pair_values[ num_pairs ][ 0 ] = '\0';

      p = skip_ws (e, txt + len);
      if (p == txt + len)
	{
	  if (debug)
	    printf ("    incomplete tag #7\n");
	  num_pairs ++;
	  return;
	}

      if (*p == '=')
	{
	  p ++;
	  p = skip_ws (p, txt + len);
	  if (p == txt + len)
	    {
	      if (debug)
		printf ("    incomplete tag #8\n");
	      num_pairs ++;
	      return;
	    }
	  e = skip_to_char_or_ws (p, '>', txt + len);

	  if (p == e)
	    {
	      if (debug)
		printf ("    incomplete tag #9\n");
	      num_pairs ++;
	      return;
	    }

	  pair_values[ num_pairs ] =
	    (char *) XtRealloc (pair_values[ num_pairs], e - p + 1);
	  strncpy (pair_values[ num_pairs ], p, e - p);
	  pair_values[ num_pairs ][ e-p ] = '\0';
	  wack_quotes (pair_values[ num_pairs ]);
	  p = e;
	}
      else
	{
	  pair_values[ num_pairs ] = (char *) XtMalloc (1);
	  pair_values[ num_pairs ][ 0 ] = '\0';
	}

      num_pairs ++;
    }
}


tag_info::~tag_info ()
{
}


void tag_info::wack_quotes (char *str)
{
  int len = strlen (str);
  int i;

  if (len == 0) return;
  if (str[ 0 ] != '"') return;

  for (i=1; i<len-1; i++)
    str[ i-1 ] = str[ i ];

  str[ i-1 ] = '\0';
}


int tag_info::size ()
{
  int i, len = 1;
  len += strlen (tag);

  for (i=0; i<num_pairs; i++)
    {
      len += 1;
      len += strlen (pair_names[ i ]);
      if (strlen (pair_values[ i ]) > 0)
	{
	  len += 3;
	  len += strlen (pair_values[ i ]);
	}
    }

  len += 1;

  return len;
}


void tag_info::sprint (char *p)
{
  int i;
  strcpy (p, "<");
  strcat (p, tag);

  for (i=0; i<num_pairs; i++)
    {
      strcat (p, " ");
      strcat (p, pair_names[ i ]);
      if (strlen (pair_values[ i ]) > 0)
	{
	  strcat (p, "=\"");
	  strcat (p, pair_values[ i ]);
	  strcat (p, "\"");
	}
    }

  strcat (p, ">");
}



void tag_info::print (FILE *fp)
{
  int i;
  fprintf (fp, "<%s", tag);

  for (i=0; i<num_pairs; i++)
    {
      fprintf (fp, " %s", pair_names[ i ]);
      if (strlen (pair_values[ i ]) > 0)
	fprintf (fp, "=\"%s\"", pair_values[ i ]);
    }

  fprintf (fp, ">");
}


void tag_info::set_tag (char *set_tag)
{
  if (tag != NULL) XtFree (tag);
  tag = (char *) XtMalloc (strlen (set_tag) + 1);
  strcpy (tag, set_tag);
}


void tag_info::set_pair (char *name, char *value)
{
  int i;
  for (i=0; i<num_pairs; i++)
    if (strcmp (pair_names[ i ], name) == 0)
      {
	XtFree (pair_values[ i ]);
	pair_values[ i ] = (char *) XtMalloc (strlen (value) + 1);
	strcpy (pair_values[ i ], value);
	return;
      }

  add_pair (name, value);
}


void tag_info::add_pair (char *name, char *value)
{
  pair_names = (char **) XtRealloc ((char *) pair_names,
				    sizeof (char *) * num_pairs + 1);
  pair_values = (char **) XtRealloc ((char *) pair_values,
				     sizeof (char *) * num_pairs + 1);

  pair_names[ num_pairs ] = (char *) XtMalloc (strlen (name) + 1);
  strcpy (pair_names[ num_pairs ], name);

  pair_values[ num_pairs ] = (char *) XtMalloc (strlen (value) + 1);
  strcpy (pair_values[ num_pairs ], value);

  num_pairs ++;
}


tag_info *tag_info::dup ()
{
  int i;
  tag_info *ti = new tag_info ();
  ti->set_tag (tag);

  for (i=0; i<num_pairs; i++)
    ti->add_pair (pair_names[ i ], pair_values[ i ]);

  ti->pix = pix;

  return ti;
}


Attr_Tag tag_info::to_number ()
{
  if (strcmp (tag, "br") == 0)
    return TAG_br;
  else if (strcmp (tag, "hr") == 0)
    return TAG_hr;
  else if (strcmp (tag, "img") == 0)
    return TAG_img;
  else
    return TAG_unknown;
}


char *tag_info::get_value (char *name)
{
  int i;
  for (i=0; i<num_pairs; i++)
    if (strcmp (pair_names[ i ], name) == 0)
      return pair_values[ i ];

  return NULL;
}


void tag_info::delete_pair (char *name)
{
  int i;
  for (i=0; i<num_pairs; i++)
    if (strcmp (pair_names[ i ], name) == 0)
      {
	int j;
	free (pair_names[ i ]);
	free (pair_values[ i ]);
	for (j=i+1; j<num_pairs; j++)
	  {
	    pair_names[ j-1 ] = pair_names[ j ];
	    pair_values[ j-1 ] = pair_values[ j ];
	  }
	num_pairs --;
      }
}


/**********************************************************************/

void test_tag_info ()
{
  tag_info *ti;
  unsigned int i;
  char *foo[] = 
  {
    "<ack a=blee b=dee>",
    " <ack a=blee b=dee>",
    "< ack a=blee b=dee>",
    "<ack  a=blee b=dee>",
    "<ack a  =blee b=dee>",
    "<ack a=  blee b=dee>",
    "<ack a=>",
    "<ack a=",
    "<ack a=blee b>",
    "<ack a=blee b=",
    "<ack a=blee b = ",
    "<ack a=blee b = >",
    "<ack a=blee   b=dee>",
    "<ack a=blee dee>",
    "<ack a=blee b =dee",
    "<ack a=blee b = dee> ",
    "<br a=foo>",
    " <br a=foo>",
    "< br a=foo>",
    "<br  a=foo>",
    "<br a =foo>",
    "<br a= foo>",
    "<br a=foo >",
    "<br a=foo",
    "<br a=foo >",
    "<br a=foo ",
    "<br a = foo ",
    "< br a = foo ",
    " < br a = foo ",
    " < br a = foo >",
    " < br a = foo>",
    "<br>",
    " <br>",
    "< br>",
    "<br >",
    " br > ",
    "br > ",
    "<br",
    "<br ",
    0
  };


  for (i=0; foo[i] != NULL; i++)
    {
      printf ("%d: ", i); fflush (stdout);
      ti = new tag_info (foo[ i ]); fflush (stdout);
      printf ("%s --> ", foo[ i ]); fflush (stdout);
      ti->print (stdout); fflush (stdout);
      printf ("\n"); fflush (stdout);
    }
}
