/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include "utils.h"
#include <X11/Intrinsic.h>


XChar2b xk_to_2b (KeySym xk)
{
  XChar2b c;
  c.byte2 = (xk & 0xFF);
  c.byte1 = (xk & 0xFF00) >> 8;

  return c;
}

KeySym xc2b_to_xk (XChar2b c)
{
  return (KeySym)((c.byte1 << 8) | c.byte2);
}


void size_scratch (char **scratch, int *max, int needed)
{
  if (needed >= *max)
    {
      *max = needed + 2;
      (*scratch) = (char *) XtRealloc ((char *) (*scratch), *max);
    }
}


char *skip_ws (char *p, char *end)
{
  while (p < end)
    {
      if (*p != ' ' &&
	  *p != '\t' &&
	  *p != '\n' &&
	  *p != '\r')
	break;

      p++;
    }

  return p;
}


char *skip_to_ws (char *p, char *end)
{
  while (p < end)
    {
      if (*p == ' ' ||
	  *p == '\t' ||
	  *p == '\n' ||
	  *p == '\r')
	break;

      p++;
    }

  return p;
}


char *skip_to_char_or_ws (char *p, char c, char *end)
{
  while (p < end)
    {
      if (*p == ' ' ||
	  *p == '\t' ||
	  *p == '\n' ||
	  *p == '\r' ||
	  *p == c)
	break;

      p++;
    }

  return p;
}


char *skip_to_char2_or_ws (char *p, char c1, char c2, char *end)
{
  while (p < end)
    {
      if (*p == ' ' ||
	  *p == '\t' ||
	  *p == '\n' ||
	  *p == '\r' ||
	  *p == c1 ||
	  *p == c2)
	break;

      p++;
    }

  return p;
}



void parse_stream (char *buf,
		   int *read_point,
		   int *write_point,
		   /**/
		   int *start,
		   int *end,
		   Container_Type *ct,
		   int bts)
{
  char *rp = buf + (*read_point);
  char *wp = buf + (*write_point);

  if (rp == wp)
    {
      /* it has all been parsed. */
      (*ct) = T_unknown;
      (*read_point) = 0;
      (*write_point) = 0;
      return;
    }

  
  /* decide if this string of characters represents a Tag or a letter */
  
  if (*rp == '<')
    {
      int i;
      char *p;
      
      /* found the start of a tag */
      
      for (p=rp; p<wp; p += bts)
	{
	  if (p == wp)
	    break;
	  
	  if (*p == '>')
	    {
	      /* got a tag */
	      (*start) = (*read_point);
	      (*end) = (*start) + (p - rp) + bts;
	      (*ct) = T_Tag;
	      (*read_point) = (*end);
	      return;
	    }
	}
      
      /* the end of the tag hasn't arrived yet. */
      /* pack the unused stuff to the bottom of the scratch area */
      
      (*ct) = T_unknown;
      // worried about overlapping args...?
      //memmove ((void *) buf, (void *) read_point, write_point - read_point);
      for (i=0; i < (*write_point)-(*read_point); i++)
	buf[ i ] = buf[ i+(*read_point) ];
      
      (*write_point) -= (*read_point);
      (*read_point) = 0;
       
      return;
    }


  /* not a tag -- must be a character */

  (*ct) = T_WordSegment;
  (*start) = (*read_point);
  (*end) = (*start) + bts;
  (*read_point) += bts;
}


char *itoa_scratch = NULL;
int max_itoa_scratch;

char *itoa (int i)
{
  int len=0, j = i;

  while (j > 10)
    {
      len ++;
      j /= 10;
    }

  len += 3;

  size_scratch (&itoa_scratch, &max_itoa_scratch, len);

  sprintf (itoa_scratch, "%d", i);

  return itoa_scratch;
}


int Iabs (int i)
{
    if (i<0)
	return -i;
    return i;
}


int ceil (double d)
{
    if ((double)((int)d) == d)
	return (int)d;

    return ((int)d + 1);
}


int highbit (unsigned long ul)
{
  /* returns position of highest set bit in 'ul' as an integer (0-31),
   or -1 if none */

  int i;
  for (i=31; ( (ul&0x80000000) == 0) && i>=0;  i--, ul<<=1);
  return i;
}


int file_exists (char *fn)
{
  /* FINISH ME -- exists doesn't mean readable. */

  return file_readable (fn);
}


int file_readable (char *fn)
{
  FILE *fp;

  fp = fopen (fn, "r");
  if (fp == NULL)
    return 0;

  fclose (fp);

  return 1;
}
