/*
 Copyright (C) 1995-1997 The Hungry Programmers - hungry@hungry.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
  

#include "utils.h"

#include <Xm/Xm.h>
#include <Xm/ScrollBar.h>

#include "g.h"
#include "motif.h"

#include "Container.h"
#include "Document.h"
#include "Page.h"
#include "WordSegment.h"
#include "Tag.h"
#include "tag_info.h"

#include "VkFont.h"
#include "XCutBuffer.h"

int global_argc;
char **global_argv;

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include "XWord.h"
#include "XDisplay.h"
#include "XWord_Window.h"


int verbose;
static int debug;
static char **xword_path;

/**********/

char *find_the_file (char *filename)
{
  static char *ret=NULL;
  static int ret_len=0;
  int i;
  FILE *fp;

  for (i=0; xword_path[ i ] != NULL; i++)
    {
      int sz = strlen (xword_path[ i ]) + strlen (filename) + 2;

      if (ret_len < sz)
	{
	  ret = (char *) realloc (ret, sz);
	  ret_len = sz;
	}
	  
      sprintf (ret, "%s/%s", xword_path[ i ], filename);


      fp = fopen (ret, "r");
      if (fp != NULL)
	{
	  fclose (fp);
	  if (verbose)
	    printf ("using '%s'\n", ret);
	  return ret;
	}
    }

  fprintf (stderr, "Couldn't locate '%s' in path:\n", filename);

  {
    int i;
    for (i=0; xword_path[ i ] != NULL; i++)
      fprintf (stderr, "  %s\n", xword_path[ i ]);
  }

  return NULL;
}


SCM scm_find_the_file (SCM scm_file_name)
{
  char *file_name = gh_scm2string (scm_file_name);
  char *full = find_the_file (file_name);
  if (full == NULL)
    return SCM_BOOL_F;
  return gh_str02scm (full);
}



int vertical_scroll_bar_position (Widget w, XWord_Window *xw)
{
  double vert_percent;
  int scroll_bar_pos;
  int /* or Dimension??? */ max_pos;
  int slider_size;

  // scroll_bar_pos = ((XmScrollBarCallbackStruct *) callData)->value;

  XtVaGetValues(w,
		XmNvalue, &scroll_bar_pos,
		XmNmaximum, &max_pos,
		XmNsliderSize, &slider_size,
		NULL);


  vert_percent = (0.0 - ((float )scroll_bar_pos)) /
    ((float) (max_pos - slider_size));


  /*
  printf ("sb_pos:%d sb_max:%d sb_size:%d doc_h:%d win_h:%d "
	  "%%:%f scroll_y:%d\n",
	  scroll_bar_pos, max_pos, slider_size,
	  xw->Document_Height (),
	  xw->height,
	  vert_percent, xw->scroll_y);
	  */

  return (int) (((double) (xw->Document_Height () - xw->height)) *
		vert_percent);
}



int horizontal_scroll_bar_position (Widget w, XWord_Window *xw)
{
  float horiz_percent;
  int scroll_bar_pos;
  int max_pos;

  XtVaGetValues(w,
		XmNvalue, &scroll_bar_pos,
		XmNmaximum, &max_pos,
		NULL);


  horiz_percent = (0.0 - (float)scroll_bar_pos)/(float)max_pos;

  return (int) (xw->Document_Width () * horiz_percent + xw->Document_Left ());
}



SCM Vertical_Scroll_Event (SCM scm_w, SCM scm_xw)
{
  Widget w = (Widget) gh_scm2ulong (scm_w);
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);

  xw->scroll_y = vertical_scroll_bar_position (w, xw);
  xw->Redraw ();

  return SCM_BOOL_T;
}


SCM Horizontal_Scroll_Event (SCM scm_w, SCM scm_xw)
{
  Widget w = (Widget) gh_scm2ulong (scm_w);
  XWord_Window *xw = (XWord_Window *) gh_scm2ulong (scm_xw);

  xw->scroll_x = horizontal_scroll_bar_position (w, xw);
  xw->Redraw ();

  return SCM_BOOL_T;
}


int process_xword_args (int argc, char *argv [])
{
  int i;

  for  (i=1; i<argc; i++)
    {
      if (strcmp (argv[ i ], "--verbose") == 0)
	verbose = 1;
      if (strcmp (argv[ i ], "--debug") == 0)
	debug = 1;
      else
	break;
    }

  return i;
}



/***********************************************************************/

/***********************************************************************/

/*
 #define GUILE_BOOT_CMD "(primitive-load-path \"ice-9/boot-9.scm\")"
 #define GUILE_BOOT_CMD "#t"
*/

/* Debugger interface (don't change the order of the following lines) */
#define GDB_TYPE SCM

extern "C"
{
#include <libguile/gdb_interface.h>
}

GDB_INTERFACE;


void hook_up_guile ()
{
  hook_up_guile_motif ();
  hook_up_guile_XCutBuffer ();
  hook_up_guile_Attributes ();
  hook_up_guile_XDisplay ();
  hook_up_guile_FontManager ();
  hook_up_guile_VkFont ();
  hook_up_guile_XWord_Window ();
  hook_up_guile_Container ();
  hook_up_guile_Document ();
  hook_up_guile_Page ();
  hook_up_guile_WordSegment ();
  hook_up_guile_Tag ();


  gh_new_procedure2_0 ("handle-horizontal-scroll-event",
		       (ghp) Horizontal_Scroll_Event);
  gh_new_procedure2_0 ("handle-vertical-scroll-event",
		       (ghp) Vertical_Scroll_Event);

  gh_new_procedure1_0 ("find-the-file",
		       (ghp) scm_find_the_file);
}



void setup_xword_path ()
{
  char *xwp_env;
  xwp_env = getenv ("XWORD_PATH");
  if (xwp_env == NULL)
    {
      xword_path = (char **) malloc (sizeof (char *) * 25);
      xword_path[ 0 ] = "/usr/local/bin";
      xword_path[ 1 ] = "/usr/public/bin";
      xword_path[ 2 ] = "/usr/bin";
      xword_path[ 3 ] = "/opt/gnu/bin";
      xword_path[ 4 ] = "/usr/local/share/xword";
      xword_path[ 5 ] = "/usr/local/lib";
      xword_path[ 6 ] = "/usr/local/lib/xword";
      xword_path[ 7 ] = "/usr/local/share";
      xword_path[ 8 ] = "/usr/local/xword";
      xword_path[ 9 ] = "/usr/share/xword";
      xword_path[ 10 ] = "/usr/share";
      xword_path[ 11 ] = "/usr/public/lib/xword";
      xword_path[ 12 ] = "/usr/public/share/xword";
      xword_path[ 13 ] = "/usr/public/lib";
      xword_path[ 14 ] = "/usr/public/share";
      xword_path[ 15 ] = "/opt/gnu/lib";
      xword_path[ 16 ] = "/opt/gnu/share";
      xword_path[ 17 ] = "/usr/lib";
      xword_path[ 18 ] = "/usr/lib/xword";
      xword_path[ 19 ] = "/lib";
      xword_path[ 20 ] = ".";
      xword_path[ 21 ] = 0;
    }
  else
    {
      int nc=0, c=0, i, len = strlen (xwp_env);
      xwp_env = strdup (xwp_env);

      /* count :'s */
      for (i=0; i<len; i++)
	if (xwp_env[ i ] == ':') nc++;

      xword_path = (char **) malloc (sizeof (char *) * (nc + 2));

      xword_path[ c++ ] = xwp_env;
      for (i=0; i<len; i++)
	{
	  if (xwp_env[ i ] == ':')
	    {
	      xwp_env[ i ] = '\0';
	      xword_path[ c++ ] = xwp_env + i + 1;
	    }
	}
      xword_path[ c ] = NULL;
    }
}

static void
debug_main (void */*closure*/, int argc, char **argv)
{
  setup_xword_path ();
  hook_up_guile ();
  printf ("For normal startup: (load \"main_window.scm\")\n");

  //(debug-enable 'backtrace)

  scm_shell (argc, argv);
}


     
void main_prog ()
{
  setup_xword_path ();
  hook_up_guile ();
  gh_eval_file (find_the_file ("main_window.scm"));

  //XtMainLoop ();
}


int main (int argc, char *argv[])
{
  int i;
  verbose = 0;
  debug = 0;

  global_argc = argc;
  global_argv = argv;

  i = process_xword_args (argc, argv);
  argc -= i;
  argv += i;

  if (debug == 1)
    scm_boot_guile (argc, argv, debug_main, 0);

  gh_enter (argc, argv, main_prog);

  return 0;
}
